//	(c) 2000-2002 J. Gans (jdg9@cornell.edu), Shalloway Lab, 
//	Cornell University. You can modify and freely distribute 
//	this code, but please keep this header intact.

#include "stdafx.h"
#include "sgraph.h"
#include "SgraphDoc.h"
#include "SgraphView.h"
#include "NonLinFitDialog.h"
#include "NonLinFitOptions.h"
#include "TextOutputDialog.h"
#include "PrestoUtil.h"
#include "MatrixInverse.h"

#include <float.h>

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CNonLinFitDialog dialog

// A wrapper function to enable multi-threaded curve fitting
UINT ThreadCurveFitLM(LPVOID pParam);

// Amount to varry function when computing discrete derivative
#define		DELTA		1.0e-6f

CNonLinFitDialog::CNonLinFitDialog(CView *m_pView)
	: CDialog(CNonLinFitDialog::IDD, NULL)
{
	pView = m_pView;

	CSgraphDoc* pDoc = (CSgraphDoc*)(m_pView->GetDocument());
	ASSERT_VALID(pDoc);

	// Get the current graph
	m_graph_ptr = &(pDoc->graph[pDoc->selected_graph]);

	// No plot has been selected yet
	fit_plot = -1;

	// By default, the user must first specify a function
	// to fit.
	m_fit_step = FIT_FUNCTION;

	// Specify the default algorithm, maximum number of
	// iterations and fit tolerance.
	fit_method = FIT_LEVENBERG_MARQUARDT;
	fit_iter = FIT_MAX_ITER;
	fit_tol = FIT_TOL;
	fit_simplify = TRUE;
	fit_use_y_error_bar = TRUE;

	// Save a plot with 100 points
	num_save_plot_points = 100;

	cur_sel_param = 0;

	m_stop_fit = FALSE;

	//{{AFX_DATA_INIT(CNonLinFitDialog)
	m_chi2 = _T("");
	m_function = _T("");
	m_max_param = _T("");
	m_min_param = _T("");
	m_value_param = _T("");
	//}}AFX_DATA_INIT
}


void CNonLinFitDialog::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CNonLinFitDialog)
	DDX_Control(pDX, IDC_EQUATION_TAB, m_eq_tab);
	DDX_Control(pDX, IDC_PARAM_LIST, m_param);
	DDX_Control(pDX, IDC_PARAM_SLIDER, m_param_slider);
	DDX_Control(pDX, IDC_DATA_LIST, m_data);
	DDX_Text(pDX, IDC_CHI_2, m_chi2);
	DDX_Text(pDX, IDC_FUNC_EDIT, m_function);
	DDX_Text(pDX, IDC_MAX_EDIT, m_max_param);
	DDX_Text(pDX, IDC_MIN_EDIT, m_min_param);
	DDX_Text(pDX, IDC_VALUE_EDIT, m_value_param);
	//}}AFX_DATA_MAP
}


BEGIN_MESSAGE_MAP(CNonLinFitDialog, CDialog)
	//{{AFX_MSG_MAP(CNonLinFitDialog)
	ON_BN_CLICKED(IDC_NEXT_BUTTON, OnNextButton)
	ON_BN_CLICKED(IDC_PREV_BUTTON, OnPrevButton)
	ON_BN_CLICKED(IDC_FIT_BUTTON, OnFitButton)
	ON_NOTIFY(NM_CLICK, IDC_PARAM_LIST, OnClickParamList)
	ON_WM_HSCROLL()
	ON_MESSAGE(WM_STOP_FIT, StopCurveFit)
	ON_MESSAGE(WM_UPDATE_FIT, UpdateCurveFit)
	ON_BN_CLICKED(IDC_OPTIONS_BUTTON, OnOptionsButton)
	ON_LBN_SELCHANGE(IDC_DATA_LIST, OnSelchangeDataList)
	ON_BN_CLICKED(IDC_SAVE_BUTTON, OnSaveButton)
	ON_NOTIFY(TCN_SELCHANGE, IDC_EQUATION_TAB, OnSelchangeEquationTab)
	ON_NOTIFY(TCN_SELCHANGING, IDC_EQUATION_TAB, OnSelchangingEquationTab)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CNonLinFitDialog message handlers
BOOL CNonLinFitDialog::OnInitDialog()
{
	CDialog::OnInitDialog();

	int i, plot_count;			// The number of displayed plots
	int max_string_length = 0;
	CDC *pDC;
	CString buffer;

	// Load all of the active plots in the data list
	for(i = 0, plot_count = 0;i < m_graph_ptr->plot.Length();i++){
		if((plot_count <= MAX_LIST_BOX_ENTRIES) && (m_graph_ptr->plot[i].isDrawn)){
			// The set number is determined when the plots are
			// first created (or read in) - hence use the index number
			// to lable the plot, not the plot_count.
			buffer.Format("Set #%d (N = %d) \"%s\"", i,
				m_graph_ptr->plot[i].data.Length(), 
				LPCTSTR(m_graph_ptr->plot[i].info));
		
			// Keep track of the maximum string length (in pixels).
			// See comments below regarding this kludge.
			int len = buffer.GetLength();
			pDC = m_data.GetDC();

			if(pDC->GetTextExtent(buffer, len).cx > max_string_length)
				max_string_length = pDC->GetTextExtent(buffer, len).cx;
			
			m_data.ReleaseDC(pDC);

			m_data.AddString(buffer);
			plot_count++;
		}
	}
	
	// Explain to users that they've encountered a limitation of the 
	// MFC CListBox class!
	if(plot_count > MAX_LIST_BOX_ENTRIES){
		AfxMessageBox("The maximum number of plots that can be displayed in a\r\n"
			"ListBox has been reached - plot list has been truncated!");
	}

	// Set the width of the list box for proper horizontal
	// scrolling (a kludge to get around horizontal scroll
	// limitations of the CListBox).
	m_data.SetHorizontalExtent(max_string_length);

	// Select the first plot for fitting (if there is at least a single plot
	// available)
	if(plot_count > 0){
		m_data.SetCurSel(0);

		// Select the plot (i.e. plot 0)
		OnSelchangeDataList();
	}

	// Load the equation history tab control
	// Save the default settings in the history list
	SaveEquationHistory();

	// Set the history tabs to reflect the number of equations in 
	// the history list
	for(i = 0;i < ((CSgraphView*)(pView))->fitEqHistory.Length();i++){
		char tmpString[64];

		// Start numbering equations from 1
		sprintf(tmpString, "Eq %d", i+1);

		m_eq_tab.InsertItem(i, tmpString);
	}

	// Select the default tab
	m_eq_tab.SetCurSel(0);

	// Build the columns in the parameter list control
	LVCOLUMN tmpCol;
	float col_width_fudge_factor = 1.2f;
	pDC = m_param.GetDC();

	tmpCol.mask = LVCF_TEXT | LVCF_FMT | LVCF_WIDTH;
	tmpCol.fmt = LVCFMT_LEFT;
	tmpCol.pszText = "name";
	tmpCol.cchTextMax = strlen(tmpCol.pszText);

	tmpCol.cx = (int)(col_width_fudge_factor*
		pDC->GetTextExtent(tmpCol.pszText, tmpCol.cchTextMax).cx);

	tmpCol.iSubItem = 0;	// Ignored
	tmpCol.iImage = 0;	// Ignored
	tmpCol.iOrder = 0;	// Ignored

	m_param.InsertColumn(0, &tmpCol);

	tmpCol.pszText = "value";
	tmpCol.cchTextMax = strlen(tmpCol.pszText);

	tmpCol.cx = (int)(col_width_fudge_factor*
		pDC->GetTextExtent(tmpCol.pszText, tmpCol.cchTextMax).cx);

	m_param.InsertColumn(1, &tmpCol);

	tmpCol.pszText = "min";
	tmpCol.cchTextMax = strlen(tmpCol.pszText);
	tmpCol.cx = (int)(col_width_fudge_factor*
		pDC->GetTextExtent(tmpCol.pszText, tmpCol.cchTextMax).cx);

	m_param.InsertColumn(2, &tmpCol);

	tmpCol.pszText = "max";
	tmpCol.cchTextMax = strlen(tmpCol.pszText);
	tmpCol.cx = (int)(col_width_fudge_factor*
		pDC->GetTextExtent(tmpCol.pszText, tmpCol.cchTextMax).cx);

	m_param.ReleaseDC(pDC);

	m_param.InsertColumn(3, &tmpCol);

	// Set the range of the parameter slider
	m_param_slider.SetRange(0, FIT_SLIDER_SIZE);

	// Make the current set of controls active
	ActivateStage();

	UpdateData(FALSE);

	return TRUE;
}

// The user has selected a plot from the list of available plots
void CNonLinFitDialog::OnSelchangeDataList() 
{
	int i, plot_count;

	// Reset the plot id to a negative value (i.e. no plot selected)
	fit_plot = -1;

	if(!m_graph_ptr)
		throw Error("NULL graph pointer in ChangeStage()");

	// Make sure the user has selected at least one plot to fit
	for(i = 0, plot_count = 0;i < m_graph_ptr->plot.Length();i++)
		if(m_graph_ptr->plot[i].isDrawn){
			if(m_data.GetSel(plot_count)){
				m_graph_ptr->plot[i].isSelected = TRUE;

				// Save the id of the selected plot
				fit_plot = i;
			}
			else{
				m_graph_ptr->plot[i].isSelected = FALSE;
			}
			plot_count ++;
		}
			
	if(fit_plot < 0){
		AfxMessageBox("Please select one data set to fit");
	}

	// Update chi^2
	ChiSquare(TRUE);
}

BOOL CNonLinFitDialog::ChangeStage() 
{
	int i;
	CString error;

	// Collect the current state of all controls
	UpdateData(TRUE);

	// Validate the data at the current stage and move to the next
	// stage
	switch(m_fit_step){
		case FIT_FUNCTION:

			// Make sure the user has entered a valid function to fit
			try{
				fit_eq.ClearError(); // Start with a clean slate

				fit_eq.Set((LPCTSTR)(m_function));
				
				if(fit_eq.GetError())
					throw fit_eq.GetError();

				// Make sure we can define the independant variable
				if(!fit_eq.DefineIndepVar("x"))
					throw "Function does not contain 'x'.";

				// Make sure there is at least a single parameter to fit!
				if(fit_eq.NumVar() == 1){
					throw "Please define at least a single parameter to fit!";
				}

				// Attempt to simplify the equation
				if(fit_simplify){
					fit_eq.Simplify();

					if(fit_eq.GetError())
						throw "Unable to simplify the equation";
				}
			}
			catch(const char* error){
				AfxMessageBox(error);
				return FALSE;
			}
			catch(...){
				AfxMessageBox("Sorry - cannot fit the given function!");
				return FALSE;
			}

			// If we get this far, we have a valid equation and data set(s)

			// Make sure the parameter list is empty
			while(param.Length() > 0)
				param.Pop();

			// Load the parameter list
			for(i = 0;i < fit_eq.NumVar();i++){
				if(!fit_eq.IsIndepVar(i)){
					// Build the parameter list
					FitParam tmpParam;

					tmpParam.index = i;

					param.Push(tmpParam);
				}
			}
			
			// Build the param list control
			SetParamList(TRUE);
			break;
		case FIT_SET_PARAM:
			
			// Load the last set of current parameter values
			SaveCurParam();

			// Make sure that the initial parameter entries
			// provided by the user are valid
			for(i = 0;i < param.Length();i++){
				if(!param[i].is_val){
					error.Format("Please provide an initial estimate for '%s'",
						fit_eq.GetVarName(param[i].index));
					AfxMessageBox(error);
					return FALSE;
				}

				if(param[i].is_min && (param[i].min_val > param[i].val)){
					error.Format("Initial estimate for '%s' is below lower bound",
						fit_eq.GetVarName(param[i].index));
					AfxMessageBox(error);
					return FALSE;
				}

				if(param[i].is_max && (param[i].max_val < param[i].val)){
					error.Format("Initial estimate for '%s' is above upper bound",
						fit_eq.GetVarName(param[i].index));
					AfxMessageBox(error);
					return FALSE;
				}

				// Set the maximum value
				if(param[i].is_max){
					param[i].adjust_max = param[i].max_val;
				}
				else{
					// Estimate the range based on fractional size of the parameter
					// value
					if(param[i].val == 0.0f){
						param[i].adjust_max = 1.0f;
					}
					else{
						param[i].adjust_max = (1.0f + FIT_SLIDER_FRACT_SIZE)*param[i].val;
					}
				}

				// Set the minimum value
				if(param[i].is_min){
					param[i].adjust_min = param[i].min_val;
				}
				else{
					// Estimate the range based on fractional size of the parameter
					// value
					if(param[i].val == 0.0f){
						param[i].adjust_min = -1.0f;
					}
					else{
						param[i].adjust_min = (1.0f - FIT_SLIDER_FRACT_SIZE)*param[i].val;
					}
				}
			}

			SetSlider();

			// If we get to this point, the parameters are valid.
			// Display the function with the initial parameter 
			// values.
			PlotFit();

			break;
		case FIT_ADJUST_PARAM:
			break;
	}

	// Turn on (and off) the appropriate controls
	ActivateStage();

	return TRUE;
}

void CNonLinFitDialog::OnNextButton() 
{
	if(ChangeStage()){
		// Change the current stage of the fitting process
		m_fit_step ++;

		// Turn on (and off) the appropriate controls
		ActivateStage();
	}
}

void CNonLinFitDialog::OnPrevButton() 
{
	if(m_fit_step > FIT_FUNCTION){
		// Change the current stage of the fitting process
		m_fit_step --;

		// Turn on (and off) the appropriate controls
		ActivateStage();
	}
}

// Build the parameter list control
void CNonLinFitDialog::SetParamList(BOOL m_initList)
{
	CString buffer;
	int max_string_length = 0;
	int i;

	if(m_initList){
		// Delete any existing elements of the list
		m_param.DeleteAllItems();
	}

	// Load all of the active plots in the data list
	for(i = 0;i < param.Length();i++){
		SetParamItem(i, m_initList);
	}

	if(m_initList){
		// Set the current parameter
		cur_sel_param = 0;	// Make the first parameter the current
							// parameter.
		LoadCurParam();
	}
}

void CNonLinFitDialog::ActivateStage()
{
	switch(m_fit_step){
		case FIT_FUNCTION:
			GetDlgItem(IDC_FUNC_EDIT)->EnableWindow(TRUE);
			GetDlgItem(IDC_EQUATION_TAB)->EnableWindow(TRUE);

			GetDlgItem(IDC_PARAM_LIST)->EnableWindow(FALSE);
			GetDlgItem(IDC_MIN_EDIT)->EnableWindow(FALSE);
			GetDlgItem(IDC_VALUE_EDIT)->EnableWindow(FALSE);
			GetDlgItem(IDC_MAX_EDIT)->EnableWindow(FALSE);

			GetDlgItem(IDC_PARAM_SLIDER)->EnableWindow(FALSE);

			GetDlgItem(IDC_PREV_BUTTON)->EnableWindow(FALSE);
			GetDlgItem(IDC_NEXT_BUTTON)->EnableWindow(TRUE);
			GetDlgItem(IDC_FIT_BUTTON)->EnableWindow(FALSE);
			GetDlgItem(IDC_SAVE_BUTTON)->EnableWindow(FALSE);
			break;
		case FIT_SET_PARAM:
			GetDlgItem(IDC_FUNC_EDIT)->EnableWindow(FALSE);
			GetDlgItem(IDC_EQUATION_TAB)->EnableWindow(FALSE);

			GetDlgItem(IDC_PARAM_LIST)->EnableWindow(TRUE);
			GetDlgItem(IDC_MIN_EDIT)->EnableWindow(TRUE);
			GetDlgItem(IDC_VALUE_EDIT)->EnableWindow(TRUE);
			GetDlgItem(IDC_MAX_EDIT)->EnableWindow(TRUE);

			GetDlgItem(IDC_PARAM_SLIDER)->EnableWindow(FALSE);

			GetDlgItem(IDC_PREV_BUTTON)->EnableWindow(TRUE);
			GetDlgItem(IDC_NEXT_BUTTON)->EnableWindow(TRUE);
			GetDlgItem(IDC_FIT_BUTTON)->EnableWindow(FALSE);
			GetDlgItem(IDC_SAVE_BUTTON)->EnableWindow(FALSE);
			break;
		case FIT_ADJUST_PARAM:
			GetDlgItem(IDC_FUNC_EDIT)->EnableWindow(FALSE);
			GetDlgItem(IDC_EQUATION_TAB)->EnableWindow(FALSE);

			GetDlgItem(IDC_PARAM_LIST)->EnableWindow(TRUE);
			GetDlgItem(IDC_MIN_EDIT)->EnableWindow(FALSE);
			GetDlgItem(IDC_VALUE_EDIT)->EnableWindow(FALSE);
			GetDlgItem(IDC_MAX_EDIT)->EnableWindow(FALSE);

			GetDlgItem(IDC_PARAM_SLIDER)->EnableWindow(TRUE);

			GetDlgItem(IDC_PREV_BUTTON)->EnableWindow(TRUE);
			GetDlgItem(IDC_NEXT_BUTTON)->EnableWindow(FALSE);
			GetDlgItem(IDC_FIT_BUTTON)->EnableWindow(TRUE);
			GetDlgItem(IDC_SAVE_BUTTON)->EnableWindow(TRUE);
			break;
		case FIT_CURVE_FIT:
			GetDlgItem(IDC_FUNC_EDIT)->EnableWindow(FALSE);
			GetDlgItem(IDC_EQUATION_TAB)->EnableWindow(FALSE);

			GetDlgItem(IDC_PARAM_LIST)->EnableWindow(FALSE);
			GetDlgItem(IDC_MIN_EDIT)->EnableWindow(FALSE);
			GetDlgItem(IDC_VALUE_EDIT)->EnableWindow(FALSE);
			GetDlgItem(IDC_MAX_EDIT)->EnableWindow(FALSE);

			GetDlgItem(IDC_PARAM_SLIDER)->EnableWindow(FALSE);

			GetDlgItem(IDC_PREV_BUTTON)->EnableWindow(FALSE);
			GetDlgItem(IDC_NEXT_BUTTON)->EnableWindow(FALSE);
			GetDlgItem(IDC_FIT_BUTTON)->EnableWindow(FALSE);
			GetDlgItem(IDC_SAVE_BUTTON)->EnableWindow(FALSE);
			break;
	}
}

void CNonLinFitDialog::OnClickParamList(NMHDR* pNMHDR, LRESULT* pResult) 
{
	// The user has clicked an entry in the list control.
	// Find out which item in the list is currently selected.

	// Save the current parameter values
	SaveCurParam();

	if(m_param.GetSelectedCount() == 0){
		// There are no currently selected parameters, so select the 
		// first parameter

		m_param.SetItemState(0, LVIS_SELECTED,LVIS_SELECTED);
		cur_sel_param = 0;

	}
	else{
		// Get the currently selected item
		POSITION pos = m_param.GetFirstSelectedItemPosition();
		
		cur_sel_param = m_param.GetNextSelectedItem(pos);
	}

	LoadCurParam();
	
	// Do we need to initialize the parameter slider?
	if(m_fit_step == FIT_ADJUST_PARAM){
		// Set the maximum value
		if(param[cur_sel_param].is_max){
			param[cur_sel_param].adjust_max = 
				param[cur_sel_param].max_val;
		}
		else{
			// Estimate the range based on fractional size of the parameter
			// value
			if(param[cur_sel_param].val == 0.0f){
				param[cur_sel_param].adjust_max = 1.0f;
			}
			else{
				param[cur_sel_param].adjust_max = 
					(1.0f + FIT_SLIDER_FRACT_SIZE)*param[cur_sel_param].val;
			}
		}

		// Set the minimum value
		if(param[cur_sel_param].is_min){
			param[cur_sel_param].adjust_min = 
				param[cur_sel_param].min_val;
		}
		else{
			// Estimate the range based on fractional size of the parameter
			// value
			if(param[cur_sel_param].val == 0.0f){
				param[cur_sel_param].adjust_min = -1.0f;
			}
			else{
				param[cur_sel_param].adjust_min = 
					(1.0f - FIT_SLIDER_FRACT_SIZE)*param[cur_sel_param].val;
			}
		}

		SetSlider();
	}

	*pResult = 0;
}


void CNonLinFitDialog::SaveCurParam()
{
	if(m_fit_step == FIT_SET_PARAM){
		UpdateData(TRUE);

		// Is the value a string or valid number?
		param[cur_sel_param].is_val = 
			StringToFloat(m_value_param, param[cur_sel_param].val);

		// Is the max value a string or valid number?
		param[cur_sel_param].is_max = 
			StringToFloat(m_max_param, param[cur_sel_param].max_val);

		// Is the min value a string or valid number?
		param[cur_sel_param].is_min = 
			StringToFloat(m_min_param, param[cur_sel_param].min_val);
	}

	SetParamItem(cur_sel_param, FALSE /* Don't insert a new item */);
}

void CNonLinFitDialog::LoadCurParam()
{
	if(param[cur_sel_param].is_val){
		m_value_param.Format("%g", param[cur_sel_param].val);
	}
	else{
		m_value_param.Format("???");
	}

	if(param[cur_sel_param].is_max){
		m_max_param.Format("%g", param[cur_sel_param].max_val);
	}
	else{
		m_max_param.Format("---");
	}

	if(param[cur_sel_param].is_min){
		m_min_param.Format("%g", param[cur_sel_param].min_val);
	}
	else{
		m_min_param.Format("---");
	}

	// Set the dialog items
	UpdateData(FALSE);
}

void CNonLinFitDialog::SetParamItem(int m_i, BOOL m_initList)
{
	CString buffer;
	LVITEM tmpLVITEM;
		
	// Zero out the LVITEM structure
	memset(&tmpLVITEM, 0, sizeof(LVITEM));

	// Add the parameter name
	tmpLVITEM.mask = LVIF_TEXT;
	tmpLVITEM.iItem = m_i;
	tmpLVITEM.iSubItem = 0; // The item itself
	tmpLVITEM.state = 0;
	tmpLVITEM.stateMask = LVIS_SELECTED;
	tmpLVITEM.pszText = fit_eq.GetVarName(param[m_i].index);

	if(m_initList)
		m_param.InsertItem(&tmpLVITEM);
	else
		m_param.SetItem(&tmpLVITEM);

	// Add the parameter value
	tmpLVITEM.mask = LVIF_TEXT;
	tmpLVITEM.iItem = m_i;
	tmpLVITEM.iSubItem = 1; // The value
	tmpLVITEM.state = 0;
	tmpLVITEM.stateMask = LVIS_SELECTED;

	if(param[m_i].is_val){
		buffer.Format("%g", param[m_i].val);
		tmpLVITEM.pszText = buffer.GetBuffer(0);
	}
	else{
		tmpLVITEM.pszText = "???";
	}

	m_param.SetItem(&tmpLVITEM);

	// Add the parameter min value
	tmpLVITEM.mask = LVIF_TEXT;
	tmpLVITEM.iItem = m_i;
	tmpLVITEM.iSubItem = 2; // The min value
	tmpLVITEM.state = 0;
	tmpLVITEM.stateMask = LVIS_SELECTED;

	if(param[m_i].is_min){
		buffer.Format("%g", param[m_i].min_val);
		tmpLVITEM.pszText = buffer.GetBuffer(0);
	}
	else{
		tmpLVITEM.pszText = "---";
	}

	m_param.SetItem(&tmpLVITEM);

	// Add the parameter min value
	tmpLVITEM.mask = LVIF_TEXT;
	tmpLVITEM.iItem = m_i;
	tmpLVITEM.iSubItem = 3; // The max value
	tmpLVITEM.state = 0;
	tmpLVITEM.stateMask = LVIS_SELECTED;

	if(param[m_i].is_max){
		buffer.Format("%g", param[m_i].max_val);
		tmpLVITEM.pszText = buffer.GetBuffer(0);
	}
	else{
		tmpLVITEM.pszText = "---";
	}

	m_param.SetItem(&tmpLVITEM);
}

void CNonLinFitDialog::SetSlider()
{
	float val = param[cur_sel_param].val;
	float max = param[cur_sel_param].adjust_max;
	float min = param[cur_sel_param].adjust_min;

	float dp = max - min;

	if(dp == 0.0f){
		// min == max, so disable the slider!
		GetDlgItem(IDC_PARAM_SLIDER)->EnableWindow(FALSE);
	}
	else{
		// min != max, enable the slider
		GetDlgItem(IDC_PARAM_SLIDER)->EnableWindow(TRUE);
	}

	int pos = (int)(FIT_SLIDER_SIZE*(val - min)/dp);

	if(pos < 0)
		pos = 0;

	if(pos > FIT_SLIDER_SIZE)
		pos = FIT_SLIDER_SIZE;

	m_param_slider.SetPos(pos);
}

void CNonLinFitDialog::OnHScroll(UINT nSBCode, UINT nPos, CScrollBar* pScrollBar) 
{
	// The user has adjusted a parameter
	float max = param[cur_sel_param].adjust_max;
	float min = param[cur_sel_param].adjust_min;

	float dp = max - min;

	param[cur_sel_param].val = m_param_slider.GetPos()*dp/FIT_SLIDER_SIZE + min;

	SaveCurParam();

	// Load the current parameter values into the text boxes
	LoadCurParam();

	PlotFit();

	CDialog::OnHScroll(nSBCode, nPos, pScrollBar);
}

float CNonLinFitDialog::PlotFit()
{
	if(!pView)
		throw Error("NULL view in PlotFit()");
	
	if(!m_graph_ptr)
		throw Error("NULL graph ptr in PlotFit()");

	// Set the parameters of the equation
	for(int i = 0;i < param.Length();i++)
		fit_eq.SetVar(param[i].index, param[i].val);

	m_graph_ptr->QuickPlot(fit_eq);
	((CSgraphView*)(pView))->ForceRedraw();

	return ChiSquare(TRUE);
}

void CNonLinFitDialog::OnCancel() 
{
	if(!pView)
		throw Error("NULL view in PlotFit()");
	
	if(!m_graph_ptr)
		throw Error("NULL graph ptr in PlotFit()");

	if(m_fit_step == FIT_CURVE_FIT){
		// We're trying to exit while performing a fit.
		// Make sure the fitting thread is closed before we quit.
		m_stop_fit = TRUE;
	}
	else{ // A "normal exit"
		m_graph_ptr->ClearQuickPlot();
		((CSgraphView*)(pView))->ForceRedraw();

		CDialog::OnCancel();
	}
}

float CNonLinFitDialog::ChiSquare(BOOL m_display)
{
	if(!m_graph_ptr)
		throw Error("NULL graph ptr in ChiSquare()");

	if(m_fit_step < FIT_ADJUST_PARAM){
		// The user is either entering a function (i.e. FIT_FUNCTION)
		// or specifying initial conditions (i.e. FIT_SET_PARAM), so
		// don't attempt to compute ChiSquare!
		return -1.0f;
	}

	if(!fit_eq.IsValid()){
		if(m_display){
			m_chi2.Format("Invalid Equation!");
			UpdateData(FALSE);
		}
		
		return -1.0f;
	}

	if(fit_plot < 0){
		if(m_display){
			m_chi2.Format("Select a plot!");
			UpdateData(FALSE);
		}
		
		return -1.0f;
	}

	stackIterator<DataPoint> next(m_graph_ptr->plot[fit_plot].data);
	DataPoint xy;
	
	float chi2 = 0.0f;

	// Does this data set contain error bars in y?
	stackIterator<float> nextError(m_graph_ptr->plot[fit_plot].yErrorBar);

	BOOL use_error_bars = fit_use_y_error_bar && 
		!(m_graph_ptr->plot[fit_plot].yErrorBar.IsEmpty());

	while(next(xy)){
		float sigma;

		if(use_error_bars){
			nextError(sigma);
		}
		else
			sigma = DEFAULT_SIGMA;

		float tmp = fit_eq(xy.x);

		if(fit_eq.GetError()){
			fit_eq.ClearError();
			
			if(m_display){
				m_chi2.Format("Could not compute");
				UpdateData(FALSE);
			}
			
			return -1.0f;
		}

		tmp = xy.y - tmp;

		if(sigma != 0.0f)
			tmp	/= sigma;

		chi2 += tmp*tmp;
	}

	if(m_display){
		m_chi2.Format("%g", chi2);
		UpdateData(FALSE);
	}
	
	return chi2;
}

void CNonLinFitDialog::OnFitButton() 
{
	if(!m_graph_ptr)
		throw Error("NULL graph ptr in OnFitButton()");

	// Make sure the equation is valid
	if(!fit_eq.IsValid()){
		AfxMessageBox("Invalid equation passed to fitting engine");
		return;
	}
	
	// Save this equation information for the equation history list
	SaveEquationHistory();

	// Deactivate controlls while we're fitting
	m_fit_step = FIT_CURVE_FIT;

	ActivateStage();

	AfxBeginThread(ThreadCurveFitLM, this);
}

void CNonLinFitDialog::StartCurveFit() 
{
	try{
		// Perform a curve fit based on the selected method
		switch(fit_method){
			case FIT_LEVENBERG_MARQUARDT:
			default:
				CurveFitLM();
				break;
		}
	}
	catch(Error &e){
		AfxMessageBox(e.message());
	}
	catch(char *error){
		AfxMessageBox(error);
	}
	catch(...){
		AfxMessageBox("A vauge and unspecified error has occured while fitting");
	}

	this->PostMessage(WM_STOP_FIT);
}

UINT ThreadCurveFitLM(LPVOID pParam)
{
	if(pParam){
		((CNonLinFitDialog*)(pParam))->StartCurveFit();
		return 0;	// Success
	}

	return 1; // Failure
}

LRESULT CNonLinFitDialog::UpdateCurveFit(WPARAM wParam, LPARAM lParam)
{
	m_graph_ptr->QuickPlot(fit_eq);
	((CSgraphView*)(pView))->ForceRedraw();

	UpdateData(FALSE);

	return 0L;
}

LRESULT CNonLinFitDialog::StopCurveFit(WPARAM wParam, LPARAM lParam)
{
	// Do we need to quit the fit ASAP?
	if(m_stop_fit){
		m_fit_step = FIT_ADJUST_PARAM;

		// Close the dialog
		OnCancel();

		return 0L;
	}

	// The curve fit has succeded! Update the parameter list
	for(int i = 0;i < param.Length();i++){
		SetParamItem(i, FALSE);
	}

	// Update the slider to make sure the user can still
	// adjust parameters.
	SetSlider();

	// Put the current parameter values into the text boxes
	LoadCurParam();

	// Restore the controls
	m_fit_step = FIT_ADJUST_PARAM;

	ActivateStage();

	// Update the graph
	m_graph_ptr->QuickPlot(fit_eq);
	((CSgraphView*)(pView))->ForceRedraw();

	return 0L;
}

void CNonLinFitDialog::CurveFitLM()
{
	int i, j;
	int iter = 0;
	int n = param.Length(); // The number of parameters
	float lambda = 0.001f;
	float chi2;
	BOOL converged = FALSE;

	// Build an array of function derivatives
	array<SEquation> deriv(n);
	
	for(i = 0;i < n;i++){
		deriv[i] = fit_eq.Deriv(param[i].index);

		if(!deriv[i].IsValid())
			throw Error("Unable to compute derivative");

		if(fit_simplify){
			deriv[i].Simplify();

			if(!deriv[i].IsValid())
				throw Error("Unable to simplify derivative");
		}
	}

	//  Solve the following set of linear equations:
	//	
	//	alpha.delta = beta
	//
	// for the vector delta (the amount we need to change the parameters).
	Matrix<float> alpha(n, n);
	Vector<float> beta(n);
	Vector<Subscript> index(n);

	// The initial value of chi^2
	chi2 = ChiSquare(FALSE);

	// Update the graph by sending a message to the worker thread (i.e. the
	// view in this case)
	m_chi2.Format("%g", chi2);
	this->PostMessage(WM_UPDATE_FIT);

	while(!converged){

		if(m_stop_fit)
			return;

		// Compute alpha
		ComputeAlpha(alpha, lambda, deriv);

		// Compute beta
		ComputeBeta(beta, deriv);

		// Solve the set of linear equations for delta
		if(LU_factor(alpha, index)){
			throw Error("LU_factor: Failed to factor alpha");
		}

		if(LU_solve(alpha, index, beta)){
			throw Error("LU_solve: Failed to solve for beta");
		}

		// Have we improved the solution?
		for(i = 0;i < n;i++){
			// Update the function but make sure any user imposed
			// constraints are not violated.
			float tmp = param[i].val + beta[i];
			
			// Reset the edge detection flags
			param[i].max_contrained = FALSE;
			param[i].min_contrained = FALSE;

			if(param[i].is_max){
				if(tmp > param[i].max_val){
					param[i].max_contrained = TRUE;
					beta[i] = param[i].max_val - param[i].val;
					tmp = param[i].max_val;
				}
			}
			
			if(param[i].is_min){
				if(tmp < param[i].min_val){
					param[i].min_contrained = TRUE;
					beta[i] = param[i].min_val - param[i].val;
					tmp = param[i].min_val;
				}
			}
			
			fit_eq.SetVar(param[i].index, tmp);
		}

		// Has the user hit cancel to terminate the thread
		if(m_stop_fit)
			return;

		float tmp_chi2 = ChiSquare(FALSE);

		// Update the graph - never mind for now. Will need a worker thread to
		// do this correctly.
		//m_graph_ptr->QuickPlot(fit_eq);
		//((CSgraphView*)(pView))->ForceRedraw();
		m_chi2.Format("%g", chi2);
		this->PostMessage(WM_UPDATE_FIT);

		if(tmp_chi2 > chi2){
			// No improvement
			lambda *= 10.0f;

			for(i = 0;i < n;i++){
				// Updating the function
				fit_eq.SetVar(param[i].index, param[i].val);
			}
		}
		else{
			// Improvement
			lambda /= 10.0f;

			// Update the parameter values and compute the value of chi^2.
			for(i = 0;i < n;i++)
				param[i].val += beta[i];

			for(i = 0;i < n;i++){
				// Updating the function
				fit_eq.SetVar(param[i].index, param[i].val);

				// Updating the function derivates
				for(j = 0;j < n;j++)
					deriv[j].SetVar(param[i].index, param[i].val);
			}

			if(fabs(tmp_chi2 - chi2) < fit_tol){
				converged = TRUE;

				// Re compute alpha with lambda == 0.0 to get the
				// covariance matrix
				ComputeAlpha(alpha, 0.0f, deriv);
			}

			// Update the current value of lambda
			chi2 = tmp_chi2;
		}

		iter++;

		if(iter > fit_iter)
			throw Error("Maximum number of iterations exceeded!");
	}

	// Has the user hit cancel to terminate the thread
	if(m_stop_fit)
		return;

	// Invert the alpha matrix (which was computed with lambda = 0).
	//alpha = alpha.Inverse();
	InvertMatrix(alpha);
	
	// Format the results for the user
	CString output;
	CString tmpbuf;

	output.Format("Performed a %d parameter Levenberg-Marquardt fit.\n\n", n);

	tmpbuf.Format("Fitting f(x) = %s\n", m_function);
	output += tmpbuf;

	tmpbuf.Format("to data set [%d]: %s.\n\n", fit_plot, m_graph_ptr->plot[fit_plot].info);
	output += tmpbuf;

	tmpbuf.Format("Iterations required for convergence = %d\n", iter);
	output += tmpbuf;

	tmpbuf.Format("Fit tolerance = %g\n\n", fit_tol);
	output += tmpbuf;

	tmpbuf.Format("Minimized value of chi squared = %g\n\n", chi2);
	output += tmpbuf;

	for(i = 0;i < n;i++){
		
		// Compute the error as the sqrt of the covariance matrix
		// The absolute value makes this robust against round off error.
		float tmp_error = (float)(sqrt(fabs(alpha[i][i])));

		if(param[i].max_contrained){
			tmpbuf.Format("** Constrained parameter: %s = max(%s) **\n", 
				fit_eq.GetVarName(param[i].index),
				fit_eq.GetVarName(param[i].index));

			output += tmpbuf;
		}

		if(param[i].min_contrained){
			tmpbuf.Format("** Constrained parameter: %s = min(%s) **\n", 
				fit_eq.GetVarName(param[i].index),
				fit_eq.GetVarName(param[i].index));

			output += tmpbuf;
		}

		tmpbuf.Format("%s = %g +/- %g\n\n", 
			fit_eq.GetVarName(param[i].index),
				param[i].val, tmp_error);

		output += tmpbuf;
	}

	CTextOutputDialog outDlg("Curve Fit Results", output);

	// Pop up a window to let the user know the good news
	outDlg.DoModal();
}

void CNonLinFitDialog::ComputeAlpha(Matrix<float> &alpha, float lambda, 
									array<SEquation> &deriv)
{
	BOOL use_error_bars = fit_use_y_error_bar && 
		!(m_graph_ptr->plot[fit_plot].yErrorBar.IsEmpty());

	int i, j;
	int n = deriv.Length();

	for(i = 0;i < n;i++){
		for(j = 0;j < n;j++){
			
			// Initialize the matrix element to zero
			alpha[i][j] = 0.0f;

			stackIterator<DataPoint> next(m_graph_ptr->plot[fit_plot].data);
			stackIterator<float> nextError(m_graph_ptr->plot[fit_plot].yErrorBar);
			DataPoint xy;

			while(next(xy)){
				float sigma;

				if(use_error_bars){
					nextError(sigma);
				}
				else
					sigma = DEFAULT_SIGMA;

				float tmp_ij, tmp;

				if(i == j){
					// We only need to compute the derivative once when i == j
					tmp = deriv[i](xy.x);

					if(deriv[i].GetError()){
						deriv[i].ClearError();

						// Compute a discrete derivative in the event
						// the analytic form runs into trouble
						tmp = DiscreteDeriv(i, xy.x);
					}

					tmp_ij = tmp*tmp;
				}
				else{
					//tmp = deriv[i](xy.x)*deriv[j](xy.x);

					tmp_ij = deriv[i](xy.x);

					if(deriv[i].GetError()){
						deriv[i].ClearError();

						// Compute a discrete derivative in the event
						// the analytic form runs into trouble
						tmp_ij = DiscreteDeriv(i, xy.x);
					}

					tmp = deriv[j](xy.x);

					if(deriv[j].GetError()){
						deriv[j].ClearError();

						// Compute a discrete derivative in the event
						// the analytic form runs into trouble
						tmp = DiscreteDeriv(j, xy.x);
					}

					tmp_ij *= tmp;
				}

				if(sigma != 0.0f)
					alpha[i][j] += tmp_ij/(sigma*sigma);
				else
					alpha[i][j] += tmp_ij;
			}

			if(i == j)
				alpha[i][i] *= (1.0f + lambda);
		}
	}
}

void CNonLinFitDialog::ComputeBeta(Vector<float> &beta, array<SEquation> &deriv)
{
	BOOL use_error_bars = fit_use_y_error_bar && 
		!(m_graph_ptr->plot[fit_plot].yErrorBar.IsEmpty());

	int i;
	int n = deriv.Length();

	for(i = 0;i < n;i++){
		stackIterator<DataPoint> next(m_graph_ptr->plot[fit_plot].data);
		stackIterator<float> nextError(m_graph_ptr->plot[fit_plot].yErrorBar);
		DataPoint xy;

		// Initialize the vector element to zero
		beta[i] = 0.0f;

		while(next(xy)){
			float sigma;

			if(use_error_bars){
				nextError(sigma);
			}
			else
				sigma = DEFAULT_SIGMA;

			float tmp = xy.y - fit_eq(xy.x);

			if(fit_eq.GetError())
				throw Error("Unable to compute function.");

			float tmp_deriv = deriv[i](xy.x);
			
			if(deriv[i].GetError()){
				deriv[i].ClearError();
				
				// Compute a discrete derivative in the event
				// the analytic form runs into trouble
				tmp_deriv = DiscreteDeriv(i, xy.x);
			}

			tmp *= tmp_deriv;

			if(sigma != 0.0f)
				beta[i] += tmp/(sigma*sigma);
			else
				beta[i] += tmp;
		}
	}
}

float CNonLinFitDialog::DiscreteDeriv(int i, float x)
{
	float tmp;
	float param_value;

	try{
		param_value = fit_eq.GetVar(i);

		fit_eq.SetVar(i, param_value + DELTA);
		
		if(fit_eq.GetError()){
			throw fit_eq.GetError();
		}

		tmp = fit_eq(x);

		fit_eq.SetVar(i, param_value - DELTA);

		tmp -= fit_eq(x);

		if(fit_eq.GetError()){
			throw fit_eq.GetError();
		}

		tmp /= 2.0f*DELTA;

		if(!_finite(tmp)){
			throw "Bad discrete derivative";
		}
	}
	catch(...){
		fit_eq.ClearError();
		fit_eq.SetVar(i, param_value);

		throw "Unable to compute discrete derivative!";
	}

	// Restore the original parameter value
	fit_eq.SetVar(i, param_value);

	return tmp;
}

void CNonLinFitDialog::OnOptionsButton() 
{
	// Display/Modify the fit options
	CNonLinFitOptions nlfoDlg(fit_method, fit_iter, fit_tol, 
		fit_simplify, fit_use_y_error_bar);

	if(nlfoDlg.DoModal() == IDOK){
		fit_iter = nlfoDlg.GetMaxIter();
		fit_tol = nlfoDlg.GetTol();
		fit_method = nlfoDlg.GetMethod();
		fit_simplify = nlfoDlg.Simplify();
		fit_use_y_error_bar = nlfoDlg.UseErrorBar();
	}
}

void CNonLinFitDialog::OnSaveButton() 
{
	// Save the current fit to the current graph
	if(!pView)
		throw Error("NULL view in PlotFit()");
	
	if(!m_graph_ptr)
		throw Error("NULL graph ptr in PlotFit()");

	// Set the parameters of the equation
	for(int i = 0;i < param.Length();i++)
		fit_eq.SetVar(param[i].index, param[i].val);


	// Prepare to evaluate the equation
	SPlot newPlot;
	DataPoint xy;
	BOOL first_time = TRUE;
	float dx, minX, maxX, minY;
	
	BOOL use_log_scale = 
		m_graph_ptr->plot[fit_plot].x_scaleType == SCALE_LOG;

	// Set the X range according to the plot that is
	// being fitted.
	minX = m_graph_ptr->plot[fit_plot].minX;
	maxX = m_graph_ptr->plot[fit_plot].maxX;
	
	dx = (maxX - minX)/num_save_plot_points;

	for(float x = minX;x <= maxX;x += dx){
		if(use_log_scale){
			// Log scale
			xy.x = (float)(pow(10.0, x));
			
		}else{
			// Linear scale
			xy.x = x;
		}

		xy.y = fit_eq(xy.x);

		// Did we successfully evaluate the function?
		if(fit_eq.GetError()){
			// Nope. A error was encountered
			fit_eq.ClearError();
		}
		else{
			// Yes! Save the function value in the new plot
			newPlot.data.Push(xy);
				
			// Track the smallest value to see if we can plot the
			// equation on a log plot.
			if(first_time){
				first_time = FALSE;

				minY = xy.y;
			}
			else{
				if(xy.y < minY)
					minY = xy.y;
			}
		}
	}
		
	// If we never evaluated the function, don't add the plot!
	if(first_time == TRUE){
		AfxMessageBox("Unable to evalute the function"
			"within the given range of x");
		return;
	}

	// Give the plot a name (use the equation)
	newPlot.info.Format("Curve fit to plot %d", fit_plot);

	// Set the plot style
	newPlot.pointType = SPlot::POINT_NONE;
	newPlot.lineColor = 0x00000000; // Black
	newPlot.lineType = SPlot::LINE_SOLID;
	
	// Set the scale type from the current graph
	newPlot.x_scaleType = m_graph_ptr->x_scaleType;
	newPlot.y_scaleType = m_graph_ptr->y_scaleType;

	// Note that the use_log_scale applies only to the x axis
	// and is set if we a currently plotting data against log(x).
	if(minX <= 0.0f && !use_log_scale){
		newPlot.canXLog = FALSE;

		// Adjust the current graph if needed
		if(m_graph_ptr->x_scaleType == SCALE_LOG)
			m_graph_ptr->x_scaleType = SCALE_LINEAR;
	}
	else
		newPlot.canXLog = TRUE;

	if(minY <= 0.0f){
		newPlot.canYLog = FALSE;

		// Adjust the current graph if needed
		if(m_graph_ptr->y_scaleType == SCALE_LOG)
			m_graph_ptr->y_scaleType = SCALE_LINEAR;
	}
	else
		newPlot.canYLog = TRUE;

	// Add the plot of the function to the current graph
	m_graph_ptr->plot.Push(newPlot);

	m_graph_ptr->SetPlotScaleType();
	m_graph_ptr->SetPlotRange();
	m_graph_ptr->SetBoundingBox();
}

void CNonLinFitDialog::SaveEquationHistory()
{
	FitEquationInfo tmp;

	tmp.m_equation = m_function;

	// Make sure the current info doesn't match an existing entry
	for(int i = 0;i < ((CSgraphView*)(pView))->fitEqHistory.Length();i++){
		if(tmp == ((CSgraphView*)(pView))->fitEqHistory[i])
			return;
	}

	// Append the most recent equation behind the head of the list
	// unless the equation is either (a) blank or (b) the first equation (in both
	// cases, this equation goes at the HEAD of the list).
	if((m_function == _T("")) || (((CSgraphView*)(pView))->fitEqHistory.Length() == 0))
		((CSgraphView*)(pView))->fitEqHistory.Insert(tmp, 0);
	else
		((CSgraphView*)(pView))->fitEqHistory.Insert(tmp, 1);
}

void CNonLinFitDialog::OnSelchangeEquationTab(NMHDR* pNMHDR, LRESULT* pResult) 
{
	// The user has selected an equation from the history list, update
	// the dialog
	m_function = 
		((CSgraphView*)(pView))->fitEqHistory[m_eq_tab.GetCurSel()].m_equation;

	// Update all the values we just set
	UpdateData(FALSE);
	
	*pResult = 0;
}

void CNonLinFitDialog::OnSelchangingEquationTab(NMHDR* pNMHDR, LRESULT* pResult) 
{
	// Read the values set by the user
	UpdateData(TRUE);

	// The user has selected an equation from the history list, update
	// the dialog before switching to the new equation. This keeps the history
	// list current and allows the user to switch back and forth without losing
	// any values.
	((CSgraphView*)(pView))->fitEqHistory[m_eq_tab.GetCurSel()].m_equation = 
		m_function;

	*pResult = 0;
}
