//	(c) 2000 J. Gans (jdg9@cornell.edu), Shalloway Lab, 
//	Cornell University. You can modify and freely distribute 
//	this code, but please keep this header intact.

// This file contains the SgraphDoc member function(s) required for
// pipe input and output. Note that we can't just connect the pipe to the
// import plot functions because the pipe is transmitting binary data.

#include "stdafx.h"
#include <math.h>
#include "MainFrm.h"
#include "sgraph.h"
#include "SGraphDoc.h"
#include "SGraphView.h"
#include "PrestoLink/PrestoLink.h"
#include "color.h"

void CSgraphDoc::ListenToPipe()
{
	char msg;

	// For now, import data to graph 0
	int curr_graph = 0;

	int curr_plot = 0;
	int curr_point = 0;
	CString tmp_text;

	// Transmission parameters
	DWORD bytes_read;
	BOOL local_status = TRUE;

	// Should we close at exit?
	BOOL send_exit_msg = FALSE;

	// Should we save the file at exit?
	// if(save_file_at_exit == -1) -> Don't save
	// if(save_file_at_exit >= 0) -> Save in the requested format
	int save_file_at_exit = -1;

	CString filename;

	// World scaling parameters
	BOOL set_scale_x = FALSE;
	BOOL set_scale_y = FALSE;
	float min_x, max_x, min_y, max_y;

	// Tick mark spacing parameters
	BOOL set_tick_x = FALSE;
	BOOL set_tick_y = FALSE;
	float major_tick_x, minor_tick_x, major_tick_y, minor_tick_y;

	char r, g, b;
	char use_x_error_bar, use_y_error_bar;
	int tmp_int;

	int text_buffer_length;
	char *text_buffer = NULL;

	DataPoint tmp_dp;
	float x_error, y_error;

	SPlot *plot_ptr = NULL;

	list<SPlot> plot_list;

	// Keep the user informed
	AfxGetApp()->GetMainWnd()->PostMessage(WM_PIPE_STATUS, PP_OPEN_PIPE);

	try{
		// Read data from the pipe as long as the pipe is valid
		// and the user hasn't caneled the pipe
		while(!haltPipe && local_status){
			if(!ReadFile(m_pipe, &msg, 1, &bytes_read, NULL))
				throw Error("Can't read message header");

			switch(msg){
				case PP_NEW_PLOT:
					if(plot_ptr){
						// Commit the current plot
						plot_list.Push(*plot_ptr);

						delete plot_ptr;
						curr_plot ++;
					}
					
					AfxGetApp()->GetMainWnd()->PostMessage(WM_PIPE_STATUS, 
						PP_NEW_PLOT, curr_plot);

					plot_ptr = new SPlot;

					// When the time comes to copy this plot, move the
					// data directly to the copy. This will (a) make copying
					// from the temporary plot more efficient and (b) maintain
					// the proper data order (as transmitted by PrestoLink).
					plot_ptr->MoveDataOnCopy(TRUE);

					plot_ptr->canXLog = TRUE;
					plot_ptr->canYLog = TRUE;

					// Should we read x error bars?
					if(!ReadFile(m_pipe, &use_x_error_bar, 1, &bytes_read, NULL))
						throw Error("Empty pipe!");
					
					// By default, show error bars above and below the point
					if(use_x_error_bar)
						plot_ptr->display_x_error_bar = SPlot::BAR_BOTH;

					// Should we read y error bars?
					if(!ReadFile(m_pipe, &use_y_error_bar, 1, &bytes_read, NULL))
						throw Error("Empty pipe!");

					// By default, show error bars above and below the point
					if(use_y_error_bar)
						plot_ptr->display_y_error_bar = SPlot::BAR_BOTH;

					// Read the info buffer size
					if(!ReadFile(m_pipe, &text_buffer_length, sizeof(int), &bytes_read, NULL))
						throw Error("Empty pipe!");

					if(text_buffer_length != 0){
						text_buffer = new char [text_buffer_length];

						// Read the info buffer
						if(!ReadFile(m_pipe, text_buffer, text_buffer_length, &bytes_read, NULL))
							throw Error("Empty pipe!");

						plot_ptr->info = CString(text_buffer);

						delete [] text_buffer;
					}
					else{
						plot_ptr->info = _T("");
					}

					text_buffer = NULL;
					
					// By default, increment the plot color
					if(plot_list.Length() > 0){
						SPlot *last_plot = &(plot_list[plot_list.Length() - 1]);

						plot_ptr->lineColor = 
							IndexToColor((ColorToIndex(last_plot->lineColor) + 1) % 
							S_CLR_NONE);

						// Exclude white from the list of colors!
						if(ColorToIndex(plot_ptr->lineColor) == S_CLR_WHITE)
							plot_ptr->lineColor = IndexToColor(S_CLR_WHITE + 1);
					}
					break;
				case PP_PRESTO_VERSION:
					// Read the version number
					if(!ReadFile(m_pipe, &tmp_int, sizeof(int), &bytes_read, NULL))
						throw Error("Empty pipe!");

					if(tmp_int != PRESTO_LINK_VERSION)
						throw Error("Wrong Presto Link version!");
					
					break;
				case PP_READ_PLOT:
					if(plot_ptr == NULL)
						throw Error("Can't read data into NULL plot");
					
					break;
				case PP_READ_POINT:
					if(plot_ptr == NULL)
						throw Error("Can't read point data into NULL plot");
					
					// Read the x data
					if(!ReadFile(m_pipe, &(tmp_dp.x), sizeof(float), &bytes_read, NULL))
						throw Error("Empty pipe!");

					if(tmp_dp.x <= 0.0f)
						plot_ptr->canXLog = FALSE;

					// Read the y data
					if(!ReadFile(m_pipe, &(tmp_dp.y), sizeof(float), &bytes_read, NULL))
						throw Error("Empty pipe!");
					
					if(tmp_dp.y <= 0.0f)
						plot_ptr->canYLog = FALSE;

					plot_ptr->data.Push(tmp_dp);

					if(use_x_error_bar){
						if(!ReadFile(m_pipe, &(x_error), sizeof(float), &bytes_read, NULL))
							throw Error("Empty pipe!");
						
						// Error bars MUST be positive values!
						plot_ptr->xErrorBar.Push((float)fabs(x_error));

						if(tmp_dp.x - x_error <= 0.0f)
							plot_ptr->canXLog = FALSE;
					}
					
					if(use_y_error_bar){
						if(!ReadFile(m_pipe, &(y_error), sizeof(float), &bytes_read, NULL))
							throw Error("Empty pipe!");

						// Error bars MUST be positive values!
						plot_ptr->yErrorBar.Push((float)fabs(y_error));

						if(tmp_dp.y - y_error <= 0.0f)
							plot_ptr->canYLog = FALSE;
					}

					break;
				case PP_CLOSE_PIPE:
					local_status = FALSE;

					if(plot_ptr){
						// Commit the current plot
						plot_list.Push(*plot_ptr);

						delete plot_ptr;
						plot_ptr = NULL;

						curr_plot ++;
					}
					
					AfxGetApp()->GetMainWnd()->PostMessage(WM_PIPE_STATUS, 
						PP_CLOSE_PIPE);

					break;
				case PP_EXIT:
					local_status = FALSE;

					if(plot_ptr){
						// Commit the current plot
						plot_list.Push(*plot_ptr);

						delete plot_ptr;
						plot_ptr = NULL;

						curr_plot ++;
					}
					
					// Shutdown!
					send_exit_msg = TRUE;
					break;

				case PP_GRAPH_TITLE:
					// Read the buffer size
					if(!ReadFile(m_pipe, &text_buffer_length, sizeof(int), &bytes_read, NULL))
						throw Error("Empty pipe!");

					text_buffer = new char [text_buffer_length];

					// Read the buffer
					if(!ReadFile(m_pipe, text_buffer, text_buffer_length, &bytes_read, NULL))
						throw Error("Empty pipe!");

					graph[curr_graph].title.SetText(text_buffer);

					delete [] text_buffer;
					text_buffer = NULL;
					break;
				case PP_X_AXIS_TITLE:
					// Read the buffer size
					if(!ReadFile(m_pipe, &text_buffer_length, sizeof(int), &bytes_read, NULL))
						throw Error("Empty pipe!");

					text_buffer = new char [text_buffer_length];

					// Read the buffer
					if(!ReadFile(m_pipe, text_buffer, text_buffer_length, &bytes_read, NULL))
						throw Error("Empty pipe!");

					graph[curr_graph].bottomAxisLable.SetText(text_buffer);

					delete [] text_buffer;
					text_buffer = NULL;
					break;
				case PP_Y_AXIS_TITLE:
					// Read the buffer size
					if(!ReadFile(m_pipe, &text_buffer_length, sizeof(int), &bytes_read, NULL))
						throw Error("Empty pipe!");

					text_buffer = new char [text_buffer_length];

					// Read the buffer
					if(!ReadFile(m_pipe, text_buffer, text_buffer_length, &bytes_read, NULL))
						throw Error("Empty pipe!");

					graph[curr_graph].leftAxisLable.SetText(text_buffer);

					delete [] text_buffer;
					text_buffer = NULL;
					break;
				case PP_SYMBOL_TYPE:
					if(plot_ptr == NULL)
						throw Error("Can't read symbol type into NULL plot");
					
					if(!ReadFile(m_pipe, &tmp_int, sizeof(int), &bytes_read, NULL))
						throw Error("Empty pipe!");

					if((tmp_int < SPlot::POINT_CIRCLE) || (tmp_int > SPlot::POINT_NONE))
						throw Error("Illegal symbol type");
					
					plot_ptr->pointType = tmp_int;
					break;
				case PP_SYMBOL_COLOR:
					if(plot_ptr == NULL)
						throw Error("Can't read symbol color into NULL plot");
					
					if(!ReadFile(m_pipe, &tmp_int, sizeof(int), &bytes_read, NULL))
						throw Error("Empty pipe!");

					if((tmp_int < S_CLR_BLACK) || (tmp_int > S_CLR_NONE))
						throw Error("Illegal symbol color");
					
					plot_ptr->brush.lbColor = IndexToColor(tmp_int);

					// Set the symbol brush
					if(plot_ptr->brush.lbColor == NO_COLOR)
						plot_ptr->brush.lbStyle = BS_HOLLOW;
					else
						plot_ptr->brush.lbStyle = BS_SOLID;

					break;
				case PP_SYMBOL_SIZE:
					if(plot_ptr == NULL)
						throw Error("Can't read symbol size into NULL plot");
					
					if(!ReadFile(m_pipe, &tmp_int, sizeof(int), &bytes_read, NULL))
						throw Error("Empty pipe!");

					if((tmp_int < 1) || (tmp_int > 10))
						throw Error("Illegal symbol size");
					
					plot_ptr->pointSize = tmp_int;

					break;
				case PP_SYMBOL_CUSTOM_COLOR:
					// Read a set of custom symbol colors [R, G, B] into the current plot
					if(plot_ptr == NULL)
						throw Error("Can't read custom symbol color into NULL plot");

					if(!ReadFile(m_pipe, &r, 1, &bytes_read, NULL))
						throw Error("Empty pipe!");

					if(!ReadFile(m_pipe, &g, 1, &bytes_read, NULL))
						throw Error("Empty pipe!");

					if(!ReadFile(m_pipe, &b, 1, &bytes_read, NULL))
						throw Error("Empty pipe!");
					
					// Check the bounds of each color index
					if((r < 0) || (r > 255) ||
					   (g < 0) || (g > 255) ||
					   (b < 0) || (b > 255))
					throw Error("Bad custom symbol color");

					// Convert the RGB values into a COLORREF 
					// structure (i.e. 32 bit int).
					plot_ptr->brush.lbColor = (b & 0x000000ff) << 8;
					plot_ptr->brush.lbColor = (plot_ptr->brush.lbColor | (g & 0x000000ff)) << 8;
					plot_ptr->brush.lbColor = plot_ptr->brush.lbColor | (r & 0x000000ff);

					// Set the symbol brush
					if(plot_ptr->brush.lbColor == NO_COLOR)
						plot_ptr->brush.lbStyle = BS_HOLLOW;
					else
						plot_ptr->brush.lbStyle = BS_SOLID;

					break;
				case PP_LINE_TYPE:
					if(plot_ptr == NULL)
						throw Error("Can't read line type into NULL plot");
					
					if(!ReadFile(m_pipe, &tmp_int, sizeof(int), &bytes_read, NULL))
						throw Error("Empty pipe!");

					if((tmp_int < SPlot::LINE_SOLID) || (tmp_int > SPlot::LINE_NONE))
						throw Error("Illegal line type");
					
					plot_ptr->lineType = tmp_int;
					break;
				case PP_LINE_COLOR:
					if(plot_ptr == NULL)
						throw Error("Can't read line color into NULL plot");
					
					if(!ReadFile(m_pipe, &tmp_int, sizeof(int), &bytes_read, NULL))
						throw Error("Empty pipe!");

					if((tmp_int < S_CLR_BLACK) || (tmp_int > S_CLR_NONE))
						throw Error("Illegal line color");
					
					plot_ptr->lineColor = IndexToColor(tmp_int);
					break;
				case PP_LINE_SIZE:
					if(plot_ptr == NULL)
						throw Error("Can't read line size into NULL plot");
					
					if(!ReadFile(m_pipe, &tmp_int, sizeof(int), &bytes_read, NULL))
						throw Error("Empty pipe!");

					if((tmp_int < 1) || (tmp_int > 5))
						throw Error("Illegal line size");
					
					plot_ptr->lineSize = tmp_int;

					break;
				case PP_LINE_CUSTOM_COLOR:
					// Read a set of custom line colors [R, G, B] into the current plot
					if(plot_ptr == NULL)
						throw Error("Can't read custom line color into NULL plot");

					if(!ReadFile(m_pipe, &r, 1, &bytes_read, NULL))
						throw Error("Empty pipe!");

					if(!ReadFile(m_pipe, &g, 1, &bytes_read, NULL))
						throw Error("Empty pipe!");

					if(!ReadFile(m_pipe, &b, 1, &bytes_read, NULL))
						throw Error("Empty pipe!");
					
					// Check the bounds of each color index
					if((r < 0) || (r > 255) ||
					   (g < 0) || (g > 255) ||
					   (b < 0) || (b > 255))
					throw Error("Bad custom symbol color");

					// Convert the RGB values into a COLORREF 
					// structure (i.e. 32 bit int).
					plot_ptr->lineColor = (b & 0x000000ff) << 8;
					plot_ptr->lineColor = (plot_ptr->lineColor | (g & 0x000000ff)) << 8;
					plot_ptr->lineColor = plot_ptr->lineColor | (r & 0x000000ff);
					break;
				case PP_GRAPH_TITLE_SIZE:
					// Set the font size of the graph title
					if(!ReadFile(m_pipe, &tmp_int, sizeof(int), &bytes_read, NULL))
						throw Error("Empty pipe!");

					if(tmp_int <= 0)
						throw Error("Font size is too small!");

					graph[curr_graph].title.SetFontSize(tmp_int);
					break;
				case PP_X_LABEL_SIZE:
					// Set the font size of the x axis label
					if(!ReadFile(m_pipe, &tmp_int, sizeof(int), &bytes_read, NULL))
						throw Error("Empty pipe!");

					if(tmp_int <= 0)
						throw Error("Font size is too small!");

					graph[curr_graph].bottomAxisLable.SetFontSize(tmp_int);
					break;
				case PP_Y_LABEL_SIZE:
					// Set the font size of the y axis label
					if(!ReadFile(m_pipe, &tmp_int, sizeof(int), &bytes_read, NULL))
						throw Error("Empty pipe!");

					if(tmp_int <= 0)
						throw Error("Font size is too small!");

					graph[curr_graph].leftAxisLable.SetFontSize(tmp_int);
					break;
				case PP_SAVE_PRESTO_EXIT:	// Fall through
				case PP_EXPORT_BMP_EXIT:	// Fall through
				case PP_EXPORT_PNG_EXIT:	// Fall through

					// The msg encodes the export/save format
					save_file_at_exit = msg;

					local_status = FALSE;

					// Allow the current document to be saved. Note that this operation
					// also commits the current plot (the user must call PrestoNewPlot to
					// create a new plot and the old plot will no longer be accessible!).
					if(plot_ptr){
						// Commit the current plot
						plot_list.Push(*plot_ptr);

						delete plot_ptr;
						plot_ptr = NULL;

						curr_plot ++;
					}

					// Read the filename buffer size
					if(!ReadFile(m_pipe, &text_buffer_length, sizeof(int), &bytes_read, NULL))
						throw Error("Empty pipe!");

					text_buffer = new char [text_buffer_length];

					// Read the filename buffer
					if(!ReadFile(m_pipe, text_buffer, text_buffer_length, &bytes_read, NULL))
						throw Error("Empty pipe!");
					
					// Save a copy of the filename
					filename = text_buffer;

					if(save_file_at_exit != PP_SAVE_PRESTO_EXIT){
						// Save the output dimensions of the exported image
						if(!ReadFile(m_pipe, &atExitExportWidth, sizeof(int), &bytes_read, NULL))
							throw Error("Empty pipe!");

						if(!ReadFile(m_pipe, &atExitExportHeight, sizeof(int), &bytes_read, NULL))
							throw Error("Empty pipe!");
					}

					send_exit_msg = TRUE;
					break;
				case PP_SCALE_X:
					// Read the scaling limits to apply to the graph
					if(!ReadFile(m_pipe, &min_x, sizeof(float), &bytes_read, NULL))
						throw Error("Empty pipe!");

					if(!ReadFile(m_pipe, &max_x, sizeof(float), &bytes_read, NULL))
						throw Error("Empty pipe!");
					
					// Set the rescale flag
					set_scale_x = TRUE;
					break;
				case PP_SCALE_Y:
					// Read the scaling limits to apply to the graph
					if(!ReadFile(m_pipe, &min_y, sizeof(float), &bytes_read, NULL))
						throw Error("Empty pipe!");

					if(!ReadFile(m_pipe, &max_y, sizeof(float), &bytes_read, NULL))
						throw Error("Empty pipe!");
					
					// Set the rescale flag
					set_scale_y = TRUE;

					break;
				case PP_TICK_X:
					// Read the major and minor tick values to apply to the graph
					if(!ReadFile(m_pipe, &major_tick_x, sizeof(float), &bytes_read, NULL))
						throw Error("Empty pipe!");

					if(!ReadFile(m_pipe, &minor_tick_x, sizeof(float), &bytes_read, NULL))
						throw Error("Empty pipe!");
					
					// Set the tick spacing flag
					set_tick_x = TRUE;
					break;
				case PP_TICK_Y:
					// Read the major and minor tick values to apply to the graph
					if(!ReadFile(m_pipe, &major_tick_y, sizeof(float), &bytes_read, NULL))
						throw Error("Empty pipe!");

					if(!ReadFile(m_pipe, &minor_tick_y, sizeof(float), &bytes_read, NULL))
						throw Error("Empty pipe!");
					
					// Set the tick spacing flag
					set_tick_y = TRUE;
					break;
				default:
					throw Error("Unknown message");
			};
		}
	}
	catch(Error &e){
		EndWaitCursor();
		AfxMessageBox(e.message());
	};

	// Clean up
	if(plot_ptr)
		delete plot_ptr;

	// Transfer the data from the temporary plot list to the plot list
	// associated with the current graph
	graph[curr_graph].plot.Move(plot_list);

	// Make sure that we're not going to attempt to
	// compute the log of a number that is <= 0.
	if((graph[curr_graph].x_scaleType == SCALE_LOG) || 
	   (graph[curr_graph].y_scaleType == SCALE_LOG)){
						
		BOOL x_log_ok = TRUE, y_log_ok = TRUE;

		for(int i = 0;i < graph[curr_graph].plot.Length();i++){

			if(!graph[curr_graph].plot[i].canXLog)
				x_log_ok = FALSE;

			if(!graph[curr_graph].plot[i].canYLog)
				y_log_ok = FALSE;
	}

	// Tell the user what has happened.
	if((!x_log_ok && (graph[curr_graph].x_scaleType == SCALE_LOG)) || 
	   (!y_log_ok && (graph[curr_graph].y_scaleType == SCALE_LOG)))
		AfxMessageBox("Can't compute log(n<=0) for piped data; "
			"Switching to a linear scale.");

		// Change the scale type(s) to allow all of the
		// data to be displayed
		if(!x_log_ok)
			graph[curr_graph].x_scaleType = SCALE_LINEAR;
					
		if(!y_log_ok)
			graph[curr_graph].y_scaleType = SCALE_LINEAR;
	}

	// Initialize the plot bounding box
	graph[curr_graph].SetBoundingBox();
	graph[curr_graph].SetPlotRange();

	// Do we need to rescale the graph range?
	if(set_scale_x){
		graph[curr_graph].SetPlotRangeX(min_x, max_x);
	}

	if(set_scale_y){
		graph[curr_graph].SetPlotRangeY(min_y, max_y);
	}

	// Do we need to adjust axis tick mark spacing?
	if(set_tick_x){
		// Make sure all tick spacing values are positive!
		graph[curr_graph].topAxis.MajorTickSpace((float)(fabs(major_tick_x)), TRUE);
		graph[curr_graph].topAxis.MinorTickSpace((float)(fabs(minor_tick_x)), TRUE);
		graph[curr_graph].bottomAxis.MajorTickSpace((float)(fabs(major_tick_x)), TRUE);
		graph[curr_graph].bottomAxis.MinorTickSpace((float)(fabs(minor_tick_x)), TRUE);
	}

	if(set_tick_y){
		graph[curr_graph].leftAxis.MajorTickSpace((float)(fabs(major_tick_y)), TRUE);
		graph[curr_graph].leftAxis.MinorTickSpace((float)(fabs(minor_tick_y)), TRUE);
		graph[curr_graph].rightAxis.MajorTickSpace((float)(fabs(major_tick_y)), TRUE);
		graph[curr_graph].rightAxis.MinorTickSpace((float)(fabs(minor_tick_y)), TRUE);
	}

	SetModifiedFlag();

	// Don't update the view by calling UpdateAllViews(NULL) -- it's
	// not thread safe. Send a message instead.
	AfxGetApp()->GetMainWnd()->PostMessage(WM_STOP_PIPE);

	m_pipe = NULL;

	if(save_file_at_exit != -1){
		try{
			switch(save_file_at_exit){
				case PP_SAVE_PRESTO_EXIT:
					// Save the current document
					OnSaveDocument(filename);
					break;
				case PP_EXPORT_BMP_EXIT:
					atExitFilename = filename;
					atExitTask = AT_EXIT_EXPORT_BMP;
					break;
				case PP_EXPORT_PNG_EXIT:
					atExitFilename = filename;
					atExitTask = AT_EXIT_EXPORT_PNG;
					break;
				default:
					throw Error("Unknown export/save option");
			};
		}
		catch(...){
			throw Error("Unable to save/export file!");
		}
	}

	if(send_exit_msg){
		// Mark the document as unmodified to prevent the user
		// from being prompted to save the document
		SetModifiedFlag(FALSE);
		AfxGetApp()->GetMainWnd()->PostMessage(WM_CLOSE);
	}
}
