//	(c) 2000-2002 J. Gans (jdg9@cornell.edu), Shalloway Lab, 
//	Cornell University. You can modify and freely distribute 
//	this code, but please keep this header intact.

#include "stdafx.h"
#include "sgraph.h"
#include "PolarEquationDialog.h"
#include "sgraphDoc.h"
#include "Equation/Equation.h"
#include "PrestoUtil.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

// The independant variable
#define		INDEPENDANT_VAR		"t" // Theta

/////////////////////////////////////////////////////////////////////////////
// CPolarEquationDialog dialog


CPolarEquationDialog::CPolarEquationDialog(CDocument* pDoc,
										   list<PolarEquationInfo> &m_polarEqHistory)
	: CDialog(CPolarEquationDialog::IDD, NULL)
{
	m_pDoc = pDoc;

	ptrPolarEqHistory = &m_polarEqHistory;

	//{{AFX_DATA_INIT(CPolarEquationDialog)
	m_equation = _T("");
	m_num_points = 1;
	m_t_max = _T("");
	m_t_min = _T("");
	//}}AFX_DATA_INIT
}


void CPolarEquationDialog::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CPolarEquationDialog)
	DDX_Control(pDX, IDC_EQUATION_TAB, m_eq_tab);
	DDX_Control(pDX, IDC_EVAL_PROGRESS, m_status);
	DDX_Text(pDX, IDC_EQUATION_EDIT, m_equation);
	DDX_Text(pDX, IDC_NUM_POINTS, m_num_points);
	DDV_MinMaxUInt(pDX, m_num_points, 1, 4294967295);
	DDX_Text(pDX, IDC_T_MAX, m_t_max);
	DDX_Text(pDX, IDC_T_MIN, m_t_min);
	//}}AFX_DATA_MAP
}


BEGIN_MESSAGE_MAP(CPolarEquationDialog, CDialog)
	//{{AFX_MSG_MAP(CPolarEquationDialog)
	ON_NOTIFY(TCN_SELCHANGE, IDC_EQUATION_TAB, OnSelchangeEquationTab)
	ON_NOTIFY(TCN_SELCHANGING, IDC_EQUATION_TAB, OnSelchangingEquationTab)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CPolarEquationDialog message handlers
BOOL CPolarEquationDialog::OnInitDialog()
{	
	CDialog::OnInitDialog();

	// Set the default number of points
	m_num_points = 100;

	// Default range of t values 
	float minT = 0.0f;	// Default theta min value
	float maxT = 6.283185307f;	// Default theta max value
	
	m_t_max.Format("%g", maxT);
	m_t_min.Format("%g", minT);
	
	// Save the default settings in the history list
	SaveEquationHistory();

	// Set the history tabs to reflect the number of equations in 
	// the history list
	for(int i = 0;i < ptrPolarEqHistory->Length();i++){
		char tmpString[64];

		// Start numbering equations from 1
		sprintf(tmpString, "Eq %d", i+1);

		m_eq_tab.InsertItem(i, tmpString);
	}

	// Select the default tab
	m_eq_tab.SetCurSel(0);

	// Update all the values we just set
	UpdateData(FALSE);

	// Give the text edit control the window focus
	GotoDlgCtrl(GetDlgItem(IDC_EQUATION_EDIT));

	return FALSE; // Return false because we are manually setting the focus
}

void CPolarEquationDialog::OnOK() 
{
	// Cast the document pointer
	SGraph *m_ptr = &(((CSgraphDoc *)m_pDoc)->
		graph[((CSgraphDoc *)m_pDoc)->selected_graph]);

	float minT;
	float maxT;

	CString error_buffer;

	// Read in all of the variables from the dialog box
	UpdateData();

	// Must plot at least a single data point
	if(m_num_points < 1)
		return;

	if(StringToFloat(m_t_max, maxT) == FALSE){
		AfxMessageBox("Invalid max t value!");
		return;
	}

	if(StringToFloat(m_t_min, minT) == FALSE){
		AfxMessageBox("Invalid min t value!");
		return;
	}

	if(minT >= maxT){
		AfxMessageBox("t min is greater than (or equal) to t max.");
		return;
	}
	
	float dt;
	
	if(m_num_points == 1){
		dt = maxT - minT;
	}
	else{
		dt = (maxT - minT)/(m_num_points - 1);
	}

	if(dt <= 0.0){
		AfxMessageBox("Invalid t step size.");
		return;
	}

	// Set the cursor to the hourglass
	HCURSOR prevCursor = ::SetCursor(LoadCursor(NULL, IDC_WAIT));

	// Prepare to evaluate the equation
	try{
		// Make a copy of the input string WITHOUT any \n or \r characters
		char *buffer = new char[m_equation.GetLength() + 1];
		int i, j = 0;
		for(i = 0;i < m_equation.GetLength();i++){
			char c = m_equation[i];

			if((c != '\n') && (c != '\r'))
				buffer[j++] = c;
		}

		// Terminate the string
		buffer[j] = '\0';

		CString eq_string(buffer);

		delete [] buffer;

		SEquation eq((LPCTSTR)(eq_string));

		if(eq.GetError())
			throw eq.GetError();

		if(!eq.DefineIndepVar(INDEPENDANT_VAR)){
			// The user has (a) not specifed the correct independant variable
			// or (b) not specified ANY independant variable. Test for the
			// latter before throwing an error
			if(eq.NumVar() != 0)
				throw "Please specify a function of "INDEPENDANT_VAR".";
		}

		// Make sure that the user has not specified any other
		// variables (only plotting functions of a single variable -- for
		// now).
		if( eq.NumVar() != eq.NumIndepVar()){
			
			error_buffer.Format("The symbol '%s' is undefined", 
				eq.GetVarName(i));
			
			throw (LPCTSTR)(error_buffer);
		}

		// Simplify the equation for efficient evaluation
		eq.Simplify();

		// Did we run into trouble simplifying the equation?
		if(eq.GetError())
			throw eq.GetError();

		// Prepare to evaluate the equation
		SPlot newPlot;
		DataPoint xy;
		BOOL first_time = TRUE;
		float minX;
		float minY;

		// Prepare the status bar
		int eval_count = 0;
		int update_every;

		if(m_num_points > 100){
			double tmp = log10(m_num_points);
			update_every = (int)(pow(10.0, tmp - 2.0));
		}
		else{
			update_every = 10;
		}

		m_status.SetRange32(0, m_num_points);
		m_status.SetStep(update_every);

		// To satisfy the revesed plot storage, plot
		// the function from maxT to minT
		//for(float t = maxT;t >= minT;t -= dt){
		for(i = 0;i < (int)m_num_points;i++){
			float t = maxT - i*dt;

			// Convert from polar to cartesian coordiantes
			float r = eq(t);
			xy.x = (float)(cos(t))*r;
			xy.y = (float)(sin(t))*r;

			// Did we successfully evaluate the function?
			if(eq.GetError()){
				// Nope. A error was encountered
				eq.ClearError();
			}
			else{
				// Yes! Save the function value in the new plot
				newPlot.data.Push(xy);
				
				// Track the smallest value to see if we can plot the
				// equation on a log plot.
				if(first_time){
					first_time = FALSE;

					minX = xy.x;
					minY = xy.y;
				}
				else{
					if(xy.y < minY)
						minY = xy.y;

					if(xy.x < minX)
						minX = xy.x;
				}
			}

			if((++eval_count) % update_every == 0)
				m_status.StepIt();
		}

		// If we never evaluated the function, don't add the plot!
		if(first_time == TRUE)
			throw "Unable to evalute the function within the given range of "INDEPENDANT_VAR;

		// Give the plot a name (use the equation)
		newPlot.info = eq_string;

		// Set the plot style
		newPlot.pointType = SPlot::POINT_NONE;
		newPlot.lineColor = 0x00000000; // Black
		newPlot.lineType = SPlot::LINE_SOLID;
	
		// Set the scale type from the current graph
		newPlot.x_scaleType = m_ptr->x_scaleType;
		newPlot.y_scaleType = m_ptr->y_scaleType;

		if(minX <= 0.0)
			newPlot.canXLog = FALSE;
		else
			newPlot.canXLog = TRUE;

		if(minY <= 0.0)
			newPlot.canYLog = FALSE;
		else
			newPlot.canYLog = TRUE;

		// Add the plot of the function to the current graph
		m_ptr->plot.Push(newPlot);

		// Update the graph scaling
		m_ptr->SetPlotScaleType();
		m_ptr->SetPlotRange();
		m_ptr->SetBoundingBox();

	}
	catch(const char *error){
		// Restore the cursor
		::SetCursor(prevCursor);

		// Invalid equation!
		AfxMessageBox(error);
		return;
	}
	catch(...){
		// Restore the cursor
		::SetCursor(prevCursor);

		// Something really, really bad happened
		AfxMessageBox("Ooof! Unable to evaluate equation!");
		return;
	}

	// Remove the default equation info from the head of the history list
	// (there is no need to store it)
	ptrPolarEqHistory->Remove(0);

	// Save this equation information for the equation history list
	SaveEquationHistory();

	// Restore the cursor
	::SetCursor(prevCursor);

	CDialog::OnOK();
}

void CPolarEquationDialog::SaveEquationHistory()
{
	PolarEquationInfo tmp;

	tmp.m_equation = m_equation;
	tmp.m_num_points = m_num_points;
	tmp.m_t_max = m_t_max;
	tmp.m_t_min = m_t_min;

	// Make sure the current info doesn't match an existing entry
	for(int i = 0;i < ptrPolarEqHistory->Length();i++){
		if(tmp == (*ptrPolarEqHistory)[i])
			return;
	}

	ptrPolarEqHistory->Insert(tmp, 0);
}

void CPolarEquationDialog::OnSelchangeEquationTab(NMHDR* pNMHDR, LRESULT* pResult) 
{
	// The user has selected an equation from the history list, update
	// the dialog
	m_equation = (*ptrPolarEqHistory)[m_eq_tab.GetCurSel()].m_equation;
	m_num_points = (*ptrPolarEqHistory)[m_eq_tab.GetCurSel()].m_num_points;
	m_t_max = (*ptrPolarEqHistory)[m_eq_tab.GetCurSel()].m_t_max;
	m_t_min = (*ptrPolarEqHistory)[m_eq_tab.GetCurSel()].m_t_min;

	// Update all the values we just set
	UpdateData(FALSE);
	
	*pResult = 0;
}

void CPolarEquationDialog::OnSelchangingEquationTab(NMHDR* pNMHDR, LRESULT* pResult) 
{
	// Read the values set by the user
	UpdateData(TRUE);

	// The user has selected an equation from the history list, update
	// the dialog before switching to the new equation. This keeps the history
	// list current and allows the user to switch back and forth without losing
	// any values.
	(*ptrPolarEqHistory)[m_eq_tab.GetCurSel()].m_equation = m_equation;
	(*ptrPolarEqHistory)[m_eq_tab.GetCurSel()].m_num_points = m_num_points;
	(*ptrPolarEqHistory)[m_eq_tab.GetCurSel()].m_t_max = m_t_max;
	(*ptrPolarEqHistory)[m_eq_tab.GetCurSel()].m_t_min = m_t_min;
	
	*pResult = 0;
}
