/*	
 * (c) 2000 J. D. Gans (jdg9@cornell.edu), Shalloway Lab, 
 * Cornell University. You can modify and freely distribute 
 * this code, but please keep this header intact.
 */


/*
 * Note that this file has a CPP extension, but should compile under C as
 * well. Note that use of AfxMessageBox for the error reporting requires
 * MFC and C++.
 *
 * Include the follwing header if using the Microsoft foundation classes
 * (otherwise, comment it out).
 */
#ifdef _WINDOWS

/* #include <stdafx.h> */
#include <afxwin.h>
#define		ERROR_MSG(X)	AfxMessageBox(X);

#else /* defined (_CONSOLE) */

#define		ERROR_MSG(X)	fprintf(stderr, X "\n");

#endif /* _AFXDLL */

#include <stdio.h>
#include <stdlib.h>
#include <io.h>
#include <fcntl.h>
#include <process.h>
#include <errno.h>
#include <string.h>
#include <float.h>
#include "PrestoLink.h"

enum {PIPE_READ = 0, PIPE_WRITE};

/* Local function */
int SendString(HANDLE pipe, char msg, const char *buff);
int SendInt(HANDLE pipe,char msg, int value);

int PrestoOpen(PrestoLink* lpPrestoLink, const char *title, const char *x_axis, const char *y_axis)
{
	char pipeName[32];
	SECURITY_ATTRIBUTES secure_attrib;

	if(!lpPrestoLink){
		ERROR_MSG("NULL handle passed to PrestoOpen!");
		return FALSE;
	}

	/*
	 * Set the security attributes to insure that we can inherit this pipe
	 */
	secure_attrib.bInheritHandle = TRUE;
	secure_attrib.lpSecurityDescriptor = NULL;
	secure_attrib.nLength = sizeof(SECURITY_ATTRIBUTES);

	if(!CreatePipe(&(lpPrestoLink->pipeHandle[PIPE_READ]), 
				   &(lpPrestoLink->pipeHandle[PIPE_WRITE]), 
				   &secure_attrib, 0)){

		ERROR_MSG("Unable to create pipe!");
		lpPrestoLink->valid = FALSE;
		return FALSE;
	}

	/* Convert the integer pipe hadle to a string */
	sprintf(pipeName, "%d", lpPrestoLink->pipeHandle[PIPE_READ]);

	/* Spawn a copy of PrestoPlot */
	if(_spawnlp(_P_NOWAIT, PRESTO_PLOT_PATH, 
		PRESTO_PLOT_PATH, "-pipe", pipeName , NULL) == -1){

		CloseHandle(lpPrestoLink->pipeHandle[PIPE_READ]);
		CloseHandle(lpPrestoLink->pipeHandle[PIPE_WRITE]);

		switch(errno){
			case E2BIG:
				ERROR_MSG("Malformed argument to spawn.");
				break;
			case EINVAL:
				ERROR_MSG("Invalid spawn mode.");
				break;
			case ENOENT:
				ERROR_MSG("Please check executable path.");
				break;
			case ENOEXEC:
				ERROR_MSG("Bad executable format.");
				break;
			case ENOMEM:
				ERROR_MSG("Insufficient memory to spawn viewer.");
				break;
		};

		lpPrestoLink->valid = FALSE;

		return FALSE;
	}
	else{
		/* 
		 * Send the version number (to make sure we are connecting to the correct
		 * version of PrestoPlot).
		 */
		if(!SendInt(lpPrestoLink->pipeHandle[PIPE_WRITE],
			PP_PRESTO_VERSION, PRESTO_LINK_VERSION)){
			lpPrestoLink->valid = FALSE;
			return FALSE;
		}

		/* The title (if specified) */
		if(title){
			if(!SendString(lpPrestoLink->pipeHandle[PIPE_WRITE], 
				PP_GRAPH_TITLE, title)){
				lpPrestoLink->valid = FALSE;
				return FALSE;
			}
		}

		/* The X Axis (if specified) */
		if(title){
			if(!SendString(lpPrestoLink->pipeHandle[PIPE_WRITE],
				PP_X_AXIS_TITLE, x_axis)){
				lpPrestoLink->valid = FALSE;
				return FALSE;
			}
		}

		/* The Y Axis (if specified) */
		if(title){
			if(!SendString(lpPrestoLink->pipeHandle[PIPE_WRITE],
				PP_Y_AXIS_TITLE, y_axis)){
				lpPrestoLink->valid = FALSE;
				return FALSE;
			}
		}
	}

	/* Reset the newPlot flag */
	lpPrestoLink->newPlot = FALSE;
	lpPrestoLink->valid = TRUE;

	return TRUE;
};

int PrestoClose(PrestoLink* lpPrestoLink)
{
	DWORD bytes_written;
	char msg = PP_CLOSE_PIPE;

	if(lpPrestoLink && (lpPrestoLink->valid == TRUE)){
		if(!WriteFile(lpPrestoLink->pipeHandle[PIPE_WRITE], &msg, 1, &bytes_written, NULL)){
			ERROR_MSG("Can't send close msg");
			lpPrestoLink->valid = FALSE;
			return FALSE;
		}

		CloseHandle(lpPrestoLink->pipeHandle[PIPE_READ]);
		CloseHandle(lpPrestoLink->pipeHandle[PIPE_WRITE]);

		/* After we close the pipes, this handle is no longer valid */
		lpPrestoLink->valid = FALSE;
	}
	else{
		ERROR_MSG("Invalid PrestoLink passed to PrestoClose");
		return FALSE;
	}

	return TRUE;
};

/* 
 * Quit the currently running copy of PrestoPlot.
 */
int PrestoExit(PrestoLink* lpPrestoLink)
{
	char tmp_char;

	DWORD bytes_written;

	if(!lpPrestoLink || (lpPrestoLink->valid != TRUE)){
		ERROR_MSG("Invalid PrestoLink passed to PrestoExit");
		return FALSE;
	}

	tmp_char = PP_EXIT;

	if(!WriteFile(lpPrestoLink->pipeHandle[PIPE_WRITE], &tmp_char, 1, &bytes_written, NULL)){
		ERROR_MSG("Can't send exit msg");
		
		lpPrestoLink->valid = FALSE;
		return FALSE;
	}
	
	CloseHandle(lpPrestoLink->pipeHandle[PIPE_READ]);
	CloseHandle(lpPrestoLink->pipeHandle[PIPE_WRITE]);

	/* After we close the pipes, this handle is no longer valid */
	lpPrestoLink->valid = FALSE;

	return TRUE;
};

/* 
 * Quit the currently running copy of PrestoPlot and
 * save the session to disk in the native PrestoPlot file format.
 */
int PrestoSaveandExit(PrestoLink* lpPrestoLink, const char *filename)
{
	char tmp_char;
	int tmp_int;

	DWORD bytes_written;

	if(!lpPrestoLink || (lpPrestoLink->valid != TRUE)){
		ERROR_MSG("Invalid PrestoLink passed to PrestoSaveandExit");
		return FALSE;
	}

	if(filename == NULL){
		ERROR_MSG("NULL filename passed to PrestoSaveandExit");
		return FALSE;
	}

	/* Computing the string length. */
	tmp_int = strlen(filename) + 1;

	if(tmp_int == 1){
		ERROR_MSG("Zero length filename passed to PrestoSaveandExit");
		return FALSE;
	}

	tmp_char = PP_SAVE_PRESTO_EXIT;

	if(!WriteFile(lpPrestoLink->pipeHandle[PIPE_WRITE], &tmp_char, 1, &bytes_written, NULL)){
		ERROR_MSG("Can't send save msg");
		
		lpPrestoLink->valid = FALSE;
		return FALSE;
	}
	
	/* Send the filename string length */
	if(!WriteFile(lpPrestoLink->pipeHandle[PIPE_WRITE], &tmp_int, sizeof(int), &bytes_written, NULL)){
		ERROR_MSG("Can't set filename string length");

		lpPrestoLink->valid = FALSE;
		return FALSE;
	}
	
	/* Write the filename buffer */
	if(!WriteFile(lpPrestoLink->pipeHandle[PIPE_WRITE], filename, tmp_int, &bytes_written, NULL)){
		ERROR_MSG("Can't set filename string");

		lpPrestoLink->valid = FALSE;
		return FALSE;
	}

	CloseHandle(lpPrestoLink->pipeHandle[PIPE_READ]);
	CloseHandle(lpPrestoLink->pipeHandle[PIPE_WRITE]);

	/* After we close the pipes, this handle is no longer valid */
	lpPrestoLink->valid = FALSE;

	return TRUE;
};

/* 
 * Quit the currently running copy of PrestoPlot and
 * export the graph as an image (currently either BMP or PNG).
 */
int PrestoExportandExit(PrestoLink* lpPrestoLink, const char *filename, 
						int width, int height, int export_format)
{
	char tmp_char;
	int tmp_int;

	DWORD bytes_written;

	if(!lpPrestoLink || (lpPrestoLink->valid != TRUE)){
		ERROR_MSG("Invalid PrestoLink passed to PrestoExportandExit");
		return FALSE;
	}

	/* Make sure that the width and height are valid */
	if(width < 1){
		ERROR_MSG("width < 1 in PrestoExportandExit");
		return FALSE;
	}

	if(width > 100000){
		ERROR_MSG("width sanity check failed in PrestoExportandExit");
		return FALSE;
	}

	if(height < 1){
		ERROR_MSG("height < 1 in PrestoExportandExit");
		return FALSE;
	}

	if(height > 100000){
		ERROR_MSG("height sanity check failed in PrestoExportandExit");
		return FALSE;
	}

	if(filename == NULL){
		ERROR_MSG("NULL filename passed to PrestoExportandExit");
		return FALSE;
	}

	/* Computing the string length. */
	tmp_int = strlen(filename) + 1;

	if(tmp_int == 1){
		ERROR_MSG("Zero length filename passed to PrestoExportandExit");
		return FALSE;
	}

	switch(export_format){
		case PRESTO_EXPORT_BMP:
			tmp_char = PP_EXPORT_BMP_EXIT;
			break;
		case PRESTO_EXPORT_PNG:
			tmp_char = PP_EXPORT_PNG_EXIT;
			break;
		default:
			ERROR_MSG("Unknown export format passed to PrestoExportandExit");
		return FALSE;
	};

	if(!WriteFile(lpPrestoLink->pipeHandle[PIPE_WRITE], &tmp_char, 1, &bytes_written, NULL)){
		ERROR_MSG("Can't send export msg");
		
		lpPrestoLink->valid = FALSE;
		return FALSE;
	}
	
	/* Send the filename string length */
	if(!WriteFile(lpPrestoLink->pipeHandle[PIPE_WRITE], &tmp_int, sizeof(int), &bytes_written, NULL)){
		ERROR_MSG("Can't set filename string length");

		lpPrestoLink->valid = FALSE;
		return FALSE;
	}
	
	/* Write the filename buffer */
	if(!WriteFile(lpPrestoLink->pipeHandle[PIPE_WRITE], filename, tmp_int, &bytes_written, NULL)){
		ERROR_MSG("Can't set filename string");

		lpPrestoLink->valid = FALSE;
		return FALSE;
	}
	
	/* Send the width */
	if(!WriteFile(lpPrestoLink->pipeHandle[PIPE_WRITE], &width, sizeof(int), &bytes_written, NULL)){
		ERROR_MSG("Can't set filename string length");

		lpPrestoLink->valid = FALSE;
		return FALSE;
	}

	/* Send the height */
	if(!WriteFile(lpPrestoLink->pipeHandle[PIPE_WRITE], &height, sizeof(int), &bytes_written, NULL)){
		ERROR_MSG("Can't set filename string length");

		lpPrestoLink->valid = FALSE;
		return FALSE;
	}

	CloseHandle(lpPrestoLink->pipeHandle[PIPE_READ]);
	CloseHandle(lpPrestoLink->pipeHandle[PIPE_WRITE]);

	/* After we close the pipes, this handle is no longer valid */
	lpPrestoLink->valid = FALSE;

	return TRUE;
};

int PrestoNewPlot(PrestoLink* lpPrestoLink, const char *info, int plotType)
{
	char tmp_char;
	int tmp_int;

	DWORD bytes_written;

	if(!lpPrestoLink || (lpPrestoLink->valid != TRUE)){
		ERROR_MSG("Invalid PrestoLink passed to PrestoNewPlot");
		return FALSE;
	}

	/* Save the users options */
	lpPrestoLink->plotType = plotType;

	tmp_char = PP_NEW_PLOT;

	if(!WriteFile(lpPrestoLink->pipeHandle[PIPE_WRITE], &tmp_char, 1, &bytes_written, NULL)){
		ERROR_MSG("Can't send new plot msg");
		
		lpPrestoLink->valid = FALSE;
		return FALSE;
	}
					
	/* Should we expect x error bars? */
	if(plotType & PRESTO_ERROR_BAR_X)
		tmp_char = 1;
	else
		tmp_char = 0;

	if(!WriteFile(lpPrestoLink->pipeHandle[PIPE_WRITE], &tmp_char, 1, &bytes_written, NULL)){
		ERROR_MSG("Can't set x error bars");

		lpPrestoLink->valid = FALSE;
		return FALSE;
	}

	/* Should we expect y error bars? */
	if(plotType & PRESTO_ERROR_BAR_Y)
		tmp_char = 1;
	else
		tmp_char = 0;

	if(!WriteFile(lpPrestoLink->pipeHandle[PIPE_WRITE], &tmp_char, 1, &bytes_written, NULL)){
		ERROR_MSG("Can't set y error bars");

		lpPrestoLink->valid = FALSE;
		return FALSE;
	}

	/* Set the info buffer size */
	if(info)
		tmp_int = strlen(info) + 1;
	else
		tmp_int = 0;
	
	if(!WriteFile(lpPrestoLink->pipeHandle[PIPE_WRITE], &tmp_int, sizeof(int), &bytes_written, NULL)){
		ERROR_MSG("Can't set info string length");

		lpPrestoLink->valid = FALSE;
		return FALSE;
	}
	
	if(info){
		/* Write the info buffer */
		if(!WriteFile(lpPrestoLink->pipeHandle[PIPE_WRITE], info, tmp_int, &bytes_written, NULL)){
			ERROR_MSG("Can't set info string");

			lpPrestoLink->valid = FALSE;
			return FALSE;
		}	
	}

	/* A new plot has been created */
	lpPrestoLink->newPlot = TRUE;

	return TRUE;
};

int PrestoWritePointXY(PrestoLink* lpPrestoLink, float x, float y)
{
	float buf[2];
	DWORD bytes_written;
	char msg = PP_READ_POINT;

	if(!lpPrestoLink || (lpPrestoLink->valid != TRUE)){
		ERROR_MSG("Invalid PrestoLink passed to PrestoWritePointXY");
		return FALSE;
	}

	if(lpPrestoLink->newPlot != TRUE){
		ERROR_MSG("You must create a new plot before sending data");
		return FALSE;
	}

	/* 
	 * Don't send any non-numbers. Note that this is a "silent" error
	 * (i.e. no message box).
	 */
	if(!_finite(y) || !_finite(x))
		return FALSE;

	if((lpPrestoLink->plotType & PRESTO_ERROR_BAR_X) || 
	   (lpPrestoLink->plotType & PRESTO_ERROR_BAR_Y)){
		ERROR_MSG("Invalid error bar specified!");
		return 0;
	}

	if(!WriteFile(lpPrestoLink->pipeHandle[PIPE_WRITE], &msg, 1, &bytes_written, NULL)){
		ERROR_MSG("Can't send read point msg");

		lpPrestoLink->valid = FALSE;
		return FALSE;
	}

	buf[0] = x;
	buf[1] = y;

	if(!WriteFile(lpPrestoLink->pipeHandle[PIPE_WRITE], buf, 8 /* 2*sizeof(float) */, 
		&bytes_written, NULL)){
		ERROR_MSG("Can't write point data");

		lpPrestoLink->valid = FALSE;
		return FALSE;
	}	

	return TRUE;
};

int PrestoWritePointXYDY(PrestoLink* lpPrestoLink, float x, float y, float dy)
{
	float buf[3];
	DWORD bytes_written;
	char msg = PP_READ_POINT;

	if(!lpPrestoLink || (lpPrestoLink->valid != TRUE)){
		ERROR_MSG("Invalid PrestoLink passed to PrestoWritePointXYDY");
		return FALSE;
	}

	if(lpPrestoLink->newPlot != TRUE){
		ERROR_MSG("You must create a new plot before sending data");
		return FALSE;
	}

	/* 
	 * Don't send any non-numbers. Note that this is a "silent" error
	 * (i.e. no message box).
	 */
	if(!_finite(dy) || !_finite(y) || !_finite(x))
		return FALSE;

	if((lpPrestoLink->plotType & PRESTO_ERROR_BAR_X) || 
	  !(lpPrestoLink->plotType & PRESTO_ERROR_BAR_Y)){
		ERROR_MSG("Invalid error bar specified!");
		return FALSE;
	}

	if(!WriteFile(lpPrestoLink->pipeHandle[PIPE_WRITE], &msg, 1, &bytes_written, NULL)){
		ERROR_MSG("Can't send read point msg");

		lpPrestoLink->valid = FALSE;
		return FALSE;
	}

	buf[0] = x;
	buf[1] = y;
	buf[2] = dy;

	if(!WriteFile(lpPrestoLink->pipeHandle[PIPE_WRITE], buf, 12 /* 3*sizeof(float) */, 
		&bytes_written, NULL)){
		ERROR_MSG("Can't write point data");

		lpPrestoLink->valid = FALSE;
		return FALSE;
	}	

	return TRUE;
};

int PrestoWritePointXYDX(PrestoLink* lpPrestoLink, float x, float y, float dx)
{
	float buf[3];
	DWORD bytes_written;
	char msg = PP_READ_POINT;

	if(!lpPrestoLink || (lpPrestoLink->valid != TRUE)){
		ERROR_MSG("Invalid PrestoLink passed to PrestoWritePointXYDX");
		return FALSE;
	}

	if(lpPrestoLink->newPlot != TRUE){
		ERROR_MSG("You must create a new plot before sending data");
		return FALSE;
	}

	/* 
	 * Don't send any non-numbers. Note that this is a "silent" error
	 * (i.e. no message box).
	 */
	if(!_finite(dx) || !_finite(y) || !_finite(x))
		return FALSE;

	if(!(lpPrestoLink->plotType & PRESTO_ERROR_BAR_X) || 
	    (lpPrestoLink->plotType & PRESTO_ERROR_BAR_Y)){
		ERROR_MSG("Invalid error bar specified!");
		return FALSE;
	}

	if(!WriteFile(lpPrestoLink->pipeHandle[PIPE_WRITE], &msg, 1, &bytes_written, NULL)){
		ERROR_MSG("Can't send read point msg");

		lpPrestoLink->valid = FALSE;
		return FALSE;
	}

	buf[0] = x;
	buf[1] = y;
	buf[2] = dx;

	if(!WriteFile(lpPrestoLink->pipeHandle[PIPE_WRITE], buf, 12 /* 3*sizeof(float) */, 
		&bytes_written, NULL)){
		ERROR_MSG("Can't write point data");

		lpPrestoLink->valid = FALSE;
		return FALSE;
	}	

	return TRUE;
};

int PrestoWritePointXYDXDY(PrestoLink* lpPrestoLink, float x, float y, float dx, float dy)
{
	float buf[4];
	DWORD bytes_written;
	char msg = PP_READ_POINT;

	if(!lpPrestoLink || (lpPrestoLink->valid != TRUE)){
		ERROR_MSG("Invalid PrestoLink passed to PrestoWritePointXYDXDY");
		return FALSE;
	}

	if(lpPrestoLink->newPlot != TRUE){
		ERROR_MSG("You must create a new plot before sending data");
		return FALSE;
	}

	/* 
	 * Don't send any non-numbers. Note that this is a "silent" error
	 * (i.e. no message box).
	 */
	if(!_finite(dy) || !_finite(dx) || !_finite(y) || !_finite(x))
		return FALSE;

	if(!(lpPrestoLink->plotType & PRESTO_ERROR_BAR_X) || 
	   !(lpPrestoLink->plotType & PRESTO_ERROR_BAR_Y)){
		ERROR_MSG("Invalid error bar specified!");
		return FALSE;
	}

	if(!WriteFile(lpPrestoLink->pipeHandle[PIPE_WRITE], &msg, 1, &bytes_written, NULL)){
		ERROR_MSG("Can't send read point msg");

		lpPrestoLink->valid = FALSE;
		return FALSE;
	}

	buf[0] = x;
	buf[1] = y;
	buf[2] = dx;
	buf[3] = dy;

	if(!WriteFile(lpPrestoLink->pipeHandle[PIPE_WRITE], buf, 16 /* 4*sizeof(float) */, 
		&bytes_written, NULL)){
		ERROR_MSG("Can't write point data");

		lpPrestoLink->valid = FALSE;
		return FALSE;
	}	

	return TRUE;
};

int PrestoWriteVector(PrestoLink* lpPrestoLink, 
					  float tail_x, float tail_y,
					  float head_x, float head_y)
{
	if(!PrestoWritePointXY(lpPrestoLink, tail_x, tail_y))
		return FALSE;

	return PrestoWritePointXY(lpPrestoLink, head_x, head_y);
}

int PrestoSetOption(PrestoLink* lpPrestoLink, int option, int value)
{
	if(!lpPrestoLink || (lpPrestoLink->valid != TRUE)){
		ERROR_MSG("Invalid PrestoLink passed to PrestoSetOption");
		return FALSE;
	}

	if(lpPrestoLink->newPlot != TRUE){
		ERROR_MSG("You must create a new plot before setting options");
		return FALSE;
	}

	switch(option){
		case PRESTO_SYMBOL_TYPE:
			if((value < PRESTO_POINT_CIRCLE) || (value > PRESTO_POINT_NONE)){
				ERROR_MSG("Illegal symbol passed to PrestoSetOption");
				return FALSE;
			}

			return SendInt(lpPrestoLink->pipeHandle[PIPE_WRITE], 
				PP_SYMBOL_TYPE, value);
		case PRESTO_SYMBOL_COLOR:
			if((value < PRESTO_CLR_BLACK) || (value > PRESTO_CLR_NONE)){
				ERROR_MSG("Illegal color passed to PrestoSetOption");
				return FALSE;
			}

			return SendInt(lpPrestoLink->pipeHandle[PIPE_WRITE], 
				PP_SYMBOL_COLOR, value);
		case PRESTO_SYMBOL_SIZE:
			if((value < 1) || (value > 10)){
				ERROR_MSG("Illegal size passed to PrestoSetOption");
				return FALSE;
			}

			return SendInt(lpPrestoLink->pipeHandle[PIPE_WRITE], 
				PP_SYMBOL_SIZE, value);
		case PRESTO_LINE_TYPE:
			if((value < PRESTO_LINE_SOLID) || (value > PRESTO_LINE_NONE)){
				ERROR_MSG("Illegal line type passed to PrestoSetOption");
				return FALSE;
			}

			return SendInt(lpPrestoLink->pipeHandle[PIPE_WRITE],
				PP_LINE_TYPE, value);
		case PRESTO_LINE_COLOR:
			if((value < PRESTO_CLR_BLACK) || (value > PRESTO_CLR_NONE)){
				ERROR_MSG("Illegal color passed to PrestoSetOption");
				return FALSE;
			}

			return SendInt(lpPrestoLink->pipeHandle[PIPE_WRITE],
				PP_LINE_COLOR, value);
		case PRESTO_LINE_SIZE:
			if((value < 1) || (value > 5)){
				ERROR_MSG("Illegal size passed to PrestoSetOption");
				return FALSE;
			}

			return SendInt(lpPrestoLink->pipeHandle[PIPE_WRITE], 
				PP_LINE_SIZE, value);
		case PRESTO_GRAPH_TITLE_FONT_SIZE:
			if((value < 0) || (value > 100)){
				ERROR_MSG("Illegal font size passed to PrestoSetOption");
				return FALSE;
			}

			return SendInt(lpPrestoLink->pipeHandle[PIPE_WRITE], 
				PP_GRAPH_TITLE_SIZE, value);
		case PRESTO_X_LABEL_FONT_SIZE:
			if((value < 0) || (value > 100)){
				ERROR_MSG("Illegal font size passed to PrestoSetOption");
				return FALSE;
			}

			return SendInt(lpPrestoLink->pipeHandle[PIPE_WRITE], 
				PP_X_LABEL_SIZE, value);
		case PRESTO_Y_LABEL_FONT_SIZE:
			if((value < 0) || (value > 100)){
				ERROR_MSG("Illegal font size passed to PrestoSetOption");
				return FALSE;
			}

			return SendInt(lpPrestoLink->pipeHandle[PIPE_WRITE], 
				PP_Y_LABEL_SIZE, value);
	};
	
	ERROR_MSG("Illegal option passed to PrestoSetOption");
	return FALSE;
}

int PrestoSetCustomColor(PrestoLink* lpPrestoLink, int option, char r, char g, char b)
{
	DWORD bytes_written;
	char msg;

	if(!lpPrestoLink || (lpPrestoLink->valid != TRUE)){
		ERROR_MSG("Invalid PrestoLink passed to PrestoSetCustomColor");
		return FALSE;
	}

	if(lpPrestoLink->newPlot != TRUE){
		ERROR_MSG("You must create a new plot before setting colors");
		return FALSE;
	}

	switch(option){
		case PRESTO_SYMBOL_COLOR:
			msg = PP_SYMBOL_CUSTOM_COLOR;
			break;
		case PRESTO_LINE_COLOR:
			msg = PP_LINE_CUSTOM_COLOR;
			break;
		default:
			ERROR_MSG("Illegal option passed to PrestoSetCustomColor");
			return FALSE;
	};

	if(!WriteFile(lpPrestoLink->pipeHandle[PIPE_WRITE], &msg, 1, &bytes_written, NULL)){
		ERROR_MSG("Can't send custom color msg");

		lpPrestoLink->valid = FALSE;
		return FALSE;
	}
	
	if(!WriteFile(lpPrestoLink->pipeHandle[PIPE_WRITE], &r, 1, &bytes_written, NULL)){
		ERROR_MSG("Can't send custom color R");
		
		lpPrestoLink->valid = FALSE;
		return FALSE;
	}

	if(!WriteFile(lpPrestoLink->pipeHandle[PIPE_WRITE], &g, 1, &bytes_written, NULL)){
		ERROR_MSG("Can't send custom color G");
		
		lpPrestoLink->valid = FALSE;
		return FALSE;
	}

	if(!WriteFile(lpPrestoLink->pipeHandle[PIPE_WRITE], &b, 1, &bytes_written, NULL)){
		ERROR_MSG("Can't send custom color B");
		
		lpPrestoLink->valid = FALSE;
		return FALSE;
	}

	return TRUE;
}


int PrestoSetScale(PrestoLink* lpPrestoLink, int axis, float min, float max)
{
	DWORD bytes_written;
	char msg;

	switch(axis){
		case PRESTO_X_AXIS:
			msg = PP_SCALE_X;
			break;
		case PRESTO_Y_AXIS:
			msg = PP_SCALE_Y;
			break;
		default:
			ERROR_MSG("Illegal axis passed to PrestoSetScale");
			return FALSE;
	};

	if(!lpPrestoLink || (lpPrestoLink->valid != TRUE)){
		ERROR_MSG("Invalid PrestoLink passed to PrestoSetScale");
		return FALSE;
	}

	if(!WriteFile(lpPrestoLink->pipeHandle[PIPE_WRITE], &msg, 1, &bytes_written, NULL)){
		ERROR_MSG("Can't send world scaling msg");

		lpPrestoLink->valid = FALSE;
		return FALSE;
	}
	
	if(!WriteFile(lpPrestoLink->pipeHandle[PIPE_WRITE], &min, sizeof(float), &bytes_written, NULL)){
		ERROR_MSG("Can't send minimum scale value");
		
		lpPrestoLink->valid = FALSE;
		return FALSE;
	}

	if(!WriteFile(lpPrestoLink->pipeHandle[PIPE_WRITE], &max, sizeof(float), &bytes_written, NULL)){
		ERROR_MSG("Can't send maximum scale value");
		
		lpPrestoLink->valid = FALSE;
		return FALSE;
	}

	return TRUE;
}

int PrestoSetTickSpacing(PrestoLink* lpPrestoLink, int axis, float major, float minor)
{
	DWORD bytes_written;
	char msg;

	switch(axis){
		case PRESTO_X_AXIS:
			msg = PP_TICK_X;
			break;
		case PRESTO_Y_AXIS:
			msg = PP_TICK_Y;
			break;
		default:
			ERROR_MSG("Illegal axis passed to PrestoSetTickSpacing");
			return FALSE;
	};

	if(major < 0.0f){
		ERROR_MSG("Negative major tick spacing in PrestoSetTickSpacing");
		return FALSE;
	}

	if(minor < 0.0f){
		ERROR_MSG("Negative minor tick spacing in PrestoSetTickSpacing");
		return FALSE;
	}

	if(!lpPrestoLink || (lpPrestoLink->valid != TRUE)){
		ERROR_MSG("Invalid PrestoLink passed to PrestoSetTickSpacing");
		return FALSE;
	}

	if(!WriteFile(lpPrestoLink->pipeHandle[PIPE_WRITE], &msg, 1, &bytes_written, NULL)){
		ERROR_MSG("Can't send tick spacing msg");

		lpPrestoLink->valid = FALSE;
		return FALSE;
	}
	
	if(!WriteFile(lpPrestoLink->pipeHandle[PIPE_WRITE], &major, sizeof(float), &bytes_written, NULL)){
		ERROR_MSG("Can't send major tick spacing");
		
		lpPrestoLink->valid = FALSE;
		return FALSE;
	}

	if(!WriteFile(lpPrestoLink->pipeHandle[PIPE_WRITE], &minor, sizeof(float), &bytes_written, NULL)){
		ERROR_MSG("Can't send minor tick spacing");
		
		lpPrestoLink->valid = FALSE;
		return FALSE;
	}

	return TRUE;
}

int SendString(HANDLE pipe, char msg, const char *buff)
{
	DWORD bytes_written;
	int tmp_int;

	/* The string (if specified) */
	if(buff){
		if(!WriteFile(pipe, &msg, 1, &bytes_written, NULL)){
			ERROR_MSG("Can't send string msg");

			return FALSE;
		}
		
		/* Set the info buffer size */
		tmp_int = strlen(buff) + 1;
	
		if(!WriteFile(pipe, &tmp_int, sizeof(int), &bytes_written, NULL)){
			ERROR_MSG("Can't send string length");
		
			return FALSE;
		}
	
		/* Write the title buffer */
		if(!WriteFile(pipe, buff, tmp_int, &bytes_written, NULL)){
			ERROR_MSG("Can't send string");

			return FALSE;
		}	
	}
	else{
		ERROR_MSG("Can't send NULL string");
		return FALSE;
	}

	return TRUE;
}

int SendInt(HANDLE pipe, char msg, int value)
{
	DWORD bytes_written;

	if(!WriteFile(pipe, &msg, 1, &bytes_written, NULL)){
		ERROR_MSG("Can't send int msg");

		return FALSE;
	}
	
	if(!WriteFile(pipe, &value, sizeof(int), &bytes_written, NULL)){
		ERROR_MSG("Can't send int value");

		return FALSE;
	}

	return TRUE;
}

