//	(c) 2000-2002 J. Gans (jdg9@cornell.edu), Shalloway Lab, 
//	Cornell University. You can modify and freely distribute 
//	this code, but please keep this header intact.
#include "EquationUtil.h"
#include <math.h>
#include <float.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdio.h>

void EqElemBinary::Set(list<EqToken> &m_token, EqElem *m_p, list<EqToken> &varstack)
{
	// Set the parent pointer
	P = m_p;

	binary_op = m_token[0].token_type;

	m_token.Remove(0);
}

void EqElemBinary::Push(list<EqToken> &m_token, list<EqToken> &varstack)
{
	if(m_token.Length() == 0)
		// Nothing to do!
		return;

#ifdef _DEBUG
	if(L == NULL){
		throw "NULL L pointer in EqElemBinary::Push";
	}
#endif // _DEBUG
	
	if(R == NULL){
		if(isBinaryOp(m_token[0].token_type))
			throw "Cannot push binary elem on binary R";

		// Allocate space for the right element
		if(isUnaryOp(m_token[0].token_type)){
			R = new EqElemUnary;

			R->Set(m_token, this, varstack);
			R->Push(m_token, varstack);

			return;
		}

		// Must be a value
		R = new EqElemVal;

		R->Set(m_token, this, varstack);
		R->Push(m_token, varstack);

		return;
	}

	// May need to insert high precidence operators here!
	if((GetType() == BINARY_ADD) || (GetType() == BINARY_SUB)){
		if((m_token[0].token_type == BINARY_DIV) || 
		   (m_token[0].token_type == BINARY_MULTI) ||
		   (m_token[0].token_type == BINARY_POW)){
				
			// Insert the new operator
			EqElem *tmp_R = R;

			R = new EqElemBinary;

			((EqElemBinary*)R)->L = tmp_R;

			tmp_R->P = R;

			R->Set(m_token, this, varstack);

			R->Push(m_token, varstack);
			return;
		}
	}
	
	if((GetType() == BINARY_DIV) || (GetType() == BINARY_MULTI)){
		if(m_token[0].token_type == BINARY_POW){
				
			// Insert the new operator
			EqElem *tmp_R = R;

			R = new EqElemBinary;

			((EqElemBinary*)R)->L = tmp_R;

			tmp_R->P = R;

			R->Set(m_token, this, varstack);

			R->Push(m_token, varstack);
			return;
		}
	}

	if(P == NULL){
		// Allocate space for the parent element (must be a binary op)
		if(!isBinaryOp(m_token[0].token_type))
			throw "NULL P pointer and binary->!binary Push";

		P = new EqElemBinary;

		((EqElemBinary*)P)->L = this;

		P->Set(m_token, NULL, varstack);
		P->Push(m_token, varstack);
		return;
	}

	// Is the parent a unary brace or unary '-'?
	if((P->GetType() == UNARY_PAREN) || 
		((P->GetType() == UNARY_NEG) &&
		((m_token[0].token_type == BINARY_DIV) || 
		 (m_token[0].token_type == BINARY_MULTI) ||
		 (m_token[0].token_type == BINARY_POW)))){
		if(!isBinaryOp(m_token[0].token_type))
			throw "Missing a binary operator between braces or before '-'";

		// The brace operator must get special treatment
		// Insert the new operator
		EqElem *tmp_P = P;

		P = new EqElemBinary;

		// Attach the old parent to this new element
		UNARY(tmp_P)->C = P;

		// Make the current element the left operatand
		BINARY(P)->L = this;

		P->Set(m_token, tmp_P, varstack);
		P->Push(m_token, varstack);
		return;
	}

	P->Push(m_token, varstack);
}

double EqElemBinary::Eval(double *m_eval_stack)
{
	if(safe_mode){
		if(L == NULL)
			throw "Binary L is NULL";

		if(R == NULL)
			throw "Binary R is NULL";
	}
	
	switch (binary_op){
		case BINARY_ADD:
			return L->Eval(m_eval_stack) + R->Eval(m_eval_stack);
		case BINARY_DIV:
			if(safe_mode){
				double tmp = L->Eval(m_eval_stack) / R->Eval(m_eval_stack);

				if(_finite(tmp))
					return tmp;

				throw "Binary divide by zero";
			}

			return L->Eval(m_eval_stack) / R->Eval(m_eval_stack);
		case BINARY_MULTI:
			return L->Eval(m_eval_stack) * R->Eval(m_eval_stack);
		case BINARY_SUB:
			return L->Eval(m_eval_stack) - R->Eval(m_eval_stack);
		case BINARY_POW:
			if(safe_mode){
				double tmp = pow(L->Eval(m_eval_stack), R->Eval(m_eval_stack));

				if(_finite(tmp))
					return tmp;

				throw "Binary pow range error";
			}

			return pow(L->Eval(m_eval_stack), R->Eval(m_eval_stack));
	};

	throw "Unknown binary operator";

	return -1;
}

#ifdef _DEBUG
void EqElemBinary::DebugEval(FILE *fout, list<EqToken> &varstack)
{
	if(safe_mode){
		if(L == NULL)
			throw "Binary L is NULL";

		if(R == NULL)
			throw "Binary R is NULL";
	}
	
	switch (binary_op){
		case BINARY_ADD:
			fprintf(fout,"(");
			L->DebugEval(fout, varstack);
			fprintf(fout,"+");
			R->DebugEval(fout, varstack);
			fprintf(fout,")");
			return;
		case BINARY_DIV:
			fprintf(fout,"(");
			L->DebugEval(fout, varstack);
			fprintf(fout,"/");
			R->DebugEval(fout, varstack);
			fprintf(fout,")");
			return;
		case BINARY_MULTI:
			fprintf(fout,"(");
			L->DebugEval(fout, varstack);
			fprintf(fout,"*");
			R->DebugEval(fout, varstack);
			fprintf(fout,")");
			return;
		case BINARY_SUB:
			fprintf(fout,"(");
			L->DebugEval(fout, varstack);
			fprintf(fout,"-");
			R->DebugEval(fout, varstack);
			fprintf(fout,")");
			return;
		case BINARY_POW:
			fprintf(fout,"(");
			L->DebugEval(fout, varstack);
			fprintf(fout,"^");
			R->DebugEval(fout, varstack);
			fprintf(fout,")");
			return;
	};

	throw "Unknown binary operator";

	return;
}
#endif // _DEBUG

EqElem* EqElemBinary::Copy(EqElem *m_p)
{
	EqElem* tmp = new EqElemBinary;

	((EqElemBinary*)tmp)->binary_op = binary_op;

	// Recursively copy the tree elements
	((EqElemBinary*)tmp)->P = m_p;

#ifdef _DEBUG
	if(L == NULL)
		throw "NULL L in EqElemBinary::Copy";
#endif // _DEBUG

	((EqElemBinary*)tmp)->L = L->Copy(tmp);

#ifdef _DEBUG
	if(R == NULL)
		throw "NULL R in EqElemBinary::Copy";
#endif // _DEBUG

	((EqElemBinary*)tmp)->R = R->Copy(tmp);

	return tmp;
}


BOOL EqElemBinary::IsEqual(EqElem *m_elem)
{
#ifdef _DEBUG
	if(m_elem == NULL)
		throw "NULL pointer in EqElemBinary::IsEqual";
#endif // _DEBUG

	if(m_elem->GetType() != binary_op)
		return FALSE;

	if(L->IsEqual(BINARY(m_elem)->L) && R->IsEqual(BINARY(m_elem)->R))
		return TRUE;

	// Explicitly handle binary operators that commute, i.e.
	// a + b = b + a
	// a * b = b * a
	if((m_elem->GetType() == BINARY_ADD) || 
	   (m_elem->GetType() == BINARY_MULTI)){
		if(L->IsEqual(BINARY(m_elem)->R) && R->IsEqual(BINARY(m_elem)->L))
			return TRUE;
	}

	return FALSE;
}

EqElem* EqElemBinary::Deriv(int m_var_index, EqElem *m_p)
{
	// The derivative of a binary operator is, in general, representable
	// as a binary operator.
	EqElem* tmpBinary = new EqElemBinary(m_p);
	EqElem* tmp;

	switch(binary_op){
		// dl/dx + dr/dx
		case BINARY_ADD:
			BINARY(tmpBinary)->binary_op = BINARY_ADD;
			BINARY(tmpBinary)->L = L->Deriv(m_var_index, tmpBinary);
			BINARY(tmpBinary)->R = R->Deriv(m_var_index, tmpBinary);
			break;
		// dl/dx - dr/dx
		case BINARY_SUB:
			BINARY(tmpBinary)->binary_op = BINARY_SUB;
			BINARY(tmpBinary)->L = L->Deriv(m_var_index, tmpBinary);
			BINARY(tmpBinary)->R = R->Deriv(m_var_index, tmpBinary);
			break;
		// (dl/dx*r - dr/dx*l)/r^2
		case BINARY_DIV:
			BINARY(tmpBinary)->binary_op = BINARY_DIV;

			BINARY(tmpBinary)->L = new EqElemBinary(tmpBinary);
			tmp = BINARY(tmpBinary)->L;

			BINARY(tmp)->binary_op = BINARY_SUB;
			BINARY(tmp)->L = new EqElemBinary(tmpBinary);
			
			BINARY(BINARY(tmp)->L)->binary_op = BINARY_MULTI;
			
			BINARY(BINARY(tmp)->L)->L = L->Deriv(m_var_index, BINARY(tmp)->L);
			BINARY(BINARY(tmp)->L)->R = R->Copy(BINARY(tmp)->L);

			BINARY(tmp)->R = new EqElemBinary(tmpBinary);

			BINARY(BINARY(tmp)->R)->binary_op = BINARY_MULTI;
			
			BINARY(BINARY(tmp)->R)->L = R->Deriv(m_var_index, BINARY(tmp)->R);
			BINARY(BINARY(tmp)->R)->R = L->Copy(BINARY(tmp)->R);

			BINARY(tmpBinary)->R = new EqElemBinary(tmpBinary);
			
			BINARY(BINARY(tmpBinary)->R)->binary_op = BINARY_POW;
			BINARY(BINARY(tmpBinary)->R)->L = R->Copy(BINARY(tmpBinary)->R);
			BINARY(BINARY(tmpBinary)->R)->R = MakeConst(2.0, BINARY(tmpBinary)->R);
			break;
		// dl/dx*r + dr/dx*l
		case BINARY_MULTI:
			BINARY(tmpBinary)->binary_op = BINARY_ADD;
			
			BINARY(tmpBinary)->L = new EqElemBinary(tmpBinary);

			tmp = BINARY(tmpBinary)->L;

			BINARY(tmp)->binary_op = BINARY_MULTI;
			BINARY(tmp)->L = L->Deriv(m_var_index, tmp);
			BINARY(tmp)->R = R->Copy(tmp);

			BINARY(tmpBinary)->R = new EqElemBinary(tmpBinary);

			tmp = BINARY(tmpBinary)->R;

			BINARY(tmp)->binary_op = BINARY_MULTI;
			BINARY(tmp)->L = L->Copy(tmp);
			BINARY(tmp)->R = R->Deriv(m_var_index, tmp);
			break;
		// (l^r)*(r/l*dl/dx + log(l)*dr/dx)
		case BINARY_POW:
			BINARY(tmpBinary)->binary_op = BINARY_MULTI;
			
			BINARY(tmpBinary)->L = Copy(tmpBinary);

			BINARY(tmpBinary)->R = new EqElemBinary(tmpBinary);

			tmp = BINARY(tmpBinary)->R;

			BINARY(tmp)->binary_op = BINARY_ADD;

			BINARY(tmp)->L = new EqElemBinary(tmp);

			BINARY(BINARY(tmp)->L)->binary_op = BINARY_MULTI;

			BINARY(BINARY(tmp)->L)->L = new EqElemBinary(tmp);

			BINARY(BINARY(BINARY(tmp)->L)->L)->binary_op = BINARY_DIV;
			BINARY(BINARY(BINARY(tmp)->L)->L)->L = 
				R->Copy(BINARY(BINARY(tmp)->L)->L);
			BINARY(BINARY(BINARY(tmp)->L)->L)->R = 
				L->Copy(BINARY(BINARY(tmp)->L)->L);
			
			BINARY(BINARY(tmp)->L)->R = 
				L->Deriv(m_var_index, BINARY(tmp)->L);

			BINARY(tmp)->R = new EqElemBinary(tmp);
			BINARY(BINARY(tmp)->R)->binary_op = BINARY_MULTI;
			
			BINARY(BINARY(tmp)->R)->L = new EqElemUnary(BINARY(tmp)->L);
			
			UNARY(BINARY(BINARY(tmp)->R)->L)->unary_op = UNARY_LOG;
			UNARY(BINARY(BINARY(tmp)->R)->L)->C = 
				L->Copy(BINARY(BINARY(tmp)->R)->L);
		
			BINARY(BINARY(tmp)->R)->R = 
				R->Deriv(m_var_index, BINARY(tmp)->R);

			break;
	};

	return tmpBinary;
}

void EqElemBinary::CleanUp()
{
	// Delete the childeren
	if(L){
		L->CleanUp();

		delete L;
		L = NULL;
	}

	if(R){
		R->CleanUp();

		delete R;
		R = NULL;
	}
}