//	(c) 2000-2002 J. Gans (jdg9@cornell.edu), Shalloway Lab, 
//	Cornell University. You can modify and freely distribute 
//	this code, but please keep this header intact.
#include "EquationUtil.h"
#include <math.h>
#include <float.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdio.h>

// Global variables
int EqElem::safe_mode = TRUE;
//////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////

SEquation::SEquation()
{
	version = EQ_VERSION;

	tree = NULL;
	last_error = NULL;
	input_string = NULL;
	eval_stack = NULL;
	stack_valid = NULL;
}

SEquation::SEquation(const char *m_input)
{
	version = EQ_VERSION;

	tree = NULL;
	last_error = NULL;
	input_string = NULL;
	eval_stack = NULL;
	stack_valid = NULL;

	Set(m_input);
}

SEquation::SEquation(SEquation &m_copy)
{
	version = EQ_VERSION;

	tree = NULL;
	last_error = NULL;
	input_string = NULL;
	eval_stack = NULL;
	stack_valid = NULL;

	(*this) = m_copy;
}

void SEquation::Reset()
{

	if(tree){
		tree->CleanUp();
		delete tree;
		tree = NULL;
	}

	if(input_string){
		delete [] input_string;
		input_string = NULL;
	}

	if(eval_stack){
		delete [] eval_stack;
		eval_stack = NULL;
	}
	
	if(stack_valid){
		delete [] stack_valid;
		stack_valid = NULL;
	}

	while(varstack.Length() > 0)
		varstack.Pop();
}

SEquation::~SEquation()
{
	Reset();
}

void SEquation::ClearError()
{
	last_error = NULL;
}

BOOL SEquation::Set(const char *m_input)
{
	// Reset the tree
	Reset();

	if(m_input == NULL){
		last_error = "Null Parse() string";
		return FALSE;
	}

	// Parsing the user's input string has several stages:
	//
	//	(a) Process the string to remove white space and convert to lower
	//		case.
	//	(b) Use the output of step (a) to produce a list of tokens. In the
	//		process, verify the syntax of equatiuon string.
	//	(c) Convert the list of tokens to an Abstract Syntax Tree (AST).

	// Make a copy of the input sting
	int len = strlen(m_input);
	int count = 0;
	char *tmp_buffer = new char[len + 1];

	// Remove all white space from the input string and convert to 
	// lower case.
	for(int i = 0;i < len;i++){
		if(!isspace(m_input[i])){
			tmp_buffer[count++] = tolower(m_input[i]);
		}
	}

	// Terminate the string
	tmp_buffer[count] = '\0';

	// Save a copy of this string for future reference
	input_string = new char[strlen(tmp_buffer) + 1];

	strcpy(input_string, tmp_buffer);

	// Deallocate the temp buffer
	delete [] tmp_buffer;

	// Turn the string into a list of token
	list<EqToken> tmplist;

	// "Tokenize" the input string
	last_error = Tokenize(input_string, tmplist);

	// Did we successfully tokenize the equation?
	if(last_error){
		Reset();
		return FALSE;
	}

	// Convert the list of tokens into an abstract syntax
	// tree (AST).
	MakeAST(tmplist);

	// Did we successfully parse the equation?
	if(last_error){
		Reset();
		return FALSE;
	}
	
	if(tree == NULL){
		last_error = "NULL tree";
		Reset();
		return FALSE;
	}

	// Make tree point to the root node
	while(tree->P != NULL)
		tree = tree->P;

	// Allocate memory for the eval_stack
	if(varstack.Length() != 0){
		eval_stack = new double [varstack.Length()];
		stack_valid = new BOOL [varstack.Length()];

		// None of the stack elements have been set yet
		for(i = 0;i < varstack.Length();i++){
			stack_valid[i] = FALSE;
		}
	}

	return TRUE;
}

void SEquation::MakeAST(list<EqToken> &tokenlist)
{
	
	if(tokenlist.Length() == 0){
		last_error = "No equation";
		return;
	}

	try{
		// The first operator is either unary or a value
		if(isUnaryOp(tokenlist[0].token_type)){
			tree = new EqElemUnary;

			tree->Set(tokenlist, NULL, varstack);	
			tree->Push(tokenlist, varstack);

			return;
		}

		// Must be a value
		tree = new EqElemVal;

		tree->Set(tokenlist, NULL, varstack);	
		tree->Push(tokenlist, varstack);
	}
	catch(char *m_error){
		last_error = m_error;
	}
}


char *SEquation::Tokenize(char *m_input, list<EqToken> &tokenlist)
{
	// Enmerate all of the valid delimeters
	char mask[] = "()[]{}+-*/^";

	// Start at the head of the string
	char *head = m_input;
	char *ptr;

	// Keep track of any errors
	char *error;

	/////////////////////////////////////////////////////////////
	// Parse the input string
	/////////////////////////////////////////////////////////////
	while(*head != '\0'){
		EqToken tmpToken;

		ptr = strpbrk(head, mask);

		if(ptr == NULL){
			int len = strlen(head);
			error = tmpToken.Set(head, len);

			if(error)
				return error;

			head = head + len;
		}
		else{ // ptr != NULL
			if(ptr == head){
				error = tmpToken.Set(head, 1);
				
				if(error)
					return error;

				head = ptr + 1;
			}
			else{
				error = tmpToken.Set(head, ptr - head);

				if(error)
					return error;

				head = ptr;
			}
		}

		tokenlist.Push(tmpToken);
	}

	/////////////////////////////////////////////////////////////
	// Check the list of tokens for correct syntax
	/////////////////////////////////////////////////////////////
	for(int i = 0;i < tokenlist.Length();i++){

		// Do we have a dangling expression?
		if((tokenlist.Length() == i + 1)){
			switch(tokenlist[i].token_type){
				case UNARY_R_BRACE:
				case VALUE_CONST:
				case VALUE_VAR:
					break;
				default:
					return "Dangling expression";
			};
		}

		switch(tokenlist[i].token_type){
			//
			// Resolve {+,-} -> {binary, unary} ambiguity
			//
			case BINARY_ADD:
				// Check to make sure that this is really a binary
				// addition and not a unary positive sign
				if(i == 0){
					// It's the first token - must be unary!
					tokenlist[i].token_type = UNARY_PLUS;
				}
				else{
					if(isBinaryOp(tokenlist[i - 1].token_type) ||
					  (tokenlist[i - 1].token_type == UNARY_L_BRACE) ||
					  (tokenlist[i - 1].token_type == UNARY_PLUS) ||
					  (tokenlist[i - 1].token_type == UNARY_NEG)){
						tokenlist[i].token_type = UNARY_PLUS;
					}
					//else{
					//	tokenlist[i].token_type = BINARY_ADD;
					//}
				}
				break;
			case BINARY_SUB:
				// Check to make sure that this is really a binary
				// subtraction and not unary negation
				if(i == 0){
					// It's the first token - must be unary!
					tokenlist[i].token_type = UNARY_NEG;
				}
				else{
					if(isBinaryOp(tokenlist[i - 1].token_type) ||
					  (tokenlist[i - 1].token_type == UNARY_L_BRACE) ||
					  (tokenlist[i - 1].token_type == UNARY_PLUS) ||
					  (tokenlist[i - 1].token_type == UNARY_NEG)){
						tokenlist[i].token_type = UNARY_NEG;
					}
					//else{
					//	tokenlist[i].token_type = BINARY_NEG;
					//}
				}
				break;
		};

		if(isBinaryOp(tokenlist[i].token_type)){
			// Make sure binary operators DONT start the equation
			if(i == 0){
				return "Binary operator at start of Equation";
			}
			else{
				// Binary operators can not follow other binary
				// operators (or the '(' unary operator).
				if(isBinaryOp(tokenlist[i-1].token_type) ||
				  (tokenlist[i-1].token_type == UNARY_L_BRACE)){
					return "Misplaced binary operator";
				}
			}

			if(i < tokenlist.Length() - 1){
				if(tokenlist[i+1].token_type == UNARY_R_BRACE)
					return "Misplaced binary operator";
			}
		}

		// Does every operator that needs a brace, have a brace?
		if(needsBrace(tokenlist[i].token_type)){
			if((i == tokenlist.Length() - 1) || 
			   (tokenlist[i + 1].token_type != UNARY_L_BRACE))
				return "Missing '('";
		}

		// Is the total sum of '(' and ')' zero?
		if(braceSum(tokenlist) != 0)
			return "Unmatched '(' or ')'";

		if(tokenlist[i].token_type == UNARY_L_BRACE){
			int brace = MatchUnaryBrace(tokenlist, i);

			if(brace == -1)
				return "Unmatched '('";

			// Every set of "()" must contain at least one const or
			// var value.
			int numval = numValue(tokenlist, i, brace);

			if(numval == 0)
				return "Empty '()'";
		}
	}

	return NULL;
}

// Define the arguments of the equation. Since the number
// of arguments is not known in advance, several overloaded
// operators are provided. This is a limitation (since a large
// number of independant variables will require a new function
// and a recompile) it avoids the nastyness variable length 
// argument lists.
BOOL SEquation::DefineIndepVar()
{
	// Flush the independant variable list
	indep_var.Init();

	return TRUE;
}

BOOL SEquation::DefineIndepVar(char *m_one)
{
	// make the symbol m_one the independant variable
	if(m_one == NULL)
		throw "NULL pointer in DefineIndepVar(1)";

	// Flush the independant variable list
	indep_var.Init();

	for(int i = 0;i < varstack.Length();i++){
		if(strcmp(varstack[i].id, m_one) == 0){
			indep_var.Push(i);

			return TRUE;
		}
	}

	return FALSE;
}

BOOL SEquation::DefineIndepVar(char *m_one, char *m_two)
{
	// make the symbols m_one and m_two the independant variables
	if(m_one == NULL)
		throw "NULL pointer in DefineIndepVar(2): arg 0";

	if(m_two == NULL)
		throw "NULL pointer in DefineIndepVar(2): arg 1";

	if(strcmp(m_one, m_two) == 0)
		throw "Repeated variable in DefineIndepVar(2)";

	// Flush the independant variable list
	indep_var.Init();

	for(int i = 0;i < varstack.Length();i++){
		if(strcmp(varstack[i].id, m_one) == 0){
			indep_var.Push(i);
		}

		if(strcmp(varstack[i].id, m_two) == 0){
			indep_var.Push(i);
		}

		if(indep_var.Length() == 2)
			return TRUE;
	}

	return FALSE;
}

BOOL SEquation::DefineIndepVar(char *m_one, char *m_two, char *m_three)
{
	// make the symbols m_one and m_two the independant variables
	if(m_one == NULL)
		throw "NULL pointer in DefineIndepVar(3): arg 0";

	if(m_two == NULL)
		throw "NULL pointer in DefineIndepVar(3): arg 1";

	if(m_three == NULL)
		throw "NULL pointer in DefineIndepVar(3): arg 2";

	if(strcmp(m_one, m_two) == 0)
		throw "Repeated variable in DefineIndepVar(3)";

	if(strcmp(m_one, m_three) == 0)
		throw "Repeated variable in DefineIndepVar(3)";

	if(strcmp(m_two, m_three) == 0)
		throw "Repeated variable in DefineIndepVar(3)";

	// Flush the independant variable list
	indep_var.Init();

	for(int i = 0;i < varstack.Length();i++){
		if(strcmp(varstack[i].id, m_one) == 0){
			indep_var.Push(i);
		}

		if(strcmp(varstack[i].id, m_two) == 0){
			indep_var.Push(i);
		}

		if(strcmp(varstack[i].id, m_three) == 0){
			indep_var.Push(i);
		}

		if(indep_var.Length() == 3)
			return TRUE;
	}

	return FALSE;
}

void SEquation::SetVar(int num, float m_value)
{
#ifdef _DEBUG
	if(eval_stack == NULL)
		throw "NULL eval_stack in SetVar";

	if(stack_valid == NULL)
		throw "NULL stack_valid in SetVar";
#endif // _DEBUG

	if(num < 0)
		throw "Index underflow in SetVar";

	if(num >= varstack.Length())
		throw "Index overflow in SetVar";

	stack_valid[num] = TRUE;

	eval_stack[num] = m_value;
}

float SEquation::GetVar(int num)
{
#ifdef _DEBUG
	if(eval_stack == NULL)
		throw "NULL eval_stack in GetVar";

	if(stack_valid == NULL)
		throw "NULL stack_valid in GetVar";
#endif // _DEBUG

	if(num < 0)
		throw "Index underflow in GetVar";

	if(num >= varstack.Length())
		throw "Index overflow in GetVar";

	return (float)(eval_stack[num]);
}

float SEquation::operator () ()
{
#ifdef _DEBUG
	if(tree == NULL)
		throw "NULL tree pointer";
#endif // _DEBUG

	if(indep_var.Length() != 0)
		throw "Incorrect number of arguments (not 0)";

	try{
		return (float)(tree->Eval(eval_stack));
	}
	catch(char *m_error){
		last_error = m_error;
		return -1.0f;
	}
}

float SEquation::operator () (float m_one)
{
#ifdef _DEBUG
	if(tree == NULL)
		throw "NULL tree pointer";
#endif // _DEBUG

	if(indep_var.Length() != 1)
		throw "Incorrect number of arguments (not 1)";

	eval_stack[indep_var[0]] = m_one;

	try{
		return (float)(tree->Eval(eval_stack));
	}
	catch(char *m_error){
		last_error = m_error;
		return -1.0f;
	}
}

float SEquation::operator () (float m_one, float m_two)
{
#ifdef _DEBUG
	if(tree == NULL)
		throw "NULL tree pointer";
#endif // _DEBUG

	if(indep_var.Length() != 2)
		throw "Incorrect number of arguments (not 2)";

	eval_stack[indep_var[0]] = m_one;
	eval_stack[indep_var[1]] = m_two;

	try{
		return (float)(tree->Eval(eval_stack));
	}
	catch(char *m_error){
		last_error = m_error;
		return -1.0f;
	}
}

float SEquation::operator () (float m_one, float m_two, float m_three)
{
#ifdef _DEBUG
	if(tree == NULL)
		throw "NULL tree pointer";
#endif // _DEBUG

	if(indep_var.Length() != 3)
		throw "Incorrect number of arguments (not 3)";

	eval_stack[indep_var[0]] = m_one;
	eval_stack[indep_var[1]] = m_two;
	eval_stack[indep_var[2]] = m_three;

	try{
		return (float)(tree->Eval(eval_stack));
	}
	catch(char *m_error){
		last_error = m_error;
		return -1.0f;
	}
}

#ifdef _DEBUG
void SEquation::operator () (FILE *fout)
{
#ifdef _DEBUG
	if(tree == NULL)
		throw "NULL tree pointer";
#endif // _DEBUG

	tree->DebugEval(fout, varstack);

	fprintf(fout, "\n");
}
#endif // _DEBUG

char *SEquation::GetVarName(int num)
{
	if(num < 0){
		last_error = "Underflow in GetVarName";
		return NULL;
	}

	if(num >= varstack.Length()){
		last_error = "Overflow in GetVarName";
		return NULL;
	}

	return varstack[num].id;
}

int SEquation::NumVar()
{
	return varstack.Length();
}

// Is the current equation free of errors and ready for 
// evaluation?
BOOL SEquation::IsValid(BOOL m_test_stack /* = TRUE */)
{
	// If an error has occured, the user MUST call ClearError()
	// before evalutating the equation.
	if(last_error)
		return FALSE;

	if(tree == NULL)
		return FALSE;

	// Does the equation contain any non-constant variables?
	if(int num = varstack.Length()){
		if(eval_stack == NULL)
			return FALSE;

		if(stack_valid == NULL)
			return FALSE;

		// Do we care if variables have been set?
		if(m_test_stack){
			for(int i = 0;i < num;i++){
				if(!IsIndepVar(i) && (stack_valid[i] == FALSE)){
					return FALSE;
				}
			}
		}
	}

	return TRUE;
}



SEquation & SEquation::operator =(SEquation &m_copy)
{
	// Remove any existing equation information
	Reset();

	// Make sure that the equation we're about to copy is valid
	if(!m_copy.IsValid(FALSE)){
		last_error = "Can not copy an invalid equation";
		return (*this);
	}
	
	// Copy the input string
	input_string = new char [strlen(m_copy.input_string) + 1];
	strcpy(input_string, m_copy.input_string);

	// Copy the variable stack
	varstack = m_copy.varstack;

	// Copy the evaluation stack
	eval_stack = new double[varstack.Length()];
	stack_valid = new BOOL [varstack.Length()];

	for(int i = 0;i < varstack.Length();i++){
		eval_stack[i] = m_copy.eval_stack[i];
		stack_valid[i] = m_copy.stack_valid[i];
	}

	// Copy the independant variable index
	indep_var = m_copy.indep_var;
	
	try{
		// Copy the tree elements
		tree = m_copy.tree->Copy(NULL);
	}
	catch(char* m_error){
		Reset();
		last_error = m_error;
	}

	return (*this);
}

SEquation SEquation::Deriv(char *m_var)
{
	if(m_var == NULL){
		last_error = "NULL variable argument to Deriv()";
		return (*this);
	}

	// Which variable are we differentiating with respect to?
	int var_index;
	
	for(var_index = 0;var_index < varstack.Length();var_index ++){
		if(strcmp(m_var, varstack[var_index].id) == 0)
			break;
	}

	return Deriv(var_index);
}

SEquation SEquation::Deriv(int m_var_index)
{
	if(!IsValid(FALSE)){
		last_error = "Unable to differentiate invalid equation!";
		return (*this);
	}

	SEquation deriv;

	try{
		if(m_var_index == varstack.Length())
			throw "Variable not found in Deriv()";

		/////////////////////////////////////////////////////////////////////
		// Copy the equation [except for the tree!]
		/////////////////////////////////////////////////////////////////////
		// Copy the input string
		deriv.input_string = new char [strlen(input_string) + 1];
		strcpy(deriv.input_string, input_string);

		// Copy the variable stack
		deriv.varstack = varstack;

		// Copy the evaluation stack
		deriv.eval_stack = new double[deriv.varstack.Length()];
		deriv.stack_valid = new BOOL [deriv.varstack.Length()];

		for(int i = 0;i < deriv.varstack.Length();i++){
			deriv.eval_stack[i] = eval_stack[i];
			deriv.stack_valid[i] = stack_valid[i];
		}

		// Copy the independant variable index
		deriv.indep_var = indep_var;

		// Build the tree with the derivative of the current equation
		deriv.tree = tree->Deriv(m_var_index, NULL);
	}
	catch(char *m_error){
		deriv.Reset();

		deriv.last_error = m_error;
	}

	return deriv;
}


void SEquation::RemoveUnaryElem(EqElemUnary *ptr)
{
	if(ptr == NULL)
		throw "NULL ptr in RemoveUnaryElem";

	// Is this the root node?
	if(ptr == tree){
		// Connect parent to child
		tree = ptr->C;

		// Make this the root node
		tree->P = NULL;

		// Sever connection between old elem and child
		ptr->C = NULL;

		// Remove element
		ptr->CleanUp();
		delete ptr;

		return;
	}

	if(isBinaryOp(ptr->P->GetType())){

		if(ptr == BINARY(ptr->P)->L){
			// Connect parent to child
			BINARY(ptr->P)->L = ptr->C;

			// Connect child to parent
			ptr->C->P = ptr->P;
			
			// Sever connection between old elem and child
			ptr->C = NULL;

			// Remove element
			ptr->CleanUp();
			delete ptr;
		}
		else{ // ptr == BINARY(ptr->P)->R
			// Connect parent to child
			BINARY(ptr->P)->R = ptr->C;

			// Connect child to parent
			ptr->C->P = ptr->P;
			
			// Sever connection between old elem and child
			ptr->C = NULL;

			// Remove element
			ptr->CleanUp();
			delete ptr;
		}

		return;
	}

	// Must be unary op parent

	// Connect parent to child
	UNARY(ptr->P)->C = ptr->C;

	// Connect child to parent
	ptr->C->P = ptr->P;
			
	// Sever connection between old elem and child
	ptr->C = NULL;

	// Remove element
	ptr->CleanUp();
	delete ptr;
}

// Make ptr_P the parent element of ptr
void SEquation::InsertUnaryParent(EqElem *ptr, EqElemUnary *ptr_P)
{
	if(ptr == NULL)
		throw "NULL ptr in InsertParent";

	if(ptr_P == NULL)
		throw "NULL ptr_P in InsertParent";

	// Is ptr the root node?
	if(ptr == tree){
		tree = ptr_P;
		ptr_P->P = NULL;
		ptr_P->C = ptr;
		ptr->P = tree;
		return;
	}
	else{
		if(isBinaryOp(ptr->P->GetType())){
			if(BINARY(ptr->P)->L == ptr){
				BINARY(ptr->P)->L = ptr_P;
			}
			else{ // BINARY(ptr->P)->R == ptr
				BINARY(ptr->P)->R = ptr_P;
			}
				
			ptr_P->P = ptr->P;

			ptr_P->C = ptr;
			ptr->P = ptr_P;

			return;
		}

		if(isUnaryOp(ptr->P->GetType())){
			UNARY(ptr->P)->C = ptr_P;
			ptr_P->P = ptr->P;
			
			ptr_P->C = ptr;
			ptr->P = ptr_P;

			return;
		}

		throw "Illegal value op in InsertUnaryParent";
	}
}