#include <string.h>
#include <stdlib.h>

// Include excepts.h for error throwing
#include "../Excepts.h"

// Use stdio for output to prevent problems will the 
// construction and destruction of iostream objects.
#include <stdio.h>

// Local functions.
void *(__memory_allocate)(int size, char *filename, int line);
void __memory_deallocate(void *p, char *filename, int line);
void __memory_freeheap(struct heap_element *cur);
void __memory_report();
long __memory_leak(struct heap_element *cur);

#ifdef MEMORY_DEBUG

// Define the log file here if using the brain damaged MsDev win32 
// compiler that can't use the '=' in the define text box!
#define		MEMORY_DEBUG_LOG	"memlog.txt"

// Overload the global new and delete functions
void *operator new (size_t nb)
{
	return __memory_allocate(nb, __FILE__, __LINE__);
}

void *operator new (size_t nb, char *filename, int line)
{
	return __memory_allocate(nb, filename, line);
}

void operator delete (void * ptr)
{
	__memory_deallocate(ptr, __FILE__, __LINE__);
}

void operator delete (void * ptr, char *filename, int line)
{
	__memory_deallocate(ptr, filename, line);
}

#endif // MEMORY_DEBUG

struct heap_element {
    struct heap_element *prev;
    struct heap_element *left;
    struct heap_element *right;
    
    // Data
    void *ptr;
    int size;
    char *name;
    int line;
} *heap = NULL;

// ******************* Global variables ******************* 
// Are we done with the memory checking?
// This flag is required because some systems (i.e. WIN32) may
// continue to deallocate memory (using delete) AFTER the
// atexit() functions have been called!
int DoneWithMemoryReporting = 0;

// Have we started loggin memory use?
int StartedMemoryReporting = 0;

void *(__memory_allocate)(int size, char *filename, int line)
{
    static first_time = 1; // True
    struct heap_element *cur;
	
    if(first_time){
		first_time = 0; // False
		
		// Set the flag to indicate that we have started 
		// tracking memory use.
		StartedMemoryReporting = 1;

		// Register deallocation function with atexit()
		if(atexit(__memory_report)){
			fprintf(stderr,"Error! Can't register memory reporting!\n"
				"Exiting to system ...\n");
			throw Mem_Error("Error! Can't register memory reporting!\n"
				"Exiting to system ...\n", __FILE__, __LINE__);
		}

#ifdef MEMORY_DEBUG_LOG
		// We will be dumping the output of the allocation heap
		// to a file specified by the user. Make sure that the 
		// file is a valid one!
		//safe_ofstream fout(MEMORY_DEBUG_LOG, ios::app);
		FILE *fout = fopen(MEMORY_DEBUG_LOG, "a");
		if(fout){
			// Every thing is ok!
			fprintf(fout, 
				"-------------- Memory Log file initialized --------------\n");
			fclose(fout);
		}
		else{
			// Woops! Something is wrong.
			fprintf(stderr,"Error! Couldn't open the memory log file '%s"
				 "'.\nExiting ...\n", MEMORY_DEBUG_LOG);
			
			throw Mem_Error("Error! Couldn't open the memory log file\n", 
				__FILE__, __LINE__);
		}
#endif // MEMORY_DEBUG_LOG
    }
    
	// ************* Allocate the memory *************
	void *p = (void *)malloc(size);

	if(!first_time && DoneWithMemoryReporting)
		// Don't attempt to monitor memory usage if
		// the program is allready working through the
		// atexit stack or allready in new.
		return p;

    if(p == NULL){
		fprintf(stderr,"Error! Can not allocate memory in file %s "
			"at line %d.\nExiting ...\n", __FILE__, __LINE__);
#ifdef MEMORY_DEBUG_LOG
		FILE *fout = fopen(MEMORY_DEBUG_LOG, "a");
		if(fout){
			// Every thing is ok!
			fprintf(stderr,"Error! Can not allocate memory in file %s "
				"at line %d.\nExiting ...\n", __FILE__, __LINE__);
			fclose(fout);
		}
		else{
			// Woops! Something is wrong.
			fprintf(stderr,"Error! Couldn't open the memory log file '%s"
				 "'.\nExiting ...\n", MEMORY_DEBUG_LOG);
		}
#endif // MEMORY_DEBUG_LOG
		throw Mem_Error("Error! Can not allocate memory!\n",
				__FILE__, __LINE__);
    }
    
    // Warn users if they attempt to allocate a block of negative bytes
    if(size < 0){
		fprintf(stderr, "Warning! A negative byte block of "
			"memory was allocated\nin file %s at line %d.\n",
			filename, line);

#ifdef MEMORY_DEBUG_LOG
		FILE *fout = fopen(MEMORY_DEBUG_LOG, "a");
		if(fout){
			// Every thing is ok!
			fprintf(fout, "Warning! A negative byte block of "
				"memory was allocated\nin file %s at line %d.\n",
				filename, line);
			fclose(fout);
		}
		else{
			// Woops! Something is wrong.
			fprintf(stderr,"Error! Couldn't open the memory log file '%s'\n."
				, MEMORY_DEBUG_LOG);
		}
#endif // MEMORY_DEBUG_LOG
	}
    if(heap == NULL){
		 // Allocate memory for the heap binary tree.
		heap = (struct heap_element*)malloc(sizeof(struct heap_element));

		if(heap == NULL){
			fprintf(stderr, "Error! Can not allocate memory in file %s"
	    		"at line %d.\nExiting ...\n", __FILE__, __LINE__);

#ifdef MEMORY_DEBUG_LOG
		FILE *fout = fopen(MEMORY_DEBUG_LOG, "a");
		if(fout){
			// Every thing is ok!
			fprintf(fout, "Error! Can not allocate memory in file %s"
	    		"at line %d.\nExiting ...\n", __FILE__, __LINE__);
			fclose(fout);
		}
		else{
			// Woops! Something is wrong.
			fprintf(stderr,"Error! Couldn't open the memory log file '%s"
				 "'.\nExiting ...\n", MEMORY_DEBUG_LOG);
		}
#endif // MEMORY_DEBUG_LOG

		throw Mem_Error("Error! Can not allocate memory!\n",
			__FILE__, __LINE__);
		}
	
		cur = heap;
		cur->prev = NULL;
    }
    else{
	
		// Perform a sorted insertion into the heap tree.
		cur = heap;
		while( ((p <= cur->ptr) && (cur->left != NULL)) ||
	       	   ((p > cur->ptr) && (cur->right != NULL)) )
	    	if(p <= cur->ptr)
				cur = cur->left;
	    	else
				cur = cur->right;
		
	
		//  Allocate memory for the next heap tree element.
		if(p <= cur->ptr){
	    	cur->left = (struct heap_element*)
		        	    malloc(sizeof(struct heap_element));

	    	if(cur->left == NULL){
				fprintf(stderr, "Error! Can not allocate memory in file %s "
	    			"at line %d.\nExiting ...\n", __FILE__, __LINE__);
#ifdef MEMORY_DEBUG_LOG
				FILE *fout = fopen(MEMORY_DEBUG_LOG, "a");
				if(fout){
					// Every thing is ok!
					fprintf(fout, "Error! Can not allocate memory in file %s "
	    				"at line %d.\nExiting ...\n", __FILE__, __LINE__);
					fclose(fout);
				}
				else{
					// Woops! Something is wrong.
					fprintf(stderr,"Error! Couldn't open the memory log file '%s"
						"'.\nExiting ...\n", MEMORY_DEBUG_LOG);
				}
#endif // MEMORY_DEBUG_LOG

				throw Mem_Error("Error! Can not allocate memory!\n",
					__FILE__, __LINE__);
			}
	    
	    	cur->left->prev = cur;
	    	cur = cur->left;
		}
		else{
	    	cur->right = (struct heap_element*)
		        	 	 malloc(sizeof(struct heap_element));

	    	if(cur->right == NULL){
				fprintf(stderr, "Error! Can not allocate memory in file %s "
	    				"at line %d.\nExiting ...\n", __FILE__, __LINE__);

#ifdef MEMORY_DEBUG_LOG
				FILE *fout = fopen(MEMORY_DEBUG_LOG, "a");
				if(fout){
					// Every thing is ok!
					fprintf(fout, "Error! Can not allocate memory in file %s "
	    				"at line %d.\nExiting ...\n", __FILE__, __LINE__);
					fclose(fout);
				}
				else{
					// Woops! Something is wrong.
					fprintf(stderr,"Error! Couldn't open the memory log file '%s"
						"'.\nExiting ...\n", MEMORY_DEBUG_LOG);
			}
#endif // MEMORY_DEBUG_LOG

			throw Mem_Error("Error! Can not allocate memory!\n",
					__FILE__, __LINE__);
	    	}
	    
	    	cur->right->prev = cur;
	    	cur = cur->right;
		}
    }
    
    cur->left = NULL;
    cur->right = NULL;
    cur->ptr = p;
    cur->line = line;
    cur->size = size;
    cur->name = (char*)malloc(sizeof(char)*(strlen(filename) + 1));

    if(cur->name == NULL){
		fprintf(stderr, "Error! Can not allocate memory in file %s "
	    				"at line %d.\nExiting ...\n", __FILE__, __LINE__);
#ifdef MEMORY_DEBUG_LOG
		FILE *fout = fopen(MEMORY_DEBUG_LOG, "a");
		if(fout){
			// Every thing is ok!
			fprintf(fout, "Error! Can not allocate memory in file %s "
	    		"at line %d.\nExiting ...\n", __FILE__, __LINE__);
			fclose(fout);
		}
		else{
			// Woops! Something is wrong.
			fprintf(stderr,"Error! Couldn't open the memory log file %s."
				"\nExiting ...\n", MEMORY_DEBUG_LOG);
		}
#endif // MEMORY_DEBUG_LOG

		throw Mem_Error("Error! Can not allocate memory!\n",
			__FILE__, __LINE__);
    }
    
    strcpy(cur->name, filename);
	
    // Return the allocated memory
    return p;    
}

// Return memory deallocated by the user
void __memory_deallocate(void *ptr, char *filename, int line)
{
	// Make sure that deleting a NULL pointer doesn't
	// cause an error.
	if(ptr == NULL)
		return;

	// All done with memory reporting!
	if(DoneWithMemoryReporting || !StartedMemoryReporting){
		// System stuff may need to be deallocated after
		// the atexit() call, so just free it with out 
		// asking any questions.
		free(ptr);

		return;
	}

    struct heap_element *cur = heap, *temp;

    while((cur != NULL) && (ptr != cur->ptr)){
		if(ptr < cur->ptr)
			cur = cur->left;
		else
			cur = cur->right;
    }
    
    if(cur == NULL){
		fprintf(stderr, "Error in file %s at line %d.\n Attempting to "
			"deallocate memory not previously allocated.\n", filename, line);
		
#ifdef MEMORY_DEBUG_LOG
		FILE *fout = fopen(MEMORY_DEBUG_LOG, "a");
		if(fout){
			// Every thing is ok!
			fprintf(fout, "Error in file %s at line %d.\n Attempting to "
				"deallocate memory not previously allocated.\n", filename, line);
			fclose(fout);
		}
		else{
			// Woops! Something is wrong.
			fprintf(stderr,"Error! Couldn't open the memory log file '%s'.\n"
				, MEMORY_DEBUG_LOG);
		}
#endif // MEMORY_DEBUG_LOG

		// Hope that the user knows what they are doing!
		free(ptr);
		return;
    }

    if((cur->ptr == ptr) && (cur == heap)){
		// Delete the element from the head of the tree...
		if((cur->left == NULL) && (cur->right == NULL)){
			heap = NULL;
		}
		else if(cur->left == NULL){
			// Only the left branch is NULL
			heap = cur->right;
			heap->prev= NULL;
		}
		else if(cur->right == NULL){
			// Only the right branch is NULL
			heap = cur->left;
			heap->prev= NULL;
		}
		else{
			// Neither branch is NULL, so insert the left
			// branch on the left most position of the right branch.
			heap = cur->right;
			heap->prev= NULL;
			temp = heap;
	    
			while(temp->left != NULL)
				temp = temp->left;
	    
			temp->left = cur->left;
			temp->left->prev = temp;
		}
	
		free(cur->name);
		free(cur);
		cur = NULL;
    }
    else if(cur->ptr == ptr){
		// Delete the element from body of the tree...
		if((cur->left == NULL) && (cur->right == NULL)){
			// Both branches are NULL
			if(cur == cur->prev->left)
				cur->prev->left = NULL;
			else
				cur->prev->right = NULL;
		}
		else if(cur->left == NULL){
			// Only the left branch is NULL
			if(cur == cur->prev->left)
				cur->prev->left = cur->right;
			else
				cur->prev->right = cur->right;

			cur->right->prev = cur->prev;
		}
		else if(cur->right == NULL){
			// Only the right branch is NULL
			if(cur == cur->prev->left)
				cur->prev->left = cur->left;
			else
				cur->prev->right = cur->left;
			cur->left->prev = cur->prev;
		}
		else{ 
			// Neither branch is NULL, so insert the left
			// branch on the left most position of the right branch.
			temp = cur->right;
			if(cur == cur->prev->left)
				cur->prev->left = cur->right;
			else
				cur->prev->right = cur->right;
	    
			cur->right->prev = cur->prev;
	    
			while(temp->left != NULL)
				temp = temp->left;
	    
			temp->left = cur->left;
			temp->left->prev = temp;
		}
	
		free(cur->name);
		free(cur);
		cur = NULL;
    }
    else{
		// Panic!
		fprintf(stderr, "PANIC! Lost control in deallocate!\n");
#ifdef MEMORY_DEBUG_LOG
		FILE *fout = fopen(MEMORY_DEBUG_LOG, "a");
		if(fout){
			// Every thing is ok!
			fprintf(fout, "PANIC! Lost control in deallocate!\n");
			fclose(fout);
		}
		else{
			// Woops! Something is wrong.
			fprintf(stderr,"Error! Couldn't open the memory log file '%s"
				"'.\nExiting ...\n", MEMORY_DEBUG_LOG);
		}
#endif // MEMORY_DEBUG_LOG

		throw Mem_Error("PANIC! Lost control in deallocate!\n", 
			__FILE__, __LINE__);
    }
}

void __memory_freeheap(struct heap_element *cur)
{
	DoneWithMemoryReporting = 1;

    if(cur != NULL){
	if(cur->left != NULL)
	    __memory_freeheap(cur->left);
	    
	if(cur->right != NULL)
	    __memory_freeheap(cur->right);
	
	free(cur->name);
	
	// Free the memory allocated by the user.
	// Note that in C++, the destructor is NOT called
	// and memory is NOT automatically returned to the
	// system - only the warning about the memory leak
	// will be provided!
	//#ifndef __cplusplus
	//free(cur->ptr);
	//#endif
	
	free(cur);
	cur = NULL;
    }
}

void __memory_report()
{
	// Once we get to this function, we're all done!
	// Set this flag to turn off memory reporting for
	// any further calls to __memory_deallocate().
	DoneWithMemoryReporting = 1;

    long total;

    total = __memory_leak(heap);
    
    if(total != 0){
		fprintf(stderr, "Total Memory Leak = %d bytes.\n", total);
#ifdef MEMORY_DEBUG_LOG
		FILE *fout = fopen(MEMORY_DEBUG_LOG, "a");
		if(fout){
			// Every thing is ok!
			fprintf(fout, "Total Memory Leak = %d bytes.\n", total);
			fclose(fout);
		}
		else{
			// Woops! Something is wrong.
			fprintf(stderr,"Error! Couldn't open the memory log file '%s'.\n"
				, MEMORY_DEBUG_LOG);
		}
#endif // MEMORY_DEBUG_LOG
    }
    
	// Delete the heap.
    __memory_freeheap(heap);
}

long __memory_leak(struct heap_element *cur)
{
    long total = 0;
    
    if(cur != NULL){
		if(cur->left != NULL)
			total += __memory_leak(cur->left);
	    
		if(cur->right != NULL)
			total += __memory_leak(cur->right);
    
		// Print out some memory info
		fprintf(stderr,"Memory Leak in file %s at line %d "
			"(%d bytes). Address = %p\n", cur->name, cur->line,
			cur->size, cur->ptr);
#ifdef MEMORY_DEBUG_LOG
		FILE *fout = fopen(MEMORY_DEBUG_LOG, "a");
		if(fout){
			// Every thing is ok!
			fprintf(fout,"Memory Leak in file %s at line %d "
				"(%d bytes). Address = %p\n", cur->name, cur->line,
				cur->size, cur->ptr);
			fclose(fout);
		}
		else{
			// Woops! Something is wrong.
			fprintf(stderr,"Error! Couldn't open the memory log file '%s'.\n"
				, MEMORY_DEBUG_LOG);
		}
#endif // MEMORY_DEBUG_LOG

		total += cur->size;
    }

    return total;
}
