//	(c) 2000 J. Gans (jdg9@cornell.edu), Shalloway Lab, 
//	Cornell University. You can modify and freely distribute 
//	this code, but please keep this header intact.

#include "stdafx.h"
#include "sgraph.h"

#include "CommandLine.h"
#include "MainFrm.h"
#include "sgraphDoc.h"
#include "sgraphView.h"

#include "ScaleDialog.h"
#include "SetControlDialog.h"
#include "color.h"
#include "Excepts.h"
#include "parseOptions.h"
#include "textoutputdialog.h"
#include "plotoptionsdialog.h"
#include <math.h>
#include <float.h>

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

// ************************** Global variables **************************
// Are we running on a OS that only allows 16 bit screen coordinates 
// (i.e. 95 98)? If false, assume 32 bit screen coordiantes (i.e. NT and 2000).
BOOL is_16_bit = FALSE;
// ************************ End Global variables ************************

/////////////////////////////////////////////////////////////////////////////
// CSgraphApp
BEGIN_MESSAGE_MAP(CSgraphApp, CWinApp)
	//{{AFX_MSG_MAP(CSgraphApp)
	ON_COMMAND(ID_APP_ABOUT, OnAppAbout)
	ON_COMMAND(ID_APP_HELP_OPTION, OnAppHelpOption)
	ON_COMMAND(ID_APP_HELP_COMMAND, OnAppHelpCommand)
	//}}AFX_MSG_MAP
	// Standard file based document commands
	ON_COMMAND(ID_FILE_NEW, CWinApp::OnFileNew)
	ON_COMMAND(ID_FILE_OPEN, CWinApp::OnFileOpen)
	// Standard print setup command
	//ON_COMMAND(ID_FILE_PRINT_SETUP, CWinApp::OnFilePrintSetup)
	ON_COMMAND(ID_FILE_PRINT_SETUP, OnFilePrintSetup)
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CSgraphApp construction


CSgraphApp::CSgraphApp()
{
	// TODO: add construction code here,
	// Place all significant initialization in InitInstance
}

/////////////////////////////////////////////////////////////////////////////
// The one and only CSgraphApp object

CSgraphApp theApp;

/////////////////////////////////////////////////////////////////////////////
// CSgraphApp Print setup
void CSgraphApp::OnFilePrintSetup()
{
	CWinApp::OnFilePrintSetup();

	// Update the WYSIWYG setting in the view
	CSingleDocTemplate* pDocTemplate;
	POSITION loc = GetFirstDocTemplatePosition();

	pDocTemplate = (CSingleDocTemplate*)GetNextDocTemplate(loc);
	loc = pDocTemplate->GetFirstDocPosition();

	CSgraphDoc *pDoc = (CSgraphDoc *)(pDocTemplate->GetNextDoc(loc));
	loc = pDoc->GetFirstViewPosition();

	CSgraphView *pView = (CSgraphView *)(pDoc->GetNextView(loc));

	pView->PostMessage(WM_UPDATE_WYSIWYG);
}

/////////////////////////////////////////////////////////////////////////////
// CSgraphApp initialization

BOOL CSgraphApp::InitInstance()
{
	// Set the registry key under which our settings are stored.
	SetRegistryKey(_T("Shalloway Lab Applications"));

	LoadStdProfileSettings();  // Load standard INI file options (including MRU)
	
	// Initialize the Rich text control in the text output dialog
	AfxInitRichEdit();

	// Determine if we are running on an OS that only supports 16 bit screen coordinates
	OSVERSIONINFO os_info;

	os_info.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);

	if(GetVersionEx(&os_info)){
		// The following info is from the MS web site and is used to determine if
		// the OS support 32 or 16 bit screen coordinates.
		// Value						Platform
		// VER_PLATFORM_WIN32s			Win32s on Windows 3.1. 
		// VER_PLATFORM_WIN32_WINDOWS	Windows 95, Windows 98, or Windows Me. 
		// VER_PLATFORM_WIN32_NT		Windows NT 3.51, Windows NT 4.0,
		//								Windows 2000 and higher ...
		if((os_info.dwPlatformId == VER_PLATFORM_WIN32s) ||
		   (os_info.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS)){
			is_16_bit = TRUE;
		}
	}
 
	// Register the application's document templates.  Document templates
	// serve as the connection between documents, frame windows and views.
	CSingleDocTemplate* pDocTemplate;
	pDocTemplate = new CSingleDocTemplate(
		IDR_MAINFRAME,
		RUNTIME_CLASS(CSgraphDoc),
		RUNTIME_CLASS(CMainFrame),       // main SDI frame window
		RUNTIME_CLASS(CSgraphView));
	AddDocTemplate(pDocTemplate);

	// Parse command line for standard shell commands, DDE, file open
	CSGraphCommandLine cmdInfo;
	try{
		ParseCommandLine(cmdInfo);
	}
	catch(Error &e){
		AfxMessageBox(e.message());
	}

	// Dispatch commands specified on the command line
	if (!ProcessShellCommand(cmdInfo))
		return FALSE;
	
	// Get a pointer to both the Document and View objects
	POSITION loc = pDocTemplate->GetFirstDocPosition();
	CSgraphDoc *pDoc = (CSgraphDoc *)(pDocTemplate->GetNextDoc(loc));
	loc = pDoc->GetFirstViewPosition();
	CSgraphView *pView = (CSgraphView *)(pDoc->GetNextView(loc));

	// Do we need to import and data?
	if(cmdInfo.ImportFile()){
		// Import all files listed on the command line (wildcards are OK).
		for(int i = 0;i < cmdInfo.m_import_filename.Length();i++)
			pDoc->FileImportXY(cmdInfo.m_import_filename[i], cmdInfo.GetImportFileOption());
	

		// Set the axis scaling (only if the user is importing a data
		// file -- do not change the axes when reading in a PrestoPlot
		// file).
		if(cmdInfo.m_logX || cmdInfo.m_logY){
			int tmpXaxis = SCALE_LINEAR;
			int tmpYaxis = SCALE_LINEAR;
			
			if(cmdInfo.m_logX)
				tmpXaxis = SCALE_LOG;
			
			if(cmdInfo.m_logY)
				tmpYaxis = SCALE_LOG;
			
			pDoc->graph[pDoc->selected_graph].SetGraphType(tmpXaxis, tmpYaxis);
		}
	
	}
	// The one and only window has been initialized, so show and update it.
	m_pMainWnd->ShowWindow(SW_SHOW);
	m_pMainWnd->UpdateWindow();

	// If PrestoPlot is run in the background and totally obscured by other 
	// windows, the screen will not be updated! To fix this, we need to force
	// a screen redraw. The call to ForceRedraw() does not actually force a 
	// redraw, but it does set a private variable in pView (i.e. flushBuffer)
	// that sets the stage for RedrawWindow(...) to force a redraw of the
	// view. Whew!
	pView->ForceRedraw();
	pView->RedrawWindow(NULL, NULL, RDW_INVALIDATE | RDW_ERASENOW | RDW_ERASE);
	
	// Do we need to display the command line options?
	if(cmdInfo.m_displayCommandLine)
		OnAppHelpCommand();
		
	if(cmdInfo.m_pipe){
		pDoc->SetPipe(cmdInfo.m_pipe_addr);
		pView->PostMessage(WM_START_PIPE);
	}

	return TRUE;
}


/////////////////////////////////////////////////////////////////////////////
// CAboutDlg dialog used for App About

class CAboutDlg : public CDialog
{
public:
	CAboutDlg();

// Dialog Data
	//{{AFX_DATA(CAboutDlg)
	enum { IDD = IDD_ABOUTBOX };
	CString	m_version;
	CString	m_time_stamp;
	CString	m_os_info;
	CString	m_info;
	//}}AFX_DATA

	// ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(CAboutDlg)
	protected:
	virtual void DoDataExchange(CDataExchange* pDX);    // DDX/DDV support
	//}}AFX_VIRTUAL

// Implementation
protected:
	//{{AFX_MSG(CAboutDlg)
		// No message handlers
	//}}AFX_MSG
	DECLARE_MESSAGE_MAP()
};

CAboutDlg::CAboutDlg() : CDialog(CAboutDlg::IDD)
{
	//{{AFX_DATA_INIT(CAboutDlg)
	m_version = _T("");
	m_time_stamp = _T("");
	m_os_info = _T("");
	m_info = _T("");
	//}}AFX_DATA_INIT

	m_version.Format("PrestoPlot version %.2f", SGRAPH_VERSION);
	m_time_stamp.Format("Build date: %s", __TIMESTAMP__);

	if(is_16_bit){
		m_os_info.Format("Using 16 bit screen coordinates");
	}
	else{
		m_os_info.Format("Using 32 bit screen coordinates");
	}

	// Give credit where credit is due ...
	m_info.Format("J.D. Gans\r\n"
		"Shalloway Lab\r\n"
		"Cornell University\r\n"
		"http://www.mbg.cornell.edu/shalloway/shalloway.html\r\n"
		"Copyright (C) 2000-2002\r\n"
		"\r\n"
		"Template Numerical Toolkit (TNT)\r\n"
		"Roldan Pozo\r\n"
		"Mathematical and Computational Sciences Division\r\n"
		"National Institute of Standards and Technology\r\n"
		"http://math.nist.gov/tnt/\r\n"
		"\r\n"
		"PrestoPlot was inspired by Grace/Xmgr\r\n"
		"Paul J Turner and the Grace Development Team\r\n"
		"http://plasma-gate.weizmann.ac.il/Grace/\r\n"
		"\r\n"
		"Data decompression performed by zlib\r\n"
		"Mark Adler and Jean-loup Gailly\r\n"
		"http://www.gzip.org/zlib/\r\n"
		"\r\n"
		"PNG graphics by libpng\r\n"
		"http://www.libpng.org/pub/png/\r\n");
}

void CAboutDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CAboutDlg)
	DDX_Text(pDX, IDC_STATIC_VERSION, m_version);
	DDX_Text(pDX, IDC_TIME_STAMP, m_time_stamp);
	DDX_Text(pDX, IDC_OS_INFO, m_os_info);
	DDX_Text(pDX, IDC_INFO_EDIT, m_info);
	//}}AFX_DATA_MAP
}

BEGIN_MESSAGE_MAP(CAboutDlg, CDialog)
	//{{AFX_MSG_MAP(CAboutDlg)
		// No message handlers
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

// App command to run the dialog
void CSgraphApp::OnAppAbout()
{
	CAboutDlg aboutDlg;
	aboutDlg.DoModal();
}

/////////////////////////////////////////////////////////////////////////////
//	*** SPlotIterator ***
/////////////////////////////////////////////////////////////////////////////
SPlotIterator::SPlotIterator(SGraph &m_graph, int m_access) :
	ListIterator<SPlot>(m_graph.plot)
{	
	access_mode = m_access;
}

SPlot* SPlotIterator::operator()()
{
	SPlot *lpPlot;

	switch(access_mode){
		case ALL:
			return ListIterator<SPlot>::operator()();
			break;
		case DRAWN:
			while(lpPlot = ListIterator<SPlot>::operator()()){
				if(lpPlot->isDrawn == TRUE)
					return lpPlot;
			}
			break;
		case NOT_DRAWN:
			while(lpPlot = ListIterator<SPlot>::operator()()){
				if(lpPlot->isDrawn == FALSE)
					return lpPlot;
			}
			break;
		default:
			throw Error("Illegal access mode in SPlotIterator");
			break;
	};

	return NULL;
}

SPlot* SPlotIterator::operator()(int &m_index)
{
	SPlot *lpPlot;

	switch(access_mode){
		case ALL:
			return ListIterator<SPlot>::operator()(m_index);
			break;
		case DRAWN:
			while(lpPlot = ListIterator<SPlot>::operator()(m_index)){
				if(lpPlot->isDrawn == TRUE)
					return lpPlot;
			}
			break;
		case NOT_DRAWN:
			while(lpPlot = ListIterator<SPlot>::operator()(m_index)){
				if(lpPlot->isDrawn == FALSE)
					return lpPlot;
			}
			break;
		default:
			throw Error("Illegal access mode in SPlotIterator");
			break;
	};

	return NULL;
}

/////////////////////////////////////////////////////////////////////////////
//	*** SGraph ***
/////////////////////////////////////////////////////////////////////////////
SGraph::SGraph() : 
title(PRESTO_DEFAULT_TITLE, 0.5f, DefaultStartY + DefaultHeight + 0.05f),
leftAxisLable(PRESTO_DEFAULT_Y_AXIS_LABLE, DefaultStartX - 0.1f, 0.5f, 900),
bottomAxisLable(PRESTO_DEFAULT_X_AXIS_LABLE, 0.5f, DefaultStartY - 0.1f),
leftAxis(DefaultStartX, 0.5f, DefaultHeight, 
		 SAxisObject::VERTICAL, 1.0f, 10.0f, SAxisObject::LEFT,
		 SAxisObject::LEFT),
rightAxis(DefaultStartX + DefaultWidth, 0.5f, DefaultHeight, 
		  SAxisObject::VERTICAL, 1.0f, 10.0f, SAxisObject::RIGHT,
		  SAxisObject::RIGHT),
topAxis(0.5f, DefaultStartY + DefaultHeight, DefaultWidth, 
		SAxisObject::HORIZONTAL, 1.0f, 10.0f, SAxisObject::UP,
		SAxisObject::UP),
bottomAxis(0.5f, DefaultStartX, DefaultWidth, 
		   SAxisObject::HORIZONTAL, 1.0f, 10.0f, SAxisObject::DOWN,
		   SAxisObject::DOWN),
legend(plot)
{
	x_scaleType = SCALE_LINEAR;
	y_scaleType = SCALE_LINEAR;
	
	x_tick_location = TICK_MARK_NORMAL;
	y_tick_location = TICK_MARK_NORMAL;
	
	x_tick_label_location = TICK_MARK_NORMAL;
	y_tick_label_location = TICK_MARK_NORMAL;

	// Set the bounding box
	x1 = DefaultStartX;
	y1 = DefaultStartY;
	x2 = DefaultStartX + DefaultWidth;
	y2 = DefaultStartY + DefaultHeight;
	
	// The center of the graph
	x = 0.5f*(x2 - x1) + x1;
	y = 0.5f*(y2 - y1) + y1;

	graph_move = GRAPH_MOVE_ALL;

	// By default, the background color is white
	background_color = 0x00ffffff;

	legend.x = x;
	legend.y = y;

	leftAxis.x = x1;
	leftAxis.y = y;

	rightAxis.x = x2;
	rightAxis.y = y;
	
	topAxis.x = x;
	topAxis.y = y2;
	
	bottomAxis.x = x;
	bottomAxis.y = y1;

	// Which objects need to be displayed?
	title.isDrawn = TRUE;
	
	// Set the default size of the graph title and axis labels.
	title.SetFontSize(-DefaultGraphTitleFontSize);
	leftAxisLable.SetFontSize(-DefaultAxisLabelFontSize);
	bottomAxisLable.SetFontSize(-DefaultAxisLabelFontSize);

	leftAxisLable.isDrawn = TRUE;
	bottomAxisLable.isDrawn = TRUE;

	leftAxis.isDrawn = TRUE;
	leftAxis.isLineDrawn = TRUE;
	leftAxis.isTickDrawn = TRUE;
	leftAxis.isTickLableDrawn = TRUE;
	
	rightAxis.isDrawn = TRUE;
	rightAxis.isLineDrawn = TRUE;
	rightAxis.isTickDrawn = FALSE;
	rightAxis.isTickLableDrawn = FALSE;

	topAxis.isDrawn = TRUE;
	topAxis.isLineDrawn = TRUE;
	topAxis.isTickDrawn = FALSE;
	topAxis.isTickLableDrawn = FALSE;

	bottomAxis.isDrawn = TRUE;
	bottomAxis.isLineDrawn = TRUE;
	bottomAxis.isTickDrawn = TRUE;
	bottomAxis.isTickLableDrawn = TRUE;

	isStickyScalingX = FALSE;
	isStickyScalingY = FALSE;

	minX = 1.0f;
	maxX = 10.0f;
	minY = 1.0f;
	maxY = 10.0f;

	defaultMinX = 1.0f;
	defaultMaxX = 10.0f;
	defaultMinY = 1.0f;
	defaultMaxY = 10.0f;

	animate_plots = FALSE;
	curr_animate_plot = -1;

	// Don't display overlay equation
	plot_quick_eq = FALSE;
}

void SGraph::PreparePrinting(CDC* pDC, long scaleMul, long scaleDiv, CRect &extent)
{
	SObject::PreparePrinting(pDC, scaleMul, scaleDiv, extent);

	title.PreparePrinting(pDC, scaleMul, scaleDiv, extent);
	leftAxisLable.PreparePrinting(pDC, scaleMul, scaleDiv, extent);
	bottomAxisLable.PreparePrinting(pDC, scaleMul, scaleDiv, extent);

	leftAxis.PreparePrinting(pDC, scaleMul, scaleDiv, extent);
	rightAxis.PreparePrinting(pDC, scaleMul, scaleDiv, extent);
	topAxis.PreparePrinting(pDC, scaleMul, scaleDiv, extent);
	bottomAxis.PreparePrinting(pDC, scaleMul, scaleDiv, extent);

	legend.PreparePrinting(pDC, scaleMul, scaleDiv, extent);

	for(int i = 0;i < plot.Length();i++)
		plot[i].PreparePrinting(pDC, scaleMul, scaleDiv, extent);
}

void SGraph::CleanUpPrinting(CDC* pDC, CRect &extent)
{
	title.CleanUpPrinting(pDC, extent);
	leftAxisLable.CleanUpPrinting(pDC, extent);
	bottomAxisLable.CleanUpPrinting(pDC, extent);

	leftAxis.CleanUpPrinting(pDC, extent);
	rightAxis.CleanUpPrinting(pDC, extent);
	topAxis.CleanUpPrinting(pDC, extent);
	bottomAxis.CleanUpPrinting(pDC, extent);

	legend.CleanUpPrinting(pDC, extent);

	for(int i = 0;i < plot.Length();i++)
		plot[i].CleanUpPrinting(pDC, extent);

	SObject::CleanUpPrinting(pDC, extent);
}

void SGraph::Printing(BOOL p)
{
	SObject::Printing(p);

	title.Printing(p);
	leftAxisLable.Printing(p);
	bottomAxisLable.Printing(p);

	leftAxis.Printing(p);
	rightAxis.Printing(p);
	topAxis.Printing(p);
	bottomAxis.Printing(p);

	legend.Printing(p);

	for(int i = 0;i < plot.Length();i++)
		plot[i].Printing(p);
}

void SGraph::Draw(CDC* pDC, CRect &extent)
{	
	if(!isDrawn)
		return;

	// *Draw the plot data*
	// But first set the clipping region to be the bounding box
	CRgn bbox;
	CRect bbox_rect(XtoScreen(x1, extent), 
					YtoScreen(y1, extent), 
					XtoScreen(x2, extent), 
					YtoScreen(y2, extent));
	
	// Draw the background region (unless the user has selected
	// "no color" as the background color). Note that we must draw
	// the graph background before bbox_rect is offset (when in
	// print preview mode).
	if(background_color != NO_COLOR){
		CBrush bkgrd_brush(background_color);
		pDC->FillRect(bbox_rect, &bkgrd_brush);
	}

	// Messy kludge (from the MS help files) to get the 
	// clipping region to work with the print preview.
	if(pDC->IsKindOf(RUNTIME_CLASS(CPreviewDC))){
		CPreviewDC *pPrevDC = (CPreviewDC *)pDC;

		pPrevDC->PrinterDPtoScreenDP(&bbox_rect.TopLeft());
        pPrevDC->PrinterDPtoScreenDP(&bbox_rect.BottomRight());

		// Now offset the result by the viewport origin of
        // the print preview window...        
		CPoint ptOrg;

		::GetViewportOrgEx(pDC->m_hDC,&ptOrg);
		bbox_rect += ptOrg;
	}

	bbox.CreateRectRgn(bbox_rect.TopLeft().x,
					   bbox_rect.TopLeft().y,
					   bbox_rect.BottomRight().x,
					   bbox_rect.BottomRight().y);

	pDC->SelectClipRgn(&bbox);

	// Draw grids (if any) BEFORE plotting any data.
	leftAxis.DrawGrid(pDC, extent, x2 - x1);
	bottomAxis.DrawGrid(pDC, extent, y2 - y1);

	if(animate_plots){
		// Make sure the user hasn't deleted plots and mucked up the
		// status of the curr_animate_plot.
		if((curr_animate_plot > -1) && (curr_animate_plot < plot.Length()))
			plot[curr_animate_plot].Draw(pDC, extent);
	}
	else{
		for(int i = 0;i < plot.Length();i++)
			plot[i].Draw(pDC, extent);
	}

	// Draw any user defined lines
	DrawLines(pDC, extent);

	if(plot_quick_eq)
		DrawQuickPlot(pDC, extent);

	// Restore the full window clipping region
	pDC->SelectClipRgn(NULL);
	
	// Draw everything else
	title.Draw(pDC, extent);
	leftAxisLable.Draw(pDC, extent);
	bottomAxisLable.Draw(pDC, extent);
	leftAxis.Draw(pDC, extent);
	rightAxis.Draw(pDC, extent);
	topAxis.Draw(pDC, extent);
	bottomAxis.Draw(pDC, extent);	

	legend.Draw(pDC, extent);

	// Draw selection boxes
	if(isSelected && !isPrinting)
		DrawSelect(pDC, extent);
}

void SGraph::DrawSelect(CDC* pDC, CRect &extent)
{
	const int focus = 3;
	
	// Select ROP2 mode
	int ropOld = pDC->SetROP2(R2_NOT);

	CBrush selectbrush(COLORREF (0) /* Black */);
	CBrush *oldbrush = pDC->SelectObject(&selectbrush); 

	pDC->Rectangle(XtoScreen(x1, extent) - focus, 
				   YtoScreen(y1, extent) - focus,
				   XtoScreen(x1, extent) + focus, 
				   YtoScreen(y1, extent) + focus);

	pDC->Rectangle(XtoScreen(x2, extent) - focus, 
				   YtoScreen(y1, extent) - focus,
				   XtoScreen(x2, extent) + focus, 
				   YtoScreen(y1, extent) + focus);

	pDC->Rectangle(XtoScreen(x1, extent) - focus, 
				   YtoScreen(y2, extent) - focus,
				   XtoScreen(x1, extent) + focus, 
				   YtoScreen(y2, extent) + focus);

	pDC->Rectangle(XtoScreen(x2, extent) - focus, 
				   YtoScreen(y2, extent) - focus,
				   XtoScreen(x2, extent) + focus, 
				   YtoScreen(y2, extent) + focus);

	// Restore the previous brush
	pDC->SelectObject(oldbrush);
	pDC->SetROP2(ropOld);
}

void SGraph::DrawLines(CDC* pDC, CRect &extent)
{
	// Draw user defined lines
	for(int j = 0;j < lines.Length();j++){
		SLineObject &tmpLine = lines[j];

		int pen_style = PS_NULL;

		switch(tmpLine.GetLineType()){
			case SLineObject::LINE_SOLID:
				pen_style = PS_SOLID;
				break;
			case SLineObject::LINE_DOTTED:
				pen_style = PS_DOT;
				break;
			case SLineObject::LINE_DASHED:
				pen_style = PS_DASH;
				break;
			case SLineObject::LINE_DASHDOT:
				pen_style = PS_DASHDOT;
				break;
			case SLineObject::LINE_DASHDOTDOT:
				pen_style = PS_DASHDOTDOT;
				break;
		};

		// Draw with a 1 pixel black line
		//CPen linepen(pen_style, 1*lineScale, RGB(0, 0, 0));
		CPen *linepen = NULL;
		DWORD *dashBuffer = NULL;

		if((pen_style != SLineObject::LINE_SOLID) && (lineScale != 1) && (!is_16_bit)){
			// Create the pen for drawing the lines. A few hoops must be jumped 
			// through in order to render thick, non-solid lines.
			LOGBRUSH penBrush;

			penBrush.lbStyle = BS_SOLID;
			penBrush.lbColor = RGB(0, 0, 0);

			int dashSize;
			dashBuffer = tmpLine.GetLineStyle(dashSize, isPrinting, printScaleMul, printScaleDiv);

			// Use the extented form of the CPen constructor to enable the use
			// of the thick dashed lines.
			linepen = new CPen(PS_GEOMETRIC | PS_ENDCAP_SQUARE | PS_JOIN_BEVEL | PS_USERSTYLE, 
				lineScale, &penBrush, dashSize, dashBuffer);
		}
		else{
			// Create the pen for drawing the lines. Use a plain old CPen for this
			// task (as it is quite fast!).
			linepen = new CPen(pen_style, lineScale, RGB(0, 0, 0));
		}

		// Make the current pen the state pen
		CPen *oldpen = pDC->SelectObject(linepen); 

		// Draw the line
		if(tmpLine.GetVarType() == SLineObject::X){
			// A vertical (constant X) line
			int line_x;
			BOOL bounds_error = FALSE;

			if(x_scaleType == SCALE_LINEAR){
				line_x = XtoScreen((tmpLine.GetValue() - minX)*(x2 - x1)/(maxX - minX) + x1, 
					extent);
			}
			else{ // x_scaleType == SCALE_LOG
				// Make sure that we don't take the log of a negative number!
				if(tmpLine.GetValue() > 0.0f)
					line_x = XtoScreen(((float)(log10(tmpLine.GetValue())) - minX)*
						(x2 - x1)/(maxX - minX) + x1, extent);
				else
					bounds_error = TRUE;
			}
	
			if(!bounds_error){
				pDC->MoveTo(line_x, YtoScreen(y1, extent)); 
				pDC->LineTo(line_x, YtoScreen(y2, extent)); 
			}
		}
		else{ // var_type == Y
			// A horizontal (constant Y) line
			int line_y;
			BOOL bounds_error = FALSE;

			if(y_scaleType == SCALE_LINEAR){
				line_y = YtoScreen((tmpLine.GetValue() - minY)*(y2 - y1)/(maxY - minY) + y1, 
					extent);
			}
			else{ // x_scaleType == SCALE_LOG
				// Make sure that we don't take the log of a negative number!
				if(tmpLine.GetValue() > 0.0f)
					line_y = YtoScreen(((float)(log10(tmpLine.GetValue())) - minY)*
						(y2 - y1)/(maxY - minY) + y1, extent);
				else
					bounds_error = TRUE;
			}

			if(!bounds_error){
				pDC->MoveTo(XtoScreen(x1, extent), line_y);
				pDC->LineTo(XtoScreen(x2, extent), line_y);
			}
		}

		pDC->SelectObject(oldpen);

		if(linepen)
			delete linepen;

		if(dashBuffer)
			delete [] dashBuffer;
	}
}

float SGraph::ScreentoDataX(int m_x, CRect &extent)
{
	if(x_scaleType == SCALE_LINEAR)
		return (ScreentoX(m_x, extent) - x1)*(maxX - minX)/(x2 - x1) 
			+ minX;
	
	// x_scaleType == SCALE_LOG
	return (float)(pow(10.0, (ScreentoX(m_x, extent) - x1)*(maxX - minX)/(x2 - x1) 
			+ minX));
}

float SGraph::ScreentoDataY(int m_y, CRect &extent)
{
	if(y_scaleType == SCALE_LINEAR)
		return (ScreentoY(m_y, extent) - y1)*(maxY - minY)/(y2 - y1) 
			+ minY;
	
	// y_scaleType == SCALE_LOG
	return (float)(pow(10.0, (ScreentoY(m_y, extent) - y1)*(maxY - minY)/(y2 - y1) 
			+ minY));
}

void SGraph::GetPlotRange(float &m_minX, float &m_maxX, 
						  float &m_minY, float &m_maxY)
{
	if(y_scaleType == SCALE_LINEAR){
		m_maxY = maxY;
		m_minY = minY;
	}
	else{ // y_scaleType == SCALE_LOG
		m_maxY = (float)(pow(10.0, maxY));
		m_minY = (float)(pow(10.0, minY));
	}
	
	if(x_scaleType == SCALE_LINEAR){
		m_maxX = maxX;
		m_minX = minX;
	}
	else{ // x_scaleType == SCALE_LOG
		m_maxX = (float)(pow(10.0, maxX));
		m_minX = (float)(pow(10.0, minX));
	}
}

// Set the plot scale types
void SGraph::SetPlotScaleType()
{
	int i;

	if(x_scaleType == SCALE_LOG){
		for(i = 0;i < plot.Length();i++){
			if(plot[i].isDrawn){
				plot[i].x_scaleType = SCALE_LOG;

				// If we have even a single active plot that
				// cannot support a log scale, switch to 
				// the linear scale.
				if(!plot[i].canXLog){
					x_scaleType = SCALE_LINEAR;
					SetPlotScaleType();
					return;
				}
			}
		}
	}

	if(x_scaleType == SCALE_LINEAR){
		for(i = 0;i < plot.Length();i++){
			plot[i].x_scaleType = SCALE_LINEAR;
		}
	}

	if(y_scaleType == SCALE_LOG){
		for(i = 0;i < plot.Length();i++){
			if(plot[i].isDrawn){
				plot[i].y_scaleType = SCALE_LOG;
				// If we have even a single active plot that
				// cannot support a log scale, switch to 
				// the linear scale.
				if(!plot[i].canYLog){
					y_scaleType = SCALE_LINEAR;
					SetPlotScaleType();
					return;
				}
			}
		}
	}

	if(y_scaleType == SCALE_LINEAR){
		for(i = 0;i < plot.Length();i++){
			plot[i].y_scaleType = SCALE_LINEAR;
		}
	}
}

void SGraph::SetPlotRange(BOOL force_sticky /* = FALSE */)
{
	SetPlotRangeX(force_sticky);
	SetPlotRangeY(force_sticky);
}

void SGraph::SetDefaultPlotRange()
{
	SetDefaultPlotRangeX();
	SetDefaultPlotRangeY();
}

void SGraph::SetPlotRangeX(BOOL force_sticky /* = FALSE */)
{
	if(isStickyScalingX && !force_sticky){
		// Return to the user defined default scaling
		SetPlotRangeX(defaultMinX, defaultMaxX, TRUE);
		return;
	}

	float range = -1.0f;
	int i;
	
	// Default values - hard coded; bad bad bad bad ...
	minX = 1.0f;
	maxX = 10.0f;
	
	// Throw away the default values if we have plots that
	// have data and are not disabled.
	for(i = 0;i < plot.Length();i++)
		if(plot[i].isDrawn){
			if(plot[i].UpdateBoundsX()){
				if(range < 0.0f){
					minX = plot[i].minX;
					maxX = plot[i].maxX;
				}
				else{
					minX = min(minX, plot[i].minX);
					maxX = max(maxX, plot[i].maxX);
				}

				range = 1.0f;
			}
		}

	// Streach the plot range to make the graph look nice
	if((range > 0.0f) && (x_scaleType == SCALE_LINEAR)){
		range = maxX - minX;
		
		// Handle the case when maxX == minX
		// (which would other wise case divide by zero 
		// errors).
		if(range == 0.0)
			range = 9.0;

		maxX += 0.05f*range;
		minX -= 0.05f*range;
	}

	if(x_scaleType == SCALE_LOG){
		minX = (float)(floor(log10(minX)));
		maxX = (float)(ceil(log10(maxX)));

		// Adjust the number of tick marks for the log plot
		topAxis.MajorTickSpace(1.0f, TRUE);
		bottomAxis.MajorTickSpace(1.0f, TRUE);
		topAxis.MinorTickSpace(1.0f, TRUE);
		bottomAxis.MinorTickSpace(1.0f, TRUE);
	}
	else{
		// Adjust the number of tick marks for the linear plot
		topAxis.MajorTickSpace((maxX - minX)/NUM_MAJOR_TICK);
		bottomAxis.MajorTickSpace(topAxis.MajorTickSpace());

		topAxis.MinorTickSpace((maxX - minX)/NUM_MINOR_TICK);
		bottomAxis.MinorTickSpace(topAxis.MinorTickSpace());
	}
		
	// Update the axis tick limits
	topAxis.tick_min = bottomAxis.tick_min = minX;
	topAxis.tick_max = bottomAxis.tick_max = maxX;

	// Update the axis scale type
	topAxis.scaleType = bottomAxis.scaleType = x_scaleType;

	// Set all of the drawn graphs to have the same plot range
	for(i = 0;i < plot.Length();i++)
		if(plot[i].isDrawn){
			plot[i].minX = minX;
			plot[i].maxX = maxX;
		}
}

void SGraph::SetDefaultPlotRangeX()
{
	isStickyScalingX = FALSE;
}

void SGraph::SetPlotRangeY(BOOL force_sticky /* = FALSE */)
{
	if(isStickyScalingY && !force_sticky){
		// Return to the user defined default scaling
		SetPlotRangeY(defaultMinY, defaultMaxY, TRUE);
		return;
	}

	float range = -1.0f;
	int i;

	// Default values - hard coded; bad bad bad bad ...
	minY = 1.0f;
	maxY = 10.0f;
	
	// Throw away the default values if we have plots that
	// have data and are not disabled.
	for(i = 0;i < plot.Length();i++)
		if(plot[i].isDrawn){
			if(plot[i].UpdateBoundsY()){
				if(range < 0.0f){
					minY = plot[i].minY;
					maxY = plot[i].maxY;
				}
				else{
					minY = min(minY, plot[i].minY);
					maxY = max(maxY, plot[i].maxY);
				}

				range = 1.0f;
			}
		}

	// Streach the plot range to make the graph look nice
	if((range > 0.0f) && (y_scaleType == SCALE_LINEAR)){
		range = maxY - minY;
		
		// Handle the case when maxX == minX
		// (which would other wise case divide by zero 
		// errors).
		if(range == 0.0)
			range = 9.0;

		maxY += 0.05f*range;
		minY -= 0.05f*range;
	}
	
	if(y_scaleType == SCALE_LOG){
		minY = (float)(floor(log10(minY)));
		maxY = (float)(ceil(log10(maxY)));

		// Adjust the number of tick marks for the log plot
		leftAxis.MajorTickSpace(1.0f, TRUE);
		rightAxis.MajorTickSpace(1.0f, TRUE);
		leftAxis.MinorTickSpace(1.0f, TRUE);
		rightAxis.MinorTickSpace(1.0f, TRUE);
	}
	else{
		// Adjust the number of tick marks for the linear plot
		leftAxis.MajorTickSpace((maxY - minY)/NUM_MAJOR_TICK);
		rightAxis.MajorTickSpace(leftAxis.MajorTickSpace());

		leftAxis.MinorTickSpace((maxY - minY)/NUM_MINOR_TICK);
		rightAxis.MinorTickSpace(leftAxis.MinorTickSpace());
	}
	
	// Update the axis tick limits
	leftAxis.tick_min = rightAxis.tick_min = minY;
	leftAxis.tick_max = rightAxis.tick_max = maxY;

	// Update the axis scale type
	leftAxis.scaleType = rightAxis.scaleType = y_scaleType;

	// Set all of the drawn graphs to have the same plot range
	for(i = 0;i < plot.Length();i++)
		if(plot[i].isDrawn){
			plot[i].minY = minY;
			plot[i].maxY = maxY;
		}
}

void SGraph::SetDefaultPlotRangeY()
{
	isStickyScalingY = FALSE;
}

void SGraph::SetPlotRange(float m_minX, float m_maxX, 
						  float m_minY, float m_maxY,
						  BOOL force_sticky /* = FALSE */)
{
	SetPlotRangeX(m_minX, m_maxX, force_sticky);
	SetPlotRangeY(m_minY, m_maxY, force_sticky);
}

void SGraph::SetDefaultPlotRange(float m_minX, float m_maxX, 
								 float m_minY, float m_maxY)
{
	SetDefaultPlotRangeX(m_minX, m_maxX);
	SetDefaultPlotRangeY(m_minY, m_maxY);
}

void SGraph::SetPlotRangeX(float m_minX, float m_maxX, 
						   BOOL force_sticky /* = FALSE */)
{
	if(isStickyScalingX && !force_sticky){
		// Return to the user defined default scaling
		SetPlotRangeX(defaultMinX, defaultMaxX, TRUE);
		return;
	}

	if(m_minX == m_maxX){
		AfxMessageBox("X coordinate range is too small! Restoring default XY range.");
		SetPlotRange();
		return;
	}

	if(x_scaleType == SCALE_LINEAR){
		minX = m_minX;
		maxX = m_maxX;
	}
	else{ // x_scaleType == SCALE_LOG
		// Make sure we don't take the log of a negative number
		if((m_minX <= 0.0f) || (m_maxX <= 0.0f)){
			AfxMessageBox("Can't have a negative range with a log scale!");
			return;
		}

		float exponent = (float)(log10(m_minX));

		if(fabs((int)(exponent) - exponent) <= FP_PRECISION){
			minX = exponent;
		}
		else{
			minX = (float)(floor(exponent));
		}

		exponent = (float)(log10(m_maxX));

		if(fabs((int)(exponent) - exponent) <= FP_PRECISION){
			maxX = (float)(exponent);
		}
		else{
			maxX = (float)(ceil(exponent));
		}
	}

	// Set the bounds for each plot
	for(int i = 0;i < plot.Length();i++){
		plot[i].minX = minX;
		plot[i].maxX = maxX;
	}

	// Update the axis tick limits
	topAxis.tick_min = bottomAxis.tick_min = minX;
	topAxis.tick_max = bottomAxis.tick_max = maxX;
	
	// Update the axis scale type
	topAxis.scaleType = bottomAxis.scaleType = x_scaleType;

	if(x_scaleType == SCALE_LOG){
		// Auto scale tick space for log graphs for now
		topAxis.MajorTickSpace(1.0f, TRUE);
		bottomAxis.MajorTickSpace(1.0f, TRUE);
		topAxis.MinorTickSpace(1.0f, TRUE);
		bottomAxis.MinorTickSpace(1.0f, TRUE);
	}
	else{ // x_scaleType == SCALE_LINEAR
		// Adjust the number of tick marks for the linear plot
		topAxis.MajorTickSpace((maxX - minX)/NUM_MAJOR_TICK);
		bottomAxis.MajorTickSpace(topAxis.MajorTickSpace());

		topAxis.MinorTickSpace((maxX - minX)/NUM_MINOR_TICK);
		bottomAxis.MinorTickSpace(topAxis.MinorTickSpace());
	}
}

void SGraph::SetDefaultPlotRangeX(float m_minX, float m_maxX)
{
	isStickyScalingX = TRUE;

	defaultMinX = m_minX;
	defaultMaxX = m_maxX;
}

void SGraph::SetPlotRangeY(float m_minY, float m_maxY,
						   BOOL force_sticky /* = FALSE */)
{
	if(isStickyScalingY && !force_sticky){
		// Return to the user defined default scaling
		SetPlotRangeY(defaultMinY, defaultMaxY, TRUE);
		return;
	}

	if(m_minY == m_maxY){
		AfxMessageBox("Y coordinate range is too small! Restoring default XY range.");
		SetPlotRange();
		return;
	}
	
	if(y_scaleType == SCALE_LINEAR){
		minY = m_minY;
		maxY = m_maxY;
	}
	else{ // y_scaleType == SCALE_LOG
		// Make sure we don't take the log of a negative number
		if((m_minY <= 0.0f) || (m_maxY <= 0.0f)){
			AfxMessageBox("Can't have a negative range with a log scale!");
			return;
		}

		float exponent = (float)(log10(m_minY));

		if(fabs((int)(exponent) - exponent) <= FP_PRECISION){
			minY = exponent;
		}
		else{
			minY = (float)(floor(exponent));
		}

		exponent = (float)(log10(m_maxY));

		if(fabs((int)(exponent) - exponent) <= FP_PRECISION){
			maxY = (float)(exponent);
		}
		else{
			maxY = (float)(ceil(exponent));
		}
	}

	// Set the bounds for each plot
	for(int i = 0;i < plot.Length();i++){
		plot[i].minY = minY;
		plot[i].maxY = maxY;
	}

	// Update the axis tick limits
	leftAxis.tick_min = rightAxis.tick_min = minY;
	leftAxis.tick_max = rightAxis.tick_max = maxY;
	
	// Update the axis scale type
	leftAxis.scaleType = rightAxis.scaleType = y_scaleType;

	if(y_scaleType == SCALE_LOG){
		// Auto scale tick space for log graphs for now
		leftAxis.MajorTickSpace(1.0f, TRUE);
		rightAxis.MajorTickSpace(1.0f, TRUE);
		leftAxis.MinorTickSpace(1.0f, TRUE);
		rightAxis.MinorTickSpace(1.0f, TRUE);
	}
	else{ // y_scaleType == SCALE_LINEAR

		// Adjust the number of tick marks for the linear plot
		leftAxis.MajorTickSpace((maxY - minY)/NUM_MAJOR_TICK);
		rightAxis.MajorTickSpace(leftAxis.MajorTickSpace());

		leftAxis.MinorTickSpace((maxY - minY)/NUM_MINOR_TICK);
		rightAxis.MinorTickSpace(leftAxis.MinorTickSpace());
	}
}

void SGraph::SetDefaultPlotRangeY(float m_minY, float m_maxY)
{
	isStickyScalingY = TRUE;

	defaultMinY = m_minY;
	defaultMaxY = m_maxY;
}

void SGraph::SetBoundingBox()
{
	for(int i = 0;i < plot.Length();i++){
		plot[i].x1 = x1;
		plot[i].x2 = x2;
		plot[i].y1 = y1;
		plot[i].y2 = y2;
	}
}

void SGraph::ShiftY(float scale)
{
	if(y_scaleType == SCALE_LOG)
		return;

	float dy = maxY - minY;
	float shift_y = scale*dy;

	// Special care must be taken when shifting an axis whose length, dx,
	// is approaching precision limits.
	minY += shift_y;
	maxY = minY + dy;

	SetPlotRangeY(minY, maxY, TRUE /* Override default graph scaling */);
}

void SGraph::ShiftX(float scale)
{
	if(x_scaleType == SCALE_LOG)
		return;

	// Special care must be taken when shifting an axis whose length, dx,
	// is approaching precision limits.
	float dx = maxX - minX;
	float shift_x = scale*dx;

	minX += shift_x;
	maxX = minX + dx;

	SetPlotRangeX(minX, maxX, TRUE /* Override default graph scaling */);
}

int SGraph::ImportPlot(CString &name, int type)
{
	// Keep the user informed of the read progress by writing
	// the current plot read to the status bar
	CMainFrame *pFrame = (CMainFrame*)AfxGetApp()->m_pMainWnd;
	CStatusBar* pStatus = &pFrame->m_wndStatusBar;
	CString statusBuffer;

	// Note that audio files are handled differently
	if(type == AUDIO_WAVE){
		if(pStatus){
			statusBuffer.Format("Reading audio data into plot %d", plot.Length());
			pStatus->SetPaneText(INFO_PANE, statusBuffer);
		}

		return ImportAudioPlot(name, type);
	}

	// Open the file
	fstream fin(name, ios::in | ios::nocreate);

	if(fin.fail())
		// Can't open the file
		return S_ERROR_IO_OPEN;

	IOBuffer io_buff(fin);

	// Keep a list of user commands (specified via the import file)
	ImportOptionList cmd;

	int curr_set = 0;
	int return_value;

	// The old way of reading data used the following recursive call to 
	// ImportPlot. However - I found that large data sets will reach
	// the stack limit for recursive function calls (thereby necessitating the
	// while loop found at the end of this function).
	//return ImportPlot(fin, 0 /* curr_set */, cmd, name, type);
	do{
		if(pStatus){
			statusBuffer.Format("Reading data into plot %d", plot.Length());
			pStatus->SetPaneText(INFO_PANE, statusBuffer);
		}
	}
	while((return_value = ImportPlot(io_buff, curr_set, cmd, name, type)) == 0);

	fin.close();

	return return_value;
}

int SGraph::ImportPlot(HANDLE memPtr, int type)
{
	// Keep the user informed of the read progress by writing
	// the current plot read to the status bar
	CMainFrame *pFrame = (CMainFrame*)AfxGetApp()->m_pMainWnd;
	CStatusBar* pStatus = &pFrame->m_wndStatusBar;
	CString statusBuffer;

	CString dummyName = _T("");

	IOBuffer io_buff(memPtr);

	// Keep a list of user commands (specified via the import file)
	ImportOptionList cmd;

	int curr_set = 0;
	int return_value;

	// The old way of reading data used the following recursive call to 
	// ImportPlot. However - I found that large data sets will reach
	// the stack limit for recursive function calls (thereby necessitating the
	// while loop found at the end of this function).
	//return ImportPlot(fin, 0 /* curr_set */, cmd, name, type);
	do{
		if(pStatus){
			statusBuffer.Format("Reading data into plot %d", plot.Length());
			pStatus->SetPaneText(INFO_PANE, statusBuffer);
		}
	}
	while((return_value = ImportPlot(io_buff, curr_set, cmd, dummyName, type)) == 0);

	return return_value;
}

int SGraph::ImportPlot(IOBuffer &io_buff, int &curr_set, ImportOptionList &cmd, CString &name, int type)
{
	// The maximum line length
	//#define		N	2048
	#define		N	524288 // 512 kb

	// A fixed length buffer (hence the max line length is N).
	char buffer[N + 1];

	// A white space mask
	char mask[] = "\t\r\n ";
	char *ptr;

	BOOL underflow = FALSE;

	// Store the data from a single line 
	stack<float> line;
	DataPoint tmp_data;
	float x;
	double tmp_double;
	int i;
	
	// Have we hit a white space break that indicates that we
	// should start reading a new plot?
	int read_a_new_plot = FALSE;

	// The current line number
	int index = 1;

	// The number of columns read on each line (for error testing)
	int num_col = -1;

	// The number of new plots to add
	int num_new_plot = 0;
	
	// The number of existing plots
	int num_old_plot = plot.Length();

	while(!io_buff.eof() && !read_a_new_plot){
		// Read in a line
		io_buff.getline(buffer, N + 1 /* +1 for the end of string terminator */);

		// Test for a buffer overflow
		if(strlen(buffer) == N)
			return S_ERROR_IO_BUFFER;
		
		// Discard any comments after checking them for user specfied
		// options
		if((ptr = strchr(buffer, S_COMMENT_CHAR)) != NULL){
			ImportOptions tmp;
			
			switch(tmp.Parse(ptr)){
				case PARSE_OPTIONS_DATA:
					cmd.Add(tmp);
					break;
				case PARSE_OPTIONS_ERROR:
					AfxMessageBox("Error reading option from import file!");
					break;
				case PARSE_OPTIONS_ERROR_MALFORMED:
					AfxMessageBox("Malformed option in import file!");
					break;
				case PARSE_OPTIONS_ERROR_BAD_INDEX:
					AfxMessageBox("Error reading set index from import file!");
					break;
				case PARSE_OPTIONS_ERROR_BAD_POINT_STYLE:
					AfxMessageBox("Bad point style in import file!");
					break;
				case PARSE_OPTIONS_ERROR_BAD_LINE_STYLE:
					AfxMessageBox("Bad line style in import file!");
					break;
				case PARSE_OPTIONS_ERROR_BAD_LINE_SIZE:
					AfxMessageBox("Bad line size in import file!");
					break;
				case PARSE_OPTIONS_ERROR_BAD_POINT_SIZE:
					AfxMessageBox("Bad point size in import file!");
					break;
				case PARSE_OPTIONS_ERROR_BAD_POINT_COLOR:
					AfxMessageBox("Bad point color in import file!");
					break;
				case PARSE_OPTIONS_ERROR_BAD_LINE_COLOR:
					AfxMessageBox("Bad line color in import file!");
					break;
				case PARSE_OPTIONS_ERROR_BAD_ERROR_BAR:
					AfxMessageBox("Bad error bar option in import file!");
					break;
				case PARSE_OPTIONS_ERROR_BAD_CUSTOM_COLOR:
					AfxMessageBox("Bad custom color in import file!");
					break;
				case PARSE_OPTIONS_ERROR_BAD_GRAPH_DATA:
					AfxMessageBox("Bad graph option in import file!");
					break;
				case PARSE_OPTIONS_ERROR_BAD_XAXIS_TILE:
					AfxMessageBox("Bad xaxis title in import file!");
					break;
				case PARSE_OPTIONS_ERROR_BAD_YAXIS_TILE:
					AfxMessageBox("Bad yaxis title in import file!");
					break;
			}
			
			// Terminate the string by overwriting the comment
			*ptr = '\0';
		}
		else
			// Save the buffer for parsing
			ptr = buffer;

		int str_len;

		while((str_len = strlen(ptr)) > 0){
			// Eat leading white space
			ptr += strspn(ptr, mask);
			
			// If we have just read a line that contains only ' ' or '\t'
			// characters, start a new set! Test to make sure that the line list
			// doesn't contain any elements (otherwise, trailing white space will
			// derail the parser).
			
			// This is the old version that did not handle trailing white space
			// correctly ...
			//if((strlen(ptr) == 0) && (line.Length() == 0)){
			//	if(str_len > 0)
			//		read_a_new_plot = TRUE;
			//
				// We just read a line with no data - so break out of this loop!
			//	break;
			//}

			// This is the new version that still needs more testing ...
			if(strlen(ptr) == 0){
				if(line.Length() == 0)
					// We just read a line with no data, so start a new plot
					read_a_new_plot = TRUE;
				// -- else --
				// we just read trailing white space, process the data we've
				// already read and move to the next line.
				
				break;
			}

			char *tmp = strpbrk(ptr, mask);

			if(tmp == NULL){
				// Last item on line
				if(line.Length() == 0){
					// Store the first number on the line
					tmp_double = atof(ptr);

					// Can we store the number as a float?
					if(fabs(tmp_double) > FLT_MAX)
						return S_ERROR_IO_FLOAT;

					// Only test for underflow until the first instance occurs
					//if(!underflow && (fabs(tmp_double) < FLT_MIN)){
					//	underflow = TRUE;
					//	if(AfxMessageBox("Underflow reading data. Set value to zero and "
					//		"Continue?", MB_YESNO) == IDNO)
					//		return S_ERROR_IO_FLOAT;
					//}

					x = (float)(tmp_double);
					try{
						line.Push(x);
					}
					catch(...){
						// Memory error!
						return S_ERROR_MEM_ALLOC;
					}
				}
				else{
					tmp_double = atof(ptr);

					// Can we store the number as a float?
					if(fabs(tmp_double) > FLT_MAX)
						return S_ERROR_IO_FLOAT;
					
					// Only test for underflow until the first instance occurs
					//if(!underflow && (fabs(tmp_double) < FLT_MIN)){
					//	underflow = TRUE;
					//	if(AfxMessageBox("Underflow reading data. Set value to zero and "
					//		"Continue?", MB_YESNO) == IDNO)
					//		return S_ERROR_IO_FLOAT;
					//}

					try{
						line.Push((float)(tmp_double));
					}
					catch(...){
						// Memory error!
						return S_ERROR_MEM_ALLOC;
					}
				}
				// Terminate the string
				*ptr = '\0';
			}
			else{
				*tmp = '\0';

				if(line.Length() == 0){
					// Store the first number on the line
					tmp_double = atof(ptr);

					// Can we store the number as a float?
					if(fabs(tmp_double) > FLT_MAX)
						return S_ERROR_IO_FLOAT;
					
					// Only test for underflow until the first instance occurs
					//if(!underflow && (fabs(tmp_double) < FLT_MIN)){
					//	underflow = TRUE;
					//	if(AfxMessageBox("Underflow reading data. Set value to zero and "
					//		"Continue?", MB_YESNO) == IDNO)
					//		return S_ERROR_IO_FLOAT;
					//}

					x = (float)(tmp_double);
					try{
						line.Push(x);
					}
					catch(...){
						// Memory error!
						return S_ERROR_MEM_ALLOC;
					}
				}
				else{
					tmp_double = atof(ptr);

					// Can we store the number as a float?
					if(fabs(tmp_double) > FLT_MAX)
						return S_ERROR_IO_FLOAT;
					
					// Only test for underflow until the first instance occurs
					//if(!underflow && (fabs(tmp_double) < FLT_MIN)){
					//	underflow = TRUE;
					//	if(AfxMessageBox("Underflow reading data. Set value to zero and "
					//		"Continue?", MB_YESNO) == IDNO)
					//		return S_ERROR_IO_FLOAT;
					//}

					try{
						line.Push((float)(tmp_double));
					}
					catch(...){
						// Memory error!
						return S_ERROR_MEM_ALLOC;
					}
				}

				ptr = tmp + 1;
			}
		}
		
		if(line.Length() > 0){
			if(num_col == -1){
				// Set num_col for the first time
				num_col = line.Length();

				// Allocate memory for the plot(s)
				if((type == SINGLE_PLOT) || (type == SINGLE_PLOT_DX) || 
				   (type == SINGLE_PLOT_DY) || (type == SINGLE_PLOT_DX_DY))
					num_new_plot = 1;
				else{
					if(num_col == 1)
						num_new_plot = 1;
					else
						num_new_plot = num_col - 1;
				}

				for(i = 0;i < num_new_plot;i++){
					// Append new empty plots to the list
					SPlot new_plot;

					// Assume data is all positive until we read a number <= 0.0
					new_plot.canXLog = TRUE;
					new_plot.canYLog = TRUE;
					
					// Save the file name
					new_plot.info = name;

					try{
						plot.Push(new_plot);
					}
					catch(...){
						// Memory error!
						return S_ERROR_MEM_ALLOC;
					}
				}

				// Initialize the plot bounding box
				SetBoundingBox();
				
				#ifdef INCREMENT_LINE_TYPE
				// Increment the line types
				if(num_old_plot != 0){
					for(i = num_old_plot;i < plot.Length();i++)
						plot[i].lineType = (plot[i - 1].lineType + 1)%SPlot::LINE_NONE;
				}
				else {
					for(i = 1;i < plot.Length();i++)
						plot[i].lineType = (plot[i - 1].lineType + 1)%SPlot::LINE_NONE;
				}
				#endif // INCREMENT_LINE_TYPE

				// Increment the line colors
				if(num_old_plot != 0){
					for(i = num_old_plot;i < plot.Length();i++){
						plot[i].lineColor = 
							IndexToColor((ColorToIndex(plot[i - 1].lineColor) + 1) % 
							S_CLR_NONE);

						// Exclude white from the list of colors!
						if(ColorToIndex(plot[i].lineColor) == S_CLR_WHITE)
							plot[i].lineColor = IndexToColor(S_CLR_WHITE + 1);
					}
				}
				else {
					for(i = 1;i < plot.Length();i++){
						plot[i].lineColor = 
							IndexToColor((ColorToIndex(plot[i - 1].lineColor) + 1) % 
							S_CLR_NONE);

						// Exclude white from the list of colors!
						if(ColorToIndex(plot[i].lineColor) == S_CLR_WHITE)
							plot[i].lineColor = IndexToColor(S_CLR_WHITE + 1);
					}
				}
			}
			else{
				// Make sure that the number of columns
				// hasn't changed! However, don't return here if
				// we need to read a new plot.
				if((num_col != line.Length()) && !read_a_new_plot){
					// Remove the added plots before returning
					for(i = 0;i < num_new_plot;i++)
						plot.Pop();

					return S_ERROR_IO_NUM_COL;
				}
			}
		}
		
		// Prepare to iterate through the data we just read
		stackIterator<float> next(line);

		// X and Y error bars
		float x_ebar, y_ebar; 

		switch(type){
			case SINGLE_PLOT:
				switch(line.Length()){
					case 0:
						// Do nothing
						break;
					case 1:
						tmp_data.x = (float)(index);
						tmp_data.y = x; // <-- We've already saved the first number
						plot[num_old_plot].data.Push(tmp_data);

						// Test for negative (or zero) numbers
						// (Don't bother with this test, since index > 0).
						//if(tmp_data.x <= 0.0)
						//	plot[num_old_plot].canXLog = FALSE;
					
						if(tmp_data.y <= 0.0)
							plot[num_old_plot].canYLog = FALSE;

						index ++;
						break;
					default:
						while(line.Length() > 2)
							line.Pop();
						
						// We have modified the stack of data points, so
						// reset the stack iterator.
						next.reset(line);

						// Get y first (FILO stack)
						next(tmp_data.y);
						tmp_data.x = x; // <-- We've already saved the first number
						plot[num_old_plot].data.Push(tmp_data);

						// Test for negative (or zero) numbers
						if(tmp_data.x <= 0.0)
							plot[num_old_plot].canXLog = FALSE;
					
						if(tmp_data.y <= 0.0)
							plot[num_old_plot].canYLog = FALSE;

						index ++;				
						break;
				};
				break;

			case MULTI_PLOT:
				switch(line.Length()){
					case 0:
						// Do nothing
						break;
					case 1:
						tmp_data.x = (float)(index);
						tmp_data.y = x; // <-- We've already saved the first number
						plot[num_old_plot].data.Push(tmp_data);

						// Test for negative (or zero) numbers
						if(tmp_data.x <= 0.0)
							plot[num_old_plot].canXLog = FALSE;
					
						if(tmp_data.y <= 0.0)
							plot[num_old_plot].canYLog = FALSE;

						index ++;
						break;
					case 2:
						// Get y first (FILO stack)
						next(tmp_data.y);
						tmp_data.x = x; // <-- We've already saved the first number
						plot[num_old_plot].data.Push(tmp_data);

						// Test for negative (or zero) numbers
						if(tmp_data.x <= 0.0)
							plot[num_old_plot].canXLog = FALSE;
					
						if(tmp_data.y <= 0.0)
							plot[num_old_plot].canYLog = FALSE;

						index ++;				
						break;
					default:
						tmp_data.x = x;
						for(i = plot.Length() - 1;i >= num_old_plot;i--){
							next(tmp_data.y);
							plot[i].data.Push(tmp_data);

							// Test for negative (or zero) numbers
							if(tmp_data.x <= 0.0)
								plot[i].canXLog = FALSE;

							if(tmp_data.y <= 0.0)
								plot[i].canYLog = FALSE;

						}
						index ++;
						break;
				};
				break;
			case SINGLE_PLOT_DX:
				if(line.Length() == 0)
					break;

				// Require at least three points on a line
				if(line.Length() < 3)
					return S_ERROR_IO_NUM_COL;
				
				while(line.Length() > 3)
					line.Pop();
				
				// We have modified the stack of data points, so
				// reset the stack iterator.
				next.reset(line);

				// Get the error bar first (FILO stack)
				next(x_ebar);
				
				// Error bars MUST be positive values!
				x_ebar = (float)(fabs(x_ebar));

				// Then read the y value
				next(tmp_data.y);

				tmp_data.x = x; // <-- We've already saved the first number

				plot[num_old_plot].data.Push(tmp_data);
				
				// Save the error bar
				plot[num_old_plot].xErrorBar.Push(x_ebar);

				// Test for negative (or zero) numbers
				if((tmp_data.x - x_ebar) <= 0.0)
					plot[num_old_plot].canXLog = FALSE;
					
				if(tmp_data.y <= 0.0)
					plot[num_old_plot].canYLog = FALSE;

				index ++;
				break;
			case SINGLE_PLOT_DY:
				if(line.Length() == 0)
					break;

				// Require at least three points on a line
				if(line.Length() < 3)
					return S_ERROR_IO_NUM_COL;

				while(line.Length() > 3)
					line.Pop();
				
				// We have modified the stack of data points, so
				// reset the stack iterator.
				next.reset(line);

				// Get the error bar first (FILO stack)
				next(y_ebar);
				
				// Error bars MUST be positive values!
				y_ebar = (float)(fabs(y_ebar));

				// Then read the y value
				next(tmp_data.y);

				tmp_data.x = x; // <-- We've already saved the first number

				plot[num_old_plot].data.Push(tmp_data);
				
				// Save the error bar
				plot[num_old_plot].yErrorBar.Push(y_ebar);

				// Test for negative (or zero) numbers
				if(tmp_data.x <= 0.0)
					plot[num_old_plot].canXLog = FALSE;
					
				if((tmp_data.y - y_ebar) <= 0.0)
					plot[num_old_plot].canYLog = FALSE;

				index ++;
				break;
			case SINGLE_PLOT_DX_DY:
				if(line.Length() == 0)
					break;

				// Require at least four points on a line
				if(line.Length() < 4)
					return S_ERROR_IO_NUM_COL;

				while(line.Length() > 4)
					line.Pop();
				
				// We have modified the stack of data points, so
				// reset the stack iterator.
				next.reset(line);

				// Get the y error bar first (FILO stack)
				next(y_ebar);
				next(x_ebar);
				
				// Error bars MUST be positive values!
				x_ebar = (float)(fabs(x_ebar));
				y_ebar = (float)(fabs(y_ebar));

				// Then read the y value
				next(tmp_data.y);

				tmp_data.x = x; // <-- We've already saved the first number

				plot[num_old_plot].data.Push(tmp_data);
				
				// Save the error bar
				plot[num_old_plot].xErrorBar.Push(x_ebar);
				plot[num_old_plot].yErrorBar.Push(y_ebar);

				// Test for negative (or zero) numbers
				if((tmp_data.x - x_ebar) <= 0.0)
					plot[num_old_plot].canXLog = FALSE;
					
				if((tmp_data.y - y_ebar) <= 0.0)
					plot[num_old_plot].canYLog = FALSE;

				index ++;
				break;
		};

		// Free stack
		while(!line.IsEmpty())
			line.Pop();
	}

	// Don't try to read a new plot from an empty file
	if(io_buff.eof())
		read_a_new_plot = FALSE;

	// Has any data been read?
	if(index == 1){
		// Remove the added plots before returning
		for(i = 0;i < num_new_plot;i++)
			plot.Pop();

		// Are we only trying to read a single set?
		if(!read_a_new_plot && (curr_set == 0))
			// We've tried to read the single data set and failed
			return S_ERROR_NO_DATA;
	}
	
	// Turn on error bars if we've read them in
	for(i = num_old_plot;i < plot.Length();i++){
		if((type == SINGLE_PLOT_DX) || (type == SINGLE_PLOT_DX_DY))
			plot[i].display_x_error_bar = SPlot::BAR_BOTH;

		if((type == SINGLE_PLOT_DY) || (type == SINGLE_PLOT_DX_DY))
			plot[i].display_y_error_bar = SPlot::BAR_BOTH;
	}

	// Apply any user options
	for(i = num_old_plot;i < plot.Length();i++){
		ImportOptions opt;

		if(cmd.GetOption(i - num_old_plot + curr_set, opt)){
			if(opt.set_lineSize != -1)
				plot[i].lineSize = opt.set_lineSize;
			
			if(opt.set_pointSize != -1)
				plot[i].pointSize = opt.set_pointSize;
			
			if(opt.set_lineType != -1)
				plot[i].lineType = opt.set_lineType;

			if(opt.set_pointType != -1)
				plot[i].pointType = opt.set_pointType;

			if(opt.set_pointColor != -1){
				plot[i].brush.lbColor = IndexToColor(opt.set_pointColor);

				// Set the symbol brush
				if(plot[i].brush.lbColor == NO_COLOR)
					plot[i].brush.lbStyle = BS_HOLLOW;
				else
					plot[i].brush.lbStyle = BS_SOLID;
			}

			if(opt.set_lineColor != -1)
				plot[i].lineColor = IndexToColor(opt.set_lineColor);

			if(opt.set_errorBarDisplay != -1){
				plot[i].display_x_error_bar = opt.set_errorBarDisplay;
				plot[i].display_y_error_bar = opt.set_errorBarDisplay;
			}

			if(opt.set_info_flag != -1)
				plot[i].info = opt.set_info;
		}
	}

	// Set any graph options if the options have not already been set
	if(title.GetText() == PRESTO_DEFAULT_TITLE)
		title.SetText(cmd.GraphTitle());

	if(bottomAxisLable.GetText() == PRESTO_DEFAULT_X_AXIS_LABLE)
		bottomAxisLable.SetText(cmd.XAxisTitle());

	if(leftAxisLable.GetText() == PRESTO_DEFAULT_Y_AXIS_LABLE)
		leftAxisLable.SetText(cmd.YAxisTitle());

	if(read_a_new_plot){
		// Make sure to pass the current set to be read (for setting user options)
		curr_set += plot.Length() - num_old_plot;

		return 0; // 0 is the code for more data to read
	}

	return -1; // -1 is the code for no more data to read
}

// Export data from a single plot to an ASCII data set
void SGraph::ExportPlot()
{
	CSetControlDialog scDlg(CSetControlDialog::SET_SELECT_MULTIPLE,
		&plot);

	if((scDlg.DoModal() == IDOK) && (scDlg.SelectedStates().Length() != 0)){
		CFileDialog fDlg(FALSE); // Save as dialog
		
		// Set the title of the file dialog
		fDlg.m_ofn.lpstrTitle = "Export";

		// Set the current directory
		char dirBuffer[1024];

		// If the buffer is not big enought, DON'T set the current 
		// directory
		if(GetCurrentDirectory(1023, dirBuffer) != 0)
			fDlg.m_ofn.lpstrInitialDir = dirBuffer;

		if(fDlg.DoModal() == IDOK){
			FILE *fout = fopen(fDlg.GetPathName(), "w");

			if(!fout){
				AfxMessageBox("Can't open " + fDlg.GetPathName() + 
					"for writing");
				return;
			}
			
			for(int i = 0;i < scDlg.SelectedStates().Length();i++){
				// Write the plot to disk
				SPlot &p = plot[scDlg.SelectedStates()[i]];

				// First write some header information
				fprintf(fout, "# %s (set %d)\n", LPCTSTR(p.info), 
					scDlg.SelectedStates()[0]);

				stackIterator<DataPoint> next(p.data);
				stackIterator<float> nextYError(p.yErrorBar);
				stackIterator<float> nextXError(p.xErrorBar);

				DataPoint xy;
				float x_ebar, y_ebar;
				
				if(!p.yErrorBar.IsEmpty()){
					if(!p.xErrorBar.IsEmpty()){
						while(next(xy) && nextXError(x_ebar) && nextYError(y_ebar)){
							fprintf(fout, "%g\t%g\t%g\t%g\n", xy.x, xy.y, x_ebar, y_ebar);
						}
					}
					else{
						while(next(xy) && nextYError(y_ebar)){
							fprintf(fout, "%g\t%g\t%g\n", xy.x, xy.y, y_ebar);
						}
					}
				}
				else{
					if(!p.xErrorBar.IsEmpty()){
						while(next(xy) && nextXError(x_ebar)){
							fprintf(fout, "%g\t%g\t%g\n", xy.x, xy.y, x_ebar);
						}
					}
					else{
						while(next(xy)){
							fprintf(fout, "%g\t%g\n", xy.x, xy.y);
						}
					}
				}

				fprintf(fout, "\t\n");
			}
			
			fclose(fout);
		}
	}
}

void SGraph::CopyPlot()
{
	CSetControlDialog scDlg(CSetControlDialog::SET_SELECT_MULTIPLE,
		&plot);

	if((scDlg.DoModal() == IDOK) && (scDlg.SelectedStates().Length() != 0)){
		
		// Format size is based on a number of the form: n.nnnnnnne+xxx
		const int format_size = 14;
		int i;
		int buffer_size = 0;
		HGLOBAL hgtext_buffer;
		char* text_buffer;

		// Prepare a buffer to hold the clipboard data
		for(i = 0;i < scDlg.SelectedStates().Length();i++){
			SPlot &p = plot[scDlg.SelectedStates()[i]];

			int num_error_bar = 0;

			if(!p.yErrorBar.IsEmpty())
				num_error_bar ++;

			if(!p.xErrorBar.IsEmpty())
				num_error_bar ++;
			
			// Make room for "%E\t%E\r\n" + error bar data
			buffer_size += format_size*(2*p.data.Length() + num_error_bar) 
				+ 3 + num_error_bar;

			// Add a "\t\r\n" between every set
			buffer_size += 3;
		}

		// Allocate memory for the buffer
		hgtext_buffer = GlobalAlloc(GMEM_MOVEABLE | GMEM_DDESHARE, buffer_size);

		if(hgtext_buffer == NULL){
			AfxMessageBox("Unable to get memory for data copy");
			return;
		}
		
		text_buffer = (char*)(GlobalLock(hgtext_buffer));

		try{
			for(i = 0;i < scDlg.SelectedStates().Length();i++){
				int tmp_size;

				// Write the plot to the text buffer
				SPlot &p = plot[scDlg.SelectedStates()[i]];

				stackIterator<DataPoint> next(p.data);
				stackIterator<float> nextYError(p.yErrorBar);
				stackIterator<float> nextXError(p.xErrorBar);

				DataPoint xy;
				float x_ebar, y_ebar;
				
				if(buffer_size < 0)
					throw "Text buffer exhausted! Copy error!";
					

				if(!p.yErrorBar.IsEmpty()){
					if(!p.xErrorBar.IsEmpty()){
						while(next(xy) && nextXError(x_ebar) && nextYError(y_ebar)){
							tmp_size = sprintf(text_buffer, "%.7g\t%.7g\t%.7g\t%.7g\r\n", 
								xy.x, xy.y, x_ebar, y_ebar);

							// Advance the buffer pointer to the next empty location
							text_buffer += tmp_size;

							buffer_size -= tmp_size;
						}
					}
					else{
						while(next(xy) && nextYError(y_ebar)){
							tmp_size = sprintf(text_buffer, "%.7g\t%.7g\t%.7g\r\n", 
								xy.x, xy.y, y_ebar);

							// Advance the buffer pointer to the next empty location
							text_buffer += tmp_size;

							buffer_size -= tmp_size;
						}
					}
				}
				else{
					if(!p.xErrorBar.IsEmpty()){
						while(next(xy) && nextXError(x_ebar)){
							tmp_size = sprintf(text_buffer, "%.7g\t%.7g\t%.7g\r\n", 
								xy.x, xy.y, x_ebar);
							
							// Advance the buffer pointer to the next empty location
							text_buffer += tmp_size;

							buffer_size -= tmp_size;
						}
					}
					else{
						while(next(xy)){
							tmp_size = sprintf(text_buffer, "%.7g\t%.7g\r\n", xy.x, xy.y);

							// Advance the buffer pointer to the next empty location
							text_buffer += tmp_size;

							buffer_size -= tmp_size;
						}
					}
				}

				if(buffer_size < 0)
					throw "Text buffer exhausted! Copy error!";

				tmp_size = sprintf(text_buffer, "\t\r\n");

				// Advance the buffer pointer to the next empty location
				text_buffer += tmp_size;
				
				buffer_size -= tmp_size;
			}
		}
		catch(char *m_error){
			// Clean up the memory
			GlobalUnlock(hgtext_buffer);
			GlobalFree(hgtext_buffer);

			AfxMessageBox(m_error);
		}

		// Copy to clipboard
		if(::OpenClipboard(NULL)){
			EmptyClipboard();
			GlobalUnlock(hgtext_buffer);
			::SetClipboardData(CF_TEXT, hgtext_buffer);
			CloseClipboard();
		}
		else{
			AfxMessageBox("Can't open the clipboard (it's already open in "
				"another application).");
		}
	}
}

BOOL SGraph::ModalActivatePlot()
{
	CSetControlDialog scDlg(CSetControlDialog::SET_ACTIVATE,
		&plot);
	
	if(scDlg.DoModal() == IDOK){
		// Make sure that all of the newly activated plots can be displayed
		// on the current axis scales.
		SetGraphType(x_scaleType, y_scaleType);

		return TRUE;
	}

	return FALSE;
}

BOOL SGraph::ModalDeactivatePlot()
{
	CSetControlDialog scDlg(CSetControlDialog::SET_DEACTIVATE,
		&plot);
	
	return (scDlg.DoModal() == IDOK);
}

BOOL SGraph::ModalPlotToBack()
{
	CSetControlDialog scDlg(CSetControlDialog::SET_SEND_TO_BACK,
		&plot);
	
	return (scDlg.DoModal() == IDOK);
}

BOOL SGraph::ModalPlotToFront()
{
	CSetControlDialog scDlg(CSetControlDialog::SET_BRING_TO_FRONT,
		&plot);
	
	return (scDlg.DoModal() == IDOK);
}

BOOL SGraph::ModalDeletePlot()
{
	CSetControlDialog scDlg(CSetControlDialog::SET_DELETE,
		&plot);
	
	return (scDlg.DoModal() == IDOK);
}

CSelectRect SGraph::SelectRect(CDC* pDC, CRect &extent)
{
	int tmp_x1 = XtoScreen(x1, extent);
	int tmp_x2 = XtoScreen(x2, extent);
	int tmp_y1 = YtoScreen(y1, extent);
	int tmp_y2 = YtoScreen(y2, extent);

	CSelectRect select_rect;

	// The angle is zero
	select_rect.X((int)(0.5*(tmp_x1 + tmp_x2)));
	select_rect.Y((int)(0.5*(tmp_y1 + tmp_y2)));

	select_rect.SizeX(tmp_x2 - tmp_x1);
	select_rect.SizeY(tmp_y1 - tmp_y2);

	return select_rect;
}

// This function contains a pointer to the back buffer (for double
// buffering) which is rather awkward. It would be nice to make
// double buffering transparant to the SGraph class, but this would
// involve passing a list of pointers back to the CSgraphView
// class and would be even more awkward.
BOOL SGraph::Select(BOOL status, CPoint &xy, CDC* pDC, BOOL doubleBuffer,
					CDC* backDC, CRect &extent)
{
	// Has the graph been selected?
	SObject::Select(xy, pDC, extent);

	// Have any text objects been selected?
	BOOL wasSelected;

	wasSelected = title.isSelected;

	if(status)
		title.isSelected = FALSE;
	else{
		status = title.Select(xy, pDC, extent);
		isSelected = status || isSelected;
	}

	if((wasSelected && !title.isSelected) || 
	   (!wasSelected && title.isSelected)){
		title.DrawSelect(pDC, extent);

		if(doubleBuffer)
			title.DrawSelect(backDC, extent);
	}

	wasSelected = leftAxisLable.isSelected;

	if(status)
		leftAxisLable.isSelected = FALSE;
	else{
		status = leftAxisLable.Select(xy, pDC, extent);
		isSelected = status || isSelected;
	}

	if((wasSelected && !leftAxisLable.isSelected) || 
	   (!wasSelected && leftAxisLable.isSelected)){
		leftAxisLable.DrawSelect(pDC, extent);

		if(doubleBuffer)
			leftAxisLable.DrawSelect(backDC, extent);
	}

	wasSelected = bottomAxisLable.isSelected;

	if(status)
		bottomAxisLable.isSelected = FALSE;
	else{
		status = bottomAxisLable.Select(xy, pDC, extent);
		isSelected = status || isSelected;
	}

	if((wasSelected && !bottomAxisLable.isSelected) || 
	   (!wasSelected && bottomAxisLable.isSelected)){
		bottomAxisLable.DrawSelect(pDC, extent);

		if(doubleBuffer)
			bottomAxisLable.DrawSelect(backDC, extent);
	}
	
	// Has the legend been selected?
	wasSelected = legend.isSelected;

	if(status)
		legend.isSelected = FALSE;
	else{
		status = legend.Select(xy, pDC, extent);
		isSelected = status || isSelected;
	}

	if((wasSelected && !legend.isSelected) || 
	   (!wasSelected && legend.isSelected)){
		legend.DrawSelect(pDC, extent);

		if(doubleBuffer)
			legend.DrawSelect(backDC, extent);
	}

	// Have any axis objects been selected?
	wasSelected = leftAxis.isSelected;

	if(status)
		leftAxis.isSelected = FALSE;
	else{
		status = leftAxis.Select(xy, pDC, extent);
		isSelected = status || isSelected;
	}

	if((wasSelected && !leftAxis.isSelected) || 
	   (!wasSelected && leftAxis.isSelected)){
		leftAxis.DrawSelect(pDC, extent);

		if(doubleBuffer)
			leftAxis.DrawSelect(backDC, extent);
	}

	wasSelected = rightAxis.isSelected;

	if(status)
		rightAxis.isSelected = FALSE;
	else{
		status = rightAxis.Select(xy, pDC, extent);
		isSelected = status || isSelected;
	}

	if((wasSelected && !rightAxis.isSelected) || 
	   (!wasSelected && rightAxis.isSelected)){
		rightAxis.DrawSelect(pDC, extent);

		if(doubleBuffer)
			rightAxis.DrawSelect(backDC, extent);
	}

	wasSelected = topAxis.isSelected;

	if(status)
		topAxis.isSelected = FALSE;
	else{
		status = topAxis.Select(xy, pDC, extent);
		isSelected = status || isSelected;
	}

	if((wasSelected && !topAxis.isSelected) || 
	   (!wasSelected && topAxis.isSelected)){
		topAxis.DrawSelect(pDC, extent);

		if(doubleBuffer)
			topAxis.DrawSelect(backDC, extent);
	}

	wasSelected = bottomAxis.isSelected;

	if(status)
		bottomAxis.isSelected = FALSE;
	else{
		status = bottomAxis.Select(xy, pDC, extent);
		isSelected = status || isSelected;
	}

	if((wasSelected && !bottomAxis.isSelected) || 
	   (!wasSelected && bottomAxis.isSelected)){
		bottomAxis.DrawSelect(pDC, extent);

		if(doubleBuffer)
			bottomAxis.DrawSelect(backDC, extent);
	}

	// If the graph or any part of the graph is selected, then the
	// graph is selected.
	return isSelected;
}

// What is the nearest active plot to the point selected 
// by the user?
int SGraph::PlotHitTest(int m_x, int m_y, float &f_x, float &f_y, CRect &extent)
{
	int plot_id = -1;
	double dist2 = DBL_MAX;
	float tmp_x, tmp_y;

	for(int i = 0;i < plot.Length();i++){
		if(plot[i].isDrawn){
			double tmp_dist2 = plot[i].HitTest(m_x, m_y, tmp_x, tmp_y, extent);

			// Distances are compared using data type double to guard against
			// overflow -- but make sure all values are greater than 0 to be safe.
			if(tmp_dist2 < 0.0)
				continue;

			if((tmp_dist2 >= 0.0) && ((tmp_dist2 < dist2) || (plot_id < 0))){
				dist2 = tmp_dist2;
				plot_id = i;
				f_x = tmp_x;
				f_y = tmp_y;
			}
		}
	}
		
	return plot_id;
}

void SGraph::Serialize(CArchive& ar)
{
	SObject::Serialize(ar);

	title.Serialize(ar);
	leftAxisLable.Serialize(ar);
	bottomAxisLable.Serialize(ar);

	leftAxis.Serialize(ar); 
	rightAxis.Serialize(ar);
	topAxis.Serialize(ar);
	bottomAxis.Serialize(ar);

	legend.Serialize(ar);

	// Serialize the list of plots
	plot.Serialize(ar);

	if(version > 1.07){
		// Read/save user defined lines
		lines.Serialize(ar);
	}

	if(ar.IsStoring()){
		ar << x_scaleType;
		ar << y_scaleType;
		ar << x_tick_location;
		ar << y_tick_location;
		ar << x_tick_label_location;
		ar << y_tick_label_location;
		ar << x1;
		ar << y1;
		ar << x2;
		ar << y2;
		
		ar << minX;
		ar << maxX;
		ar << minY;
		ar << maxY;
		ar << background_color;

		ar << isStickyScalingX;
		ar << isStickyScalingY;
	}
	else{
		ar >> x_scaleType;
		ar >> y_scaleType;
		ar >> x_tick_location;
		ar >> y_tick_location;
		ar >> x_tick_label_location;
		ar >> y_tick_label_location;
		ar >> x1;
		ar >> y1;
		ar >> x2;
		ar >> y2;
		
		ar >> minX;
		ar >> maxX;
		ar >> minY;
		ar >> maxY;

		ar >> background_color;

		if(version > 1.18){
			ar >> isStickyScalingX;
			ar >> isStickyScalingY;
		}
		else{
			isStickyScalingX = FALSE;
			isStickyScalingY = FALSE;
		}


		// If these values are ever negative, an error must have
		// occured reading in the file.
		if((x_scaleType < 0) || (y_scaleType < 0) || 
		   (x_tick_location < 0) || (y_tick_location < 0))
			AfxThrowArchiveException(CArchiveException::badIndex,
				ar.m_pDocument->GetTitle());

		if(isStickyScalingX){
			defaultMinX = minX;
			defaultMaxX = maxX;
		}

		if(isStickyScalingY){
			defaultMinY = minY;
			defaultMaxY = maxY;
		}
	}
}

// SetGraphType is used for setting new axis scale types AND for checking that 
// newly activated plots can be displayed on the CURRENT axis scales.
BOOL SGraph::SetGraphType(int xAxisType, int yAxisType)
{
	// Can we display the data on a log scale?
	if(xAxisType == SCALE_LOG){
		for(int i = 0;i < plot.Length();i++){
			if(plot[i].isDrawn){
				if(!plot[i].canXLog){
					AfxMessageBox("Warning! Can't use log scale, x data <= 0.0");

					// Reset all plots to a linear scale
					SetGraphType(SCALE_LINEAR, yAxisType);
					return FALSE;
				}

				if(plot[i].pointType == SPlot::POINT_HISTOGRAM_Y){
					AfxMessageBox("Warning! Can't use log scale while plotting a histogram!");
					
					// Reset all plots to a linear scale
					SetGraphType(SCALE_LINEAR, yAxisType);
					return FALSE;
				}
			}
		}

	}
	
	if(yAxisType == SCALE_LOG){
		for(int i = 0;i < plot.Length();i++){
			if(plot[i].isDrawn){
				if(!plot[i].canYLog){
					AfxMessageBox("Warning! Can't use log scale, y data <= 0.0");

					// Reset all plots to a linear scale
					SetGraphType(xAxisType, SCALE_LINEAR);
					return FALSE;
				}

				if(plot[i].pointType == SPlot::POINT_HISTOGRAM_X){
					AfxMessageBox("Warning! Can't use log scale while plotting a histogram!");

					// Reset all plots to a linear scale
					SetGraphType(xAxisType, SCALE_LINEAR);
					return FALSE;
				}
			}
		}
	}

	// Adjust the default scaling if needed
	if(isStickyScalingX && (x_scaleType != xAxisType)){
		if(xAxisType == SCALE_LOG){
			if((defaultMinX <= 0.0f) || (defaultMaxX <= 0.0f)){
				// "un-stick" the axis range
				isStickyScalingX = FALSE;
			}
		}
	}

	x_scaleType = xAxisType;

	if(isStickyScalingY && (y_scaleType != yAxisType)){
		if(yAxisType == SCALE_LOG){
			if((defaultMinY <= 0.0f) || (defaultMaxY <= 0.0f)){
				// "un-stick" the axis range
				isStickyScalingY = FALSE;
			}
		}
	}

	y_scaleType = yAxisType;
	
	SetPlotScaleType();

	// Rescan the data to display everything
	SetPlotRange();

	return TRUE;
}

void SGraph::QuickPlot(SEquation &m_quick_eq)
{
	plot_quick_eq = TRUE;

	if(quick_eq.GetError())
		quick_eq.ClearError();

	quick_eq = m_quick_eq;
}

void SGraph::ClearQuickPlot()
{
	plot_quick_eq = FALSE;
}

void SGraph::DrawQuickPlot(CDC *pDC, CRect &extent)
{
	// Plot the graph from minX to maxX in QUICK_PLOT_STEPS steps
	float tmp_y, tmp_x, dx, final_x;

	if(x_scaleType == SCALE_LINEAR){
		tmp_x = minX;
		final_x = maxX;
		dx = (maxX - minX)/QUICK_PLOT_STEPS;
	}
	else{ // x_scaleType == SCALE_LOG
		// When using a log scale, the minX and maxX values are stored
		// as log10(minX) and log10(maxX).
		tmp_x = (float)(pow(10.0f, minX));
		final_x = (float)(pow(10.0f, maxX));
		dx = (float)(pow(10.0f, (maxX - minX)/QUICK_PLOT_STEPS));
	}

	if(dx <= 0.0)
		return;

	if(!quick_eq.IsValid())
		return;
	
	BOOL first_point = FALSE;

	while(!first_point){

		// If the function NEVER evaluates between the required range
		// of minX and maxX, just return without plotting anything.
		if(tmp_x > final_x)
			return;

		tmp_y = quick_eq(tmp_x);
		
		if(quick_eq.GetError()){
			quick_eq.ClearError();
		}
		else{
			if((y_scaleType == SCALE_LOG) && (tmp_y <= 0.0f)){
				// Woops! Trying to plot a negative value on a 
				// log scale
				plot_quick_eq = FALSE;

				//AfxMessageBox("Trying to plot a negative function value\n\r"
				//	"on a log scale.");
				return;
			}

			first_point = TRUE;

			pDC->MoveTo(PlotX(tmp_x, extent), PlotY(tmp_y, extent)); 
		}
		
		if(x_scaleType == SCALE_LINEAR){
			tmp_x += dx;
		}
		else{ // x_scaleType == SCALE_LOG
			tmp_x *= dx;
		}
	}

	// Build a pen for drawing the function
	CPen funcpen(PS_SOLID, 1, RGB(0, 0, 0));
	CPen *oldpen = pDC->SelectObject(&funcpen); 

	while(tmp_x <= final_x){
		tmp_y = quick_eq(tmp_x);

		if(quick_eq.GetError()){
			quick_eq.ClearError();
		}
		else{
			if((y_scaleType == SCALE_LOG) && (tmp_y <= 0.0f)){
				// Woops! Trying to plot a negative value on a 
				// log scale
				plot_quick_eq = FALSE;

				pDC->SelectObject(oldpen);

				//AfxMessageBox("Trying to plot a negative function value\n\r"
				//	"on a log scale.");
				return;
			}

			pDC->LineTo(PlotX(tmp_x, extent), PlotY(tmp_y, extent)); 

			if(x_scaleType == SCALE_LINEAR){
				tmp_x += dx;
			}
			else{ // x_scaleType == SCALE_LOG
				tmp_x *= dx;
			}
		}
	}

	pDC->SelectObject(oldpen);
}

int SGraph::PlotX(float m_x, CRect &extent)
{
	if(x_scaleType == SCALE_LINEAR)
		return XtoScreen((m_x - minX)*(x2 - x1)/(maxX - minX) + x1, extent);

	return XtoScreen(((float)(log10(m_x)) - minX)*(x2 - x1)/(maxX - minX) + x1, extent);
}

int SGraph::PlotY(float m_y, CRect &extent)
{
	if(y_scaleType == SCALE_LINEAR)
		return YtoScreen((m_y - minY)*(y2 - y1)/(maxY - minY) + y1, extent);

	return YtoScreen(((float)(log10(m_y)) - minY)*(y2 - y1)/(maxY - minY) + y1, extent);
}

float SGraph::UpdateX(float m_x)
{
	float dx = m_x - x;
	
	switch(graph_move){
		case GRAPH_MOVE_ALL:
			x1 += dx;
			x2 += dx;

			title.MoveX(dx);
			leftAxisLable.MoveX(dx);
			bottomAxisLable.MoveX(dx);

			topAxis.MoveX(dx);
			bottomAxis.MoveX(dx);
			leftAxis.MoveX(dx);
			rightAxis.MoveX(dx);

			legend.MoveX(dx);

			x = 0.5f*(x2 + x1);
			break;
		case GRAPH_MOVE_LEFT_AXIS:
			x1 += dx;
			topAxis.line_length = bottomAxis.line_length = x2 - x1;

			title.MoveX(dx*0.5f);
			leftAxisLable.MoveX(dx);
			bottomAxisLable.MoveX(dx*0.5f);

			topAxis.MoveX(dx*0.5f);
			bottomAxis.MoveX(dx*0.5f);

			leftAxis.MoveX(dx);

			legend.MoveX(dx*0.5f);
			break;
		case GRAPH_MOVE_RIGHT_AXIS:
			x2 += dx;
			topAxis.line_length = bottomAxis.line_length = x2 - x1;

			title.MoveX(dx*0.5f);
			bottomAxisLable.MoveX(dx*0.5f);

			topAxis.MoveX(dx*0.5f);
			bottomAxis.MoveX(dx*0.5f);

			rightAxis.MoveX(dx);

			legend.MoveX(dx*0.5f);
			break;
		case GRAPH_MOVE_TOP_AXIS:
			return x;
		case GRAPH_MOVE_BOTTOM_AXIS:
			return x;
	};

	// Propagate the bounding box change to all of the plots
	SetBoundingBox();

	return x;
}

float SGraph::UpdateY(float m_y)
{
	float dy = m_y - y;
	
	switch(graph_move){
		case GRAPH_MOVE_ALL:
			y1 += dy;
			y2 += dy;

			title.MoveY(dy);
			leftAxisLable.MoveY(dy);
			bottomAxisLable.MoveY(dy);

			leftAxis.MoveY(dy);
			rightAxis.MoveY(dy);
			topAxis.MoveY(dy);
			bottomAxis.MoveY(dy);
			
			legend.MoveY(dy);

			y = 0.5f*(y2 + y1);
			break;
		case GRAPH_MOVE_LEFT_AXIS:
			return y;
		case GRAPH_MOVE_RIGHT_AXIS:
			return y;
		case GRAPH_MOVE_TOP_AXIS:
			y2 += dy;
			leftAxis.line_length = rightAxis.line_length = y2 - y1;

			title.MoveY(dy);
			leftAxisLable.MoveY(dy*0.5f);

			leftAxis.MoveY(dy*0.5f);
			rightAxis.MoveY(dy*0.5f);

			topAxis.MoveY(dy);

			legend.MoveY(dy*0.5f);
			break;
		case GRAPH_MOVE_BOTTOM_AXIS:
			y1 += dy;
			leftAxis.line_length = rightAxis.line_length = y2 - y1;

			leftAxisLable.MoveY(dy*0.5f);
			bottomAxisLable.MoveY(dy);

			leftAxis.MoveY(dy*0.5f);
			rightAxis.MoveY(dy*0.5f);

			bottomAxis.MoveY(dy);

			legend.MoveY(dy*0.5f);
			break;
	};

	// Propagate the bounding box change to all of the plots
	SetBoundingBox();

	return y;
}

float SGraph::MoveX(float m_dx)
{	
	switch(graph_move){
		case GRAPH_MOVE_ALL:
			x1 += m_dx;
			x2 += m_dx;

			title.MoveX(m_dx);
			leftAxisLable.MoveX(m_dx);
			bottomAxisLable.MoveX(m_dx);

			topAxis.MoveX(m_dx);
			bottomAxis.MoveX(m_dx);
			leftAxis.MoveX(m_dx);
			rightAxis.MoveX(m_dx);

			legend.MoveX(m_dx);

			x = 0.5f*(x2 + x1);
			break;
		case GRAPH_MOVE_LEFT_AXIS:
			x1 += m_dx;
			topAxis.line_length = bottomAxis.line_length = x2 - x1;

			title.MoveX(m_dx*0.5f);
			leftAxisLable.MoveX(m_dx);
			bottomAxisLable.MoveX(m_dx*0.5f);

			topAxis.MoveX(m_dx*0.5f);
			bottomAxis.MoveX(m_dx*0.5f);

			leftAxis.MoveX(m_dx);

			legend.MoveX(m_dx*0.5f);
			break;
		case GRAPH_MOVE_RIGHT_AXIS:
			x2 += m_dx;
			topAxis.line_length = bottomAxis.line_length = x2 - x1;

			title.MoveX(m_dx*0.5f);
			bottomAxisLable.MoveX(m_dx*0.5f);

			topAxis.MoveX(m_dx*0.5f);
			bottomAxis.MoveX(m_dx*0.5f);
			
			rightAxis.MoveX(m_dx);

			legend.MoveX(m_dx*0.5f);
			break;
		case GRAPH_MOVE_TOP_AXIS:
			return x;
		case GRAPH_MOVE_BOTTOM_AXIS:
			return x;
	};

	// Propagate the bounding box change to all of the plots
	SetBoundingBox();

	return x;
}

float SGraph::MoveY(float m_dy)
{
	switch(graph_move){
		case GRAPH_MOVE_ALL:
			y1 += m_dy;
			y2 += m_dy;

			title.MoveY(m_dy);
			leftAxisLable.MoveY(m_dy);
			bottomAxisLable.MoveY(m_dy);

			leftAxis.MoveY(m_dy);
			rightAxis.MoveY(m_dy);
			topAxis.MoveY(m_dy);
			bottomAxis.MoveY(m_dy);
			
			legend.MoveY(m_dy);

			y = 0.5f*(y2 + y1);
			break;
		case GRAPH_MOVE_LEFT_AXIS:
			return y;
		case GRAPH_MOVE_RIGHT_AXIS:
			return y;
		case GRAPH_MOVE_TOP_AXIS:
			y2 += m_dy;
			leftAxis.line_length = rightAxis.line_length = y2 - y1;

			title.MoveY(m_dy);
			leftAxisLable.MoveY(m_dy*0.5f);

			leftAxis.MoveY(m_dy*0.5f);
			rightAxis.MoveY(m_dy*0.5f);
			
			topAxis.MoveY(m_dy);

			legend.MoveY(m_dy*0.5f);
			break;
		case GRAPH_MOVE_BOTTOM_AXIS:
			y1 += m_dy;
			leftAxis.line_length = rightAxis.line_length = y2 - y1;

			leftAxisLable.MoveY(m_dy*0.5f);
			bottomAxisLable.MoveY(m_dy);

			leftAxis.MoveY(m_dy*0.5f);
			rightAxis.MoveY(m_dy*0.5f);

			bottomAxis.MoveY(m_dy);

			legend.MoveY(m_dy*0.5f);
			break;
	};

	// Propagate the bounding box change to all of the plots
	SetBoundingBox();

	return y;
}

void SGraph::PrepareMove()
{
	if(leftAxis.isSelected){
		graph_move = GRAPH_MOVE_LEFT_AXIS;
		return;
	}

	if(rightAxis.isSelected){
		graph_move = GRAPH_MOVE_RIGHT_AXIS;
		return;
	}
	
	if(topAxis.isSelected){
		graph_move = GRAPH_MOVE_TOP_AXIS;
		return;
	}

	if(bottomAxis.isSelected){
		graph_move = GRAPH_MOVE_BOTTOM_AXIS;
		return;
	}

	graph_move = GRAPH_MOVE_ALL;
}

// Overlap the current graph on top of m_graph.
void SGraph::Overlap(SGraph &m_graph)
{

	// Overlap the graphs
	x = m_graph.x;
	y = m_graph.y;

	x1 = m_graph.x1;
	y1 = m_graph.y1;
	x2 = m_graph.x2;
	y2 = m_graph.y2;
	
	// Overlap the associated graph text
	title.x = m_graph.title.x;
	title.y = m_graph.title.y;

	legend.x = m_graph.legend.x;
	legend.y = m_graph.legend.y;

	leftAxisLable.x = m_graph.leftAxisLable.x;
	leftAxisLable.y = m_graph.leftAxisLable.y;

	bottomAxisLable.x = m_graph.bottomAxisLable.x;
	bottomAxisLable.y = m_graph.bottomAxisLable.y;

	// Overlap the left, right, top and bottom axes
	leftAxis.x = m_graph.leftAxis.x;
	leftAxis.y = m_graph.leftAxis.y;
	leftAxis.line_length = m_graph.leftAxis.line_length;	

	rightAxis.x = m_graph.rightAxis.x;
	rightAxis.y = m_graph.rightAxis.y;
	rightAxis.line_length = m_graph.rightAxis.line_length;

	topAxis.x = m_graph.topAxis.x;
	topAxis.y = m_graph.topAxis.y;
	topAxis.line_length = m_graph.topAxis.line_length;

	bottomAxis.x = m_graph.bottomAxis.x;
	bottomAxis.y = m_graph.bottomAxis.y;
	bottomAxis.line_length = m_graph.bottomAxis.line_length;
}

void CSgraphApp::OnAppHelpOption() 
{
	CTextOutputDialog toDlg("Import file options",
		"The following commands will be read from the head of\r\n"
		"an imported text file. Note that 'i' refers to the zero based\r\n"
		"plot index\r\n"
		"______________________________________________\r\n"
		"#@set[i].point.style OPTION\r\n"
		"\tWhere OPTION is one of the following\r\n"
		"\t{circle, square, triangle, dot,\r\n\thistogramx, histogramy, vector, none}\r\n"
		"#@set[i].point.color OPTION\r\n"
		"\tWhere OPTION is one of the following\r\n"
		"\t{black, white, red, green, blue,\r\n\tcyan, grey, none}\r\n"
		"#@set[i].point.size OPTION\r\n"
		"\tWhere OPTION is an integer,\r\n\t0 <= OPTION <= 10\r\n"
		"#@set[i].line.style OPTION\r\n"
		"\tWhere OPTION is one of the following\r\n\t"
		"{solid, dotted, dashed, dashdot,\r\n\tdashdotdot, none}\r\n"
		"#@set[i].line.color OPTION\r\n"
		"\tWhere OPTION is one of the following\r\n\t"
		"{black, white, red, green, blue,\r\n\tcyan, grey, none}\r\n"
		"#@set[i].line.color custom  R G B\r\n"
		"\tWhere R, G abd B are integers and \r\n"
		"\t0 <= R <= 255, 0 <= G <= 255, 0 <= B <= 255.\r\n"
		"#@set[i].line.size OPTION\r\n"
		"\tWhere OPTION is an integer,\r\n\t0 <= OPTION <= 10\r\n"
		"#@set[i].errorbar.display OPTION\r\n"
		"\tWhere OPTION is one of the following\r\n\t"
		"{top, left, bottom, right, none}\r\n"
		"#@set[i].info TEXT\r\n"
		"\tWhere TEXT is any user defined text\r\n"
		"______________________________________________\r\n"
		"#@graph.title TEXT\r\n"
		"\tWhere TEXT is the graph title to use\r\n"
		"#@graph.xaxis.title TEXT\r\n"
		"\tWhere TEXT is the X axis title to use\r\n"
		"#@graph.yaxis.title TEXT\r\n"
		"\tWhere TEXT is the Y axis title to use\r\n");

	toDlg.DoModal();
}

void CSgraphApp::OnAppHelpCommand() 
{
	CTextOutputDialog helpDlg("Command Line Options",
		"The following options can be specified on the \r\n"
		"command line. [filename] can contain wildcard characters.\r\n\r\n"
		"presto [filename]\t\tOpen filename\r\n"
		"presto -nxy [filename]\tImport nxy data from filename\r\n"
		"presto -xy [filename]\t\tImport xy data from filename\r\n"
		"presto -xydx [filename]\tImport xydx data from filename\r\n"
		"presto -xydy [filename]\tImport xydy data from filename\r\n"
		"presto -xydxdy [filename]\tImport xydxdy data from filename\r\n"
		"presto -wav [filename]\tImport audio data from filename\r\n"
		".... -logx  ....\t\tUse a log scale for the x axis\r\n"
		".... -logy  ....\t\tUse a log scale for the y axis\r\n"
		".... -logxy ....\t\tUse a log scale for the x and y axes\r\n"
		"presto -help\t\tList the allowed command line options\r\n"
		"presto -?\t\t\tList the allowed command line options\r\n");

	helpDlg.DoModal();
}
